/***************************************
 *                                     *
 *  JBoss: The OpenSource J2EE WebOS   *
 *                                     *
 *  Distributable under LGPL license.  *
 *  See terms of license at gnu.org.   *
 *                                     *
 ***************************************/

package org.jboss.aop;

import EDU.oswego.cs.dl.util.concurrent.ConcurrentReaderHashMap;
import EDU.oswego.cs.dl.util.concurrent.CopyOnWriteArraySet;
import javassist.CtClass;
import javassist.CtConstructor;
import javassist.CtField;
import javassist.CtMethod;
import org.jboss.aop.advice.AdviceBinding;
import org.jboss.aop.advice.AspectDefinition;
import org.jboss.aop.advice.Interceptor;
import org.jboss.aop.annotation.AnnotationElement;
import org.jboss.aop.annotation.AnnotationRepository;
import org.jboss.aop.introduction.InterfaceIntroduction;
import org.jboss.aop.joinpoint.Joinpoint;
import org.jboss.aop.metadata.ClassMetaDataBinding;
import org.jboss.aop.metadata.ConstructorMetaData;
import org.jboss.aop.metadata.FieldMetaData;
import org.jboss.aop.metadata.MethodMetaData;
import org.jboss.aop.metadata.SimpleMetaData;

import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * @author <a href="mailto:bill@jboss.org">Bill Burke</a>
 * @version $Revision: 1.39.2.4 $
 */
public abstract class Advisor
{
   protected HashSet adviceBindings = new HashSet();
   protected ArrayList interfaceIntroductions = new ArrayList();
   protected ArrayList classMetaDataBindings = new ArrayList();
   protected SimpleMetaData defaultMetaData = new SimpleMetaData();
   protected MethodMetaData methodMetaData = new MethodMetaData();
   protected FieldMetaData fieldMetaData = new FieldMetaData();
   protected SimpleMetaData classMetaData = new SimpleMetaData();
   protected ConstructorMetaData constructorMetaData = new ConstructorMetaData();
   protected HashMap classAnnotations = new HashMap();
   protected AnnotationRepository annotations = new AnnotationRepository();
   public boolean doesHaveAspects = false;

   protected String name;
   protected ConcurrentReaderHashMap aspects = new ConcurrentReaderHashMap();
   protected HashMap adviceInterceptors = new HashMap();
   protected CopyOnWriteArraySet perInstanceAspectDefinitions = new CopyOnWriteArraySet();
   protected ConcurrentReaderHashMap perInstanceJoinpointAspectDefinitions = new ConcurrentReaderHashMap();

   static Class cl = java.lang.String.class;
   public Advisor(String name)
   {
      this.name = name;
   }

   public List getClassMetadataBindings()
   {
      return classMetaDataBindings;
   }

   public SimpleMetaData getClassMetaData()
   {
      return classMetaData;
   }

   public SimpleMetaData getDefaultMetaData()
   {
      return defaultMetaData;
   }

   public MethodMetaData getMethodMetaData()
   {
      return methodMetaData;
   }

   public FieldMetaData getFieldMetaData()
   {
      return fieldMetaData;
   }

   public ConstructorMetaData getConstructorMetaData()
   {
      return constructorMetaData;
   }

   public AnnotationRepository getAnnotations()
   {
      return annotations;
   }

   public HashMap getClassAnnotations()
   {
      return classAnnotations;
   }

   public Object resolveAnnotation(Class annotation)
   {
      return classAnnotations.get(annotation.getName());
   }

   public Object resolveAnnotation(Method m, Class annotation)
   {
      Object value = annotations.resolveAnnotation(m, annotation);
      if (value == null) value = AnnotationElement.getAnyAnnotation(m, annotation);
      return value;
   }

   public Object resolveAnnotation(Field f, Class annotation)
   {
      Object value = annotations.resolveAnnotation(f, annotation);
      if (value == null) value = AnnotationElement.getAnyAnnotation(f, annotation);
      return value;
   }

   public Object resolveAnnotation(Constructor c, Class annotation)
   {
      Object value = annotations.resolveAnnotation(c, annotation);
      if (value == null) value = AnnotationElement.getAnyAnnotation(c, annotation);
      return value;
   }

   public boolean hasAnnotation(Method m, String annotation)
   {
      if (annotations.hasAnnotation(m, annotation)) return true;
      try
      {
         return AnnotationElement.isAnyAnnotationPresent(m, annotation);
      }
      catch (Exception e)
      {
         throw new RuntimeException(e);  //To change body of catch statement use Options | File Templates.
      }
   }

   public boolean hasAnnotation(Field m, String annotation)
   {
      if (annotations.hasAnnotation(m, annotation)) return true;
      try
      {
         return AnnotationElement.isAnyAnnotationPresent(m, annotation);
      }
      catch (Exception e)
      {
         throw new RuntimeException(e);  //To change body of catch statement use Options | File Templates.
      }
   }

   public boolean hasAnnotation(Constructor m, String annotation)
   {
      if (annotations.hasAnnotation(m, annotation)) return true;
      try
      {
         return AnnotationElement.isAnyAnnotationPresent(m, annotation);
      }
      catch (Exception e)
      {
         throw new RuntimeException(e);  //To change body of catch statement use Options | File Templates.
      }
   }

   public boolean hasAnnotation(CtClass clazz, String annotation)
   {
      if (classAnnotations.containsKey(annotation)) return true;
      try
      {
         return AnnotationElement.isAnyAnnotationPresent(clazz, annotation);
      }
      catch (Exception e)
      {
         throw new RuntimeException(e);  //To change body of catch statement use Options | File Templates.
      }
   }

   public boolean hasAnnotation(String annotation)
   {
      return classAnnotations.containsKey(annotation);
   }

   public boolean hasAnnotation(CtMethod member, String annotation)
   {
      // todo these are here so that we can chain configuration domains
      if (annotations.hasAnnotation(member, annotation)) return true;
      return AnnotationElement.isAnyAnnotationPresent(member, annotation);
   }

   public boolean hasAnnotation(CtField member, String annotation)
   {
      // todo these are here so that we can chain configuration domains
      if (annotations.hasAnnotation(member, annotation)) return true;
      return AnnotationElement.isAnyAnnotationPresent(member, annotation);
   }

   public boolean hasAnnotation(CtConstructor member, String annotation)
   {
      // todo these are here so that we can chain configuration domains
      if (annotations.hasAnnotation(member, annotation)) return true;
      return AnnotationElement.isAnyAnnotationPresent(member, annotation);
   }

   public String getName() { return name; }

   public final boolean hasAspects()
   {
      return doesHaveAspects;
   }

   public synchronized void removeAdviceBinding(AdviceBinding binding)
   {
      adviceBindings.remove(binding);
      rebuildInterceptors();
      doesHaveAspects = adviceBindings.size() > 0;
   }

   /**
    * a new binding has been added to the AspectManager, recalculate interceptors
    */
   public synchronized void newBindingAdded()
   {
      rebuildInterceptors();
      doesHaveAspects = adviceBindings.size() > 0;
   }

   public ArrayList getInterfaceIntroductions()
   {
      return interfaceIntroductions;
   }

   public synchronized void addInterfaceIntroduction(InterfaceIntroduction pointcut)
   {
      interfaceIntroductions.add(pointcut);
   }

   public synchronized void removeInterfaceIntroduction(InterfaceIntroduction pointcut)
   {
      interfaceIntroductions.remove(pointcut);
   }

   protected abstract void rebuildInterceptors();

   ////////////////////////////////
   // Metadata.  Metadata will be used for things like Transaction attributes (Required, RequiresNew, etc...)
   //

   public abstract void addClassMetaData(ClassMetaDataBinding data);

   public abstract void removeClassMetaData(ClassMetaDataBinding data);

   // This is aspect stuff.  Aspect again, is a class that encapsulates advices

   public void addPerInstanceAspect(AspectDefinition def)
   {
      perInstanceAspectDefinitions.add(def);
   }

   public void removePerInstanceAspect(AspectDefinition def)
   {
      perInstanceAspectDefinitions.remove(def);
   }

   public Set getPerInstanceAspectDefinitions() { return perInstanceAspectDefinitions; }

   // This is aspect stuff.  Aspect again, is a class that encapsulates advices

   public void addPerInstanceJoinpointAspect(Joinpoint joinpoint, AspectDefinition def)
   {
      Set joinpoints = (Set)perInstanceJoinpointAspectDefinitions.get(def);
      if (joinpoints == null)
      {
         joinpoints = new CopyOnWriteArraySet();
         perInstanceJoinpointAspectDefinitions.put(def, joinpoints);
      }
      joinpoints.add(joinpoint);
   }

   public void removePerInstanceJoinpointAspect(AspectDefinition def)
   {
     perInstanceJoinpointAspectDefinitions.remove(def);
   }

   public Map getPerInstanceJoinpointAspectDefinitions() { return perInstanceJoinpointAspectDefinitions; }

   public Object getPerClassAspect(AspectDefinition def)
   {
      return aspects.get(def.getName());
   }

   public Object getPerClassAspect(String def)
   {
      return aspects.get(def);
   }

   public void addPerClassAspect(AspectDefinition def)
   {
      if (aspects.containsKey(def.getName())) return;
      Object aspect = def.getFactory().createPerClass(this);
      aspects.put(def.getName(), aspect);
   }

   public void removePerClassAspect(AspectDefinition def)
   {
      aspects.remove(def.getName());
      adviceInterceptors.remove(def);
   }

   public Interceptor getAdviceInterceptor(AspectDefinition def, String adviceName)
   {
      synchronized (adviceInterceptors)
      {
         Map map = null;
         if (adviceInterceptors.containsKey(def))
         {
            map = (Map) adviceInterceptors.get(def);
            if (map != null)
            {
               if (map.containsKey(adviceName))
               {
                  return (Interceptor) map.get(adviceName);
               }
            }
            else
            {
               map = new HashMap();
               adviceInterceptors.put(def, map);
            }
         }
         else
         {
            map = new HashMap();
            adviceInterceptors.put(def, map);
         }
      }
      return null;
   }

   public void addAdviceInterceptor(AspectDefinition def, String adviceName, Interceptor interceptor)
   {
      synchronized (adviceInterceptors)
      {
         Map map = (Map) adviceInterceptors.get(def);
         map.put(adviceName, interceptor);
      }
   }


}
