/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.aop;

import java.io.DataInputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.InputStream;
import java.io.StringReader;
import java.io.StringWriter;
import java.util.Iterator;

import javassist.bytecode.AnnotationsAttribute;
import javassist.bytecode.ClassFile;
import javassist.bytecode.FieldInfo;
import javassist.bytecode.MethodInfo;
import javassist.bytecode.annotation.ArrayMemberValue;
import javassist.bytecode.annotation.BooleanMemberValue;
import javassist.bytecode.annotation.ClassMemberValue;
import javassist.bytecode.annotation.MemberValue;
import javassist.bytecode.annotation.StringMemberValue;

import org.jboss.aop.advice.AdviceBinding;
import org.jboss.aop.advice.AdviceFactory;
import org.jboss.aop.advice.AspectDefinition;
import org.jboss.aop.advice.AspectFactory;
import org.jboss.aop.advice.AspectFactoryDelegator;
import org.jboss.aop.advice.DynamicCFlowDefinition;
import org.jboss.aop.advice.GenericAspectFactory;
import org.jboss.aop.advice.Interceptor;
import org.jboss.aop.advice.InterceptorFactory;
import org.jboss.aop.advice.Scope;
import org.jboss.aop.advice.ScopedInterceptorFactory;
import org.jboss.aop.annotation.AnnotationProxy;
import org.jboss.aop.introduction.AnnotationIntroduction;
import org.jboss.aop.introduction.InterfaceIntroduction;
import org.jboss.aop.pointcut.CFlow;
import org.jboss.aop.pointcut.CFlowStack;
import org.jboss.aop.pointcut.DynamicCFlow;
import org.jboss.aop.pointcut.Pointcut;
import org.jboss.aop.pointcut.PointcutExpression;
import org.jboss.aop.pointcut.Typedef;
import org.jboss.aop.pointcut.TypedefExpression;
import org.jboss.aop.pointcut.ast.ASTCFlowExpression;
import org.jboss.aop.pointcut.ast.ASTStart;
import org.jboss.aop.pointcut.ast.PointcutExpressionParser;
import org.jboss.aop.pointcut.ast.TypeExpressionParser;
import org.jboss.aop.util.XmlHelper;

/**
 * Comment
 *
 * @author <a href="mailto:bill@jboss.org">Bill Burke</a>
 * @version $Revision: 1.2.2.5 $
 */
public class AspectAnnotationLoader
{
   //TODO: We need something to undeploy everything...
   
   public void deployInputStreamIterator(Iterator it) throws Exception
   {
      while (it.hasNext())
      {
         InputStream stream = (InputStream) it.next();
         DataInputStream dstream = new DataInputStream(stream);
         ClassFile cf = null;
         try
         {
            cf = new ClassFile(dstream);
         }
         finally
         {
            dstream.close();
            stream.close();
         }
         if (AspectManager.verbose) System.out.println("[debug] Looking for aspects in: " + cf.getName());
         AnnotationsAttribute visible = (AnnotationsAttribute) cf.getAttribute(AnnotationsAttribute.visibleTag);
         if (visible != null)
         {
            AspectDefinition def = deployAspect(visible, cf);
            
            if (def == null)
            {
               def = deployInterceptor(visible, cf);
            }
            
            if (def == null)
            {
               deployDynamicCFlow(visible, cf);
            }
            
            if (def != null)
            {
      	      deployPointcuts(cf);
               deployMixins(cf);
               deployIntroductions(cf);
               deployTypedefs(cf);
               deployCFlowStackDefs(cf);
               deployPrepares(cf);
               deployAnnotationIntroductions(cf);
            }
         }
      }
   }

   private AspectDefinition deployAspect(AnnotationsAttribute visible, ClassFile cf)throws Exception
   {
      //Check for Aspect
      javassist.bytecode.annotation.Annotation info = visible.getAnnotation(Aspect.class.getName());
      if (info != null)
      {
	      if (AspectManager.verbose) System.out.println("[debug] Found @Aspect in: " + cf.getName());
	      Aspect aspect = (Aspect) AnnotationProxy.createProxy(info, Aspect.class);
	      Scope scope = aspect.scope();
	      AspectDefinition def = new AspectDefinition(cf.getName(), scope, new GenericAspectFactory(cf.getName(), null));
	      AspectManager.instance().addAspectDefinition(def);

         deployAspectMethodBindings(cf, def);

         return def;
      }
      return null;
   }

   private AspectDefinition deployInterceptor(AnnotationsAttribute visible, ClassFile cf)throws Exception
   {
      //Check for InterceptorDef
      javassist.bytecode.annotation.Annotation info = visible.getAnnotation(InterceptorDef.class.getName());
      if (info != null)
      {
	      if (AspectManager.verbose) System.out.println("[debug] Found @InterceptorDef in: " + cf.getName());
	      Aspect aspect = (Aspect) AnnotationProxy.createProxy(info, Aspect.class);
	      Scope scope = aspect.scope();
	      
	      String[] interfaces = cf.getInterfaces();
	      boolean isFactory = false;
	      for (int i = 0 ; i < interfaces.length ; i++)
	      {
	         if (interfaces[i].equals(AspectFactory.class.getName()))
	         {
	            isFactory = true;
	            break;
	         }
	         else if (interfaces[i].equals(Interceptor.class.getName()))
	         {
	            break;
	         }
	      }

	      AspectFactory aspectFactory;
	      if (isFactory)
	      {
            aspectFactory = new AspectFactoryDelegator(cf.getName(), null);	         
	      }
	      else
	      {
            aspectFactory = new GenericAspectFactory(cf.getName(), null);	         
	      }
	      
	      AspectDefinition def = new AspectDefinition(cf.getName(), scope, aspectFactory);
         AspectManager.instance().addAspectDefinition(def);
         ScopedInterceptorFactory factory = new ScopedInterceptorFactory(def);
         AspectManager.instance().addInterceptorFactory(factory.getName(), factory);
         
         deployInterceptorBindings(visible, cf, factory);

	      return def;
      }
      
      return null;
   }
   
   private void deployDynamicCFlow(AnnotationsAttribute visible, ClassFile cf) throws Exception
   {
      javassist.bytecode.annotation.Annotation info = visible.getAnnotation(DynamicCFlowDef.class.getName());
      if (info != null)
      {
	      if (AspectManager.verbose) System.out.println("[debug] Found @InterceptorDef in: " + cf.getName());
	      DynamicCFlowDef dynCFlow = (DynamicCFlowDef) AnnotationProxy.createProxy(info, DynamicCFlowDef.class);

         String name = cf.getName();
         String clazz = cf.getName();
	      
	      String[] interfaces = cf.getInterfaces();
	      boolean foundDCFlow = false;
	      for (int i = 0 ; i < interfaces.length ; i++)
	      {
	         if (interfaces[i].equals(DynamicCFlow.class.getName()))
	         {
	            foundDCFlow = true;
	            break;
	         }
	      }	         
	      if (!foundDCFlow) throw new RuntimeException("@DynamicCFlow annotated class: " + clazz + " must implement " + DynamicCFlow.class.getName());
	      
	      AspectManager.instance().addDynamicCFlow(name, new DynamicCFlowDefinition(null, clazz, name));
      }
   }
   

   private void deployAspectMethodBindings(ClassFile cf, AspectDefinition def)
           throws Exception
   {
      Iterator methods = cf.getMethods().iterator();
      while (methods.hasNext())
      {
         javassist.bytecode.MethodInfo minfo = (javassist.bytecode.MethodInfo) methods.next();
         AnnotationsAttribute mgroup = (AnnotationsAttribute) minfo.getAttribute(AnnotationsAttribute.visibleTag);
         if (mgroup == null) continue;
         javassist.bytecode.annotation.Annotation binfo = mgroup.getAnnotation(Bind.class.getName());
         if (binfo == null) continue;
         Bind binding = (Bind) AnnotationProxy.createProxy(binfo, Bind.class);
         String pointcutString = binding.pointcut();
         String cflow = binding.cflow();
         if (cflow == null || cflow.trim().equals("")) cflow = null;
         ASTCFlowExpression cflowExpression = null;
         if (cflow != null)
         {
            cflowExpression = new PointcutExpressionParser(new StringReader(cflow)).CFlowExpression();

         }
         AdviceFactory factory = new AdviceFactory(def, minfo.getName());
         AspectManager.instance().addInterceptorFactory(factory.getName(), factory);
         InterceptorFactory[] fact = {factory};
         String name = cf.getName() + "." + minfo.getName() + " " + minfo.hashCode();
         PointcutExpression pointcut = new PointcutExpression(name, pointcutString);
         AspectManager.instance().addPointcut(pointcut);
         AdviceBinding abinding = new AdviceBinding(name, pointcut, cflowExpression, cflow, fact);
         AspectManager.instance().addBinding(abinding);
      }
   }

   private void deployInterceptorBindings(AnnotationsAttribute visible, ClassFile cf, InterceptorFactory factory)
   			throws Exception
   {
      javassist.bytecode.annotation.Annotation binfo = visible.getAnnotation(Bind.class.getName());
      Bind bind = (Bind) AnnotationProxy.createProxy(binfo, Bind.class);
      String pointcutString = bind.pointcut();
      String cflow = bind.cflow();
      if (cflow == null || cflow.trim().equals("")) cflow = null;
      ASTCFlowExpression cflowExpression = null;
      if (cflow != null)
      {
         cflowExpression = new PointcutExpressionParser(new StringReader(cflow)).CFlowExpression();

      }

      String name = cf.getName();
      InterceptorFactory[] inters = {factory};
      Pointcut p = null;
      p = new PointcutExpression(name, pointcutString);
      AspectManager.instance().addPointcut(p);

      AdviceBinding binding = new AdviceBinding(name, p, cflowExpression, cflow, inters);
      AspectManager.instance().addBinding(binding);
   }
   
   
   private void deployPointcuts(ClassFile cf)
           throws Exception
   {
      Iterator fields = cf.getFields().iterator();
      while (fields.hasNext())
      {
         javassist.bytecode.FieldInfo finfo = (javassist.bytecode.FieldInfo) fields.next();
         AnnotationsAttribute mgroup = (AnnotationsAttribute) finfo.getAttribute(AnnotationsAttribute.visibleTag);
         if (mgroup == null) continue;
         javassist.bytecode.annotation.Annotation binfo = mgroup.getAnnotation(PointcutDef.class.getName());
         if (binfo == null) continue;
         PointcutDef pdef = (PointcutDef) AnnotationProxy.createProxy(binfo, PointcutDef.class);

         PointcutExpression pointcut = new PointcutExpression(cf.getName() + "." + finfo.getName(), pdef.value());

         AspectManager.instance().addPointcut(pointcut);
      }
   }

   private void deployMixins(ClassFile cf)
   throws Exception
   {
      Iterator methods = cf.getMethods().iterator();
      while (methods.hasNext())
      {
         javassist.bytecode.MethodInfo minfo = (javassist.bytecode.MethodInfo) methods.next();
         AnnotationsAttribute mgroup = (AnnotationsAttribute) minfo.getAttribute(AnnotationsAttribute.visibleTag);
         if (mgroup == null) continue;
         javassist.bytecode.annotation.Annotation binfo = mgroup.getAnnotation(Mixin.class.getName());
         if (binfo == null) continue;
         
         //Since some of the values are of type Class, and method gets called by the system classloader (via Agent)
         // for loadtime aop in jdk 1.5 the classes we try to look at might not yet have been loaded. We're 
         //only after the names anyway, so we bypass the AnnotationProxy/ProxyMapCreator which tries to validate
         //class names by loading the (so far) non-existent class.
         
         //Mixin mixin = (Mixin) AnnotationProxy.createProxy(binfo, Mixin.class);
         //String target = mixin.target();
         //String typeExpression = mixin.typeExpression();
         //String[] interfaces = mixin.interfaces();
         //boolean isTransient = mixin.isTransient();
         
         MemberValue mv = binfo.getMemberValue("target");
         String target = (mv != null) ? ((ClassMemberValue)mv).getValue() : "java.lang.Class";//Note! this should be the same as the default in @Mixin 
         
         mv = binfo.getMemberValue("typeExpression");
         String  typeExpression = (mv != null) ? ((StringMemberValue)mv).getValue() : "";//Note! this should be the same as the default in @Mixin
         
         mv = binfo.getMemberValue("interfaces");
         MemberValue[] values = ((ArrayMemberValue)mv).getValue();
         String[] interfaces = new String[values.length];
         for (int i = 0 ; i < values.length ; i++) interfaces[i] = ((ClassMemberValue)values[i]).getValue();
         
         mv = binfo.getMemberValue("isTransient");
         boolean isTransient = (mv != null) ? ((BooleanMemberValue)mv).getValue() : true;//Note! this should be the same as the default in @Mixin 

         String name = cf.getName() + "." + minfo.getName(); //Name of the method defined on

         InterfaceIntroduction intro = createIntroduction(name, target, typeExpression, null);

         
         String construction = name + "(this)";
         
         //Parse the descriptor to get the returntype of the method.
         String classname = getReturnType(minfo);
         intro.getMixins().add(new InterfaceIntroduction.Mixin(classname, interfaces, construction, isTransient));

         AspectManager.instance().addInterfaceIntroduction(intro);
      }
   }

   private void deployIntroductions(ClassFile cf)
   throws Exception
   {
      Iterator fields = cf.getFields().iterator();
      while (fields.hasNext())
      {
         FieldInfo finfo = (javassist.bytecode.FieldInfo) fields.next();
         AnnotationsAttribute mgroup = (AnnotationsAttribute) finfo.getAttribute(AnnotationsAttribute.visibleTag);
         if (mgroup == null) continue;
         javassist.bytecode.annotation.Annotation binfo = mgroup.getAnnotation(Introduction.class.getName());
         if (binfo == null) continue;
         
         //Since some of the values are of type Class, and method gets called by the system classloader (via Agent)
         // for loadtime aop in jdk 1.5 the classes we try to look at might not yet have been loaded. We're 
         //only after the names anyway, so we bypass the AnnotationProxy/ProxyMapCreator which tries to validate
         //class names by loading the (so far) non-existent class.
         
         //Introduction introduction = (Introduction) AnnotationProxy.createProxy(binfo, Introduction.class);
         //String target = introduction.target();
         //String  typeExpression = introduction.typeExpression();
         //String[] interfaces = introduction.interfaces();
         
         MemberValue mv = binfo.getMemberValue("target");
         String target = (mv != null) ? ((ClassMemberValue)mv).getValue() : "java.lang.Class";//Note! this should be the same as the default in @Interceptor 
         
         mv = binfo.getMemberValue("typeExpression");
         String  typeExpression = (mv != null) ? ((StringMemberValue)mv).getValue() : "";//Note! this should be the same as the default in @Interceptor
         
         mv = binfo.getMemberValue("interfaces");
         MemberValue[] values = ((ArrayMemberValue)mv).getValue();
         String[] interfaces = new String[values.length];
         for (int i = 0 ; i < values.length ; i++) interfaces[i] = ((ClassMemberValue)values[i]).getValue();

         String name = cf.getName() + "." + finfo.getName(); //Name of the field defined on
         
         InterfaceIntroduction interfaceIntro = createIntroduction(name, target, typeExpression, interfaces);
         AspectManager.instance().addInterfaceIntroduction(interfaceIntro);
      }
   }

   private void deployTypedefs(ClassFile cf)throws Exception
   {
      Iterator fields = cf.getFields().iterator();
      while (fields.hasNext())
      {
         FieldInfo finfo = (javassist.bytecode.FieldInfo) fields.next();
         AnnotationsAttribute mgroup = (AnnotationsAttribute) finfo.getAttribute(AnnotationsAttribute.visibleTag);
         if (mgroup == null) continue;
         javassist.bytecode.annotation.Annotation binfo = mgroup.getAnnotation(TypeDef.class.getName());
         if (binfo == null) continue;
         TypeDef typeDefinition = (TypeDef) AnnotationProxy.createProxy(binfo, TypeDef.class);
 
         String name = cf.getName() + "." + finfo.getName(); 
         String expr = typeDefinition.value();
         Typedef typedef = new TypedefExpression(name, expr);
         AspectManager.instance().addTypedef(typedef);
         
      }
   }
   
   private void deployCFlowStackDefs(ClassFile cf)throws Exception
   {
      Iterator fields = cf.getFields().iterator();
      while (fields.hasNext())
      {
         FieldInfo finfo = (javassist.bytecode.FieldInfo) fields.next();
         AnnotationsAttribute mgroup = (AnnotationsAttribute) finfo.getAttribute(AnnotationsAttribute.visibleTag);
         if (mgroup == null) continue;
         javassist.bytecode.annotation.Annotation binfo = mgroup.getAnnotation(CFlowStackDef.class.getName());
         if (binfo == null) continue;
         CFlowStackDef stackDef = (CFlowStackDef) AnnotationProxy.createProxy(binfo, CFlowStackDef.class);
 
         String name = cf.getName() + "." + finfo.getName();
         CFlowDef[] cflows = stackDef.cflows();
         CFlowStack stack = new CFlowStack(name);

         for (int i = 0 ; i < cflows.length ; i++)
         {
            CFlowDef cflow = cflows[i];
            boolean not = !cflow.called();
            stack.addCFlow(new CFlow(cflow.expr(), not));
         }

         AspectManager.instance().addCFlowStack(stack);
      }
   }
   
   private void deployPrepares(ClassFile cf) throws Exception
   {
      Iterator fields = cf.getFields().iterator();
      while (fields.hasNext())
      {
         FieldInfo finfo = (javassist.bytecode.FieldInfo) fields.next();
         AnnotationsAttribute mgroup = (AnnotationsAttribute) finfo.getAttribute(AnnotationsAttribute.visibleTag);
         if (mgroup == null) continue;
         javassist.bytecode.annotation.Annotation binfo = mgroup.getAnnotation(Prepare.class.getName());
         if (binfo == null) continue;
         Prepare prepare = (Prepare) AnnotationProxy.createProxy(binfo, Prepare.class);

         String name = cf.getName() + "." + finfo.getName(); 
         String expr = prepare.value();
         Pointcut p = new PointcutExpression(name, expr);
         AspectManager.instance().addPointcut(p);
      }
   }
   
   private void deployAnnotationIntroductions(ClassFile cf) throws Exception
   {
      Iterator fields = cf.getFields().iterator();
      while (fields.hasNext())
      {
         FieldInfo finfo = (javassist.bytecode.FieldInfo) fields.next();
         AnnotationsAttribute mgroup = (AnnotationsAttribute) finfo.getAttribute(AnnotationsAttribute.visibleTag);
         if (mgroup == null) continue;
         javassist.bytecode.annotation.Annotation binfo = mgroup.getAnnotation(AnnotationIntroductionDef.class.getName());
         if (binfo == null) continue;
         AnnotationIntroductionDef intro = (AnnotationIntroductionDef) AnnotationProxy.createProxy(binfo, AnnotationIntroductionDef.class);
         
         String expr = intro.expr();
         boolean invisible = intro.invisible();
         String annotation = intro.annotation();
         
         annotation = annotation.replace('\'', '"');
         
         AnnotationIntroduction annIntro = AnnotationIntroduction.createComplexAnnotationIntroduction(expr, annotation, invisible);
         AspectManager.instance().addAnnotationIntroduction(annIntro);
      }
   }
      
   private InterfaceIntroduction createIntroduction(
         String name, String target, String typeExpression, String[] interfaces)
   		throws Exception
   {
      if (typeExpression != null && typeExpression.trim().equals(""))
      {
         typeExpression = null;
      }
      
      if (typeExpression != null && target!= null && target.equals("java.lang.Class"))
      {
         target = null;
      }
      
      if (target == null && typeExpression == null)
      {
         throw new RuntimeException("No target nor a typeExpression attribute is defined for this @Mixin");
      }

      if (target == null && typeExpression == null)
      {
         throw new RuntimeException("You cannot define both a target and typeExpression attribute in the same @Mixin");
      }
      

      InterfaceIntroduction intro = null;
      
      if (target != null)
      {
         intro = new InterfaceIntroduction(name, target, interfaces);
      }
      else
      {
         ASTStart start = new TypeExpressionParser(new StringReader(typeExpression)).Start();
         intro = new InterfaceIntroduction(name, start, interfaces);
      }
      
      return intro;
   }
   
   
   private String[] makeClassnameArray(Class[] classes)
   {
      String[] names = new String[classes.length]; 
      for (int i = 0 ; i < classes.length ; i++)
      {
         names[i] = classes[i].getName();
      }   
      return names;
   }
   
   private String getReturnType(MethodInfo minfo)
   {
      String descriptor = minfo.getDescriptor();
      int paramsEnd = descriptor.indexOf(";)");
      String classname = descriptor.substring(paramsEnd + 3, descriptor.length() -1);
      classname = classname.replace('/', '.');
      return classname;
   }
   
}
