/***************************************
 *                                     *
 *  JBoss: The OpenSource J2EE WebOS   *
 *                                     *
 *  Distributable under LGPL license.  *
 *  See terms of license at gnu.org.   *
 *                                     *
 ***************************************/

package org.jboss.aop;

import EDU.oswego.cs.dl.util.concurrent.ConcurrentReaderHashMap;
import javassist.ByteArrayClassPath;
import javassist.CannotCompileException;
import javassist.ClassPool;
import javassist.CtClass;
import javassist.LoaderClassPath;
import javassist.NotFoundException;
import javassist.expr.ExprEditor;
import javassist.expr.MethodCall;
import org.jboss.aop.advice.AdviceBinding;
import org.jboss.aop.advice.AdviceStack;
import org.jboss.aop.advice.AspectDefinition;
import org.jboss.aop.advice.DynamicCFlowDefinition;
import org.jboss.aop.advice.InterceptorFactory;
import org.jboss.aop.advice.Scope;
import org.jboss.aop.instrument.Instrumentor;
import org.jboss.aop.introduction.AnnotationIntroduction;
import org.jboss.aop.introduction.InterfaceIntroduction;
import org.jboss.aop.metadata.ClassMetaDataBinding;
import org.jboss.aop.metadata.ClassMetaDataLoader;
import org.jboss.aop.metadata.SimpleClassMetaDataLoader;
import org.jboss.aop.pointcut.CFlowStack;
import org.jboss.aop.pointcut.DynamicCFlow;
import org.jboss.aop.pointcut.Pointcut;
import org.jboss.aop.pointcut.Typedef;
import org.jboss.util.loading.Translatable;
import org.jboss.util.loading.Translator;

import java.io.IOException;
import java.security.ProtectionDomain;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Map;

/**
 * This singleton class manages all pointcuts and metadata.
 * Coders can access it via the AspectManager.instance() method.
 * <p/>
 * It is also the middle man between the ClassLoader and
 * the actual class instrumentation as well.
 * <p/>
 * App Developers that want to create and apply, interceptors,
 * pointcuts, or metadata at runtime can also use this object
 * to do that.
 *
 * @author <a href="mailto:bill@jboss.org">Bill Burke</a>
 * @version $Revision: 1.83.2.1 $
 */
public class AspectManager
        implements Translator
{
   private static ThreadLocal REENTRY = new ThreadLocal()
   {
      protected Object initialValue()
      {
         return Boolean.FALSE;
      }
   };

   private final CtClass advised;
   // Attributes ---------------------------------------------------
   protected ClassPool classpool;
   protected HashMap registeredCLs = new HashMap();
   protected HashMap ucl2classes = new HashMap();
   protected LinkedHashMap interfaceIntroductions = new LinkedHashMap();
   protected LinkedHashMap annotationIntroductions = new LinkedHashMap();
   protected LinkedHashMap bindings = new LinkedHashMap();
   protected LinkedHashMap typedefs = new LinkedHashMap();
   protected HashMap interceptorFactories = new HashMap();
   protected HashMap advisors = new HashMap();
   protected Hashtable classMetaDataLoaders = new Hashtable();
   protected HashMap interceptorStacks = new HashMap();
   protected ConcurrentReaderHashMap cflowStacks = new ConcurrentReaderHashMap();
   protected ConcurrentReaderHashMap dynamicCFlows = new ConcurrentReaderHashMap();
   protected ConcurrentReaderHashMap aspectDefinitions = new ConcurrentReaderHashMap();
   protected ConcurrentReaderHashMap perVMAspects = new ConcurrentReaderHashMap();

   protected LinkedHashMap pointcuts = new LinkedHashMap();
   protected LinkedHashMap classMetaData = new LinkedHashMap();


   // Static -------------------------------------------------------

   protected static AspectManager manager;
   public static boolean optimize = true;

   /**
    * logging switch.  We don't use log4j to avoid another heavy library
    */
   public static boolean verbose = false;
   public static ClassPoolFactory classPoolFactory = new AOPClassPoolFactory();

   public static synchronized AspectManager instance()
   {
      if (manager != null) return manager;
      String optimized = System.getProperty("jboss.aop.optimized", null);
      if (optimized != null)
      {
         optimize = (new Boolean(optimized)).booleanValue();
      }
      manager = new AspectManager();

      if (!verbose)
      {
         verbose = (new Boolean(System.getProperty("jboss.aop.verbose", "false"))).booleanValue();
      }
      Deployment.deploy();
      return manager;
   }


   public LinkedHashMap getPointcuts()
   {
      return pointcuts;
   }

   public CFlowStack getCFlowStack(String name)
   {
      return (CFlowStack) cflowStacks.get(name);
   }

   public void addCFlowStack(CFlowStack stack)
   {
      cflowStacks.put(stack.getName(), stack);
   }

   public void removeCFlowStack(String name)
   {
      cflowStacks.remove(name);
   }

   public DynamicCFlow getDynamicCFlow(String name)
   {
      DynamicCFlowDefinition def = (DynamicCFlowDefinition) dynamicCFlows.get(name);
      return def.create();
   }

   public void addDynamicCFlow(String name, DynamicCFlowDefinition cflow)
   {
      dynamicCFlows.put(name, cflow);
   }

   public void removeDynamicCFlow(String name)
   {
      dynamicCFlows.remove(name);
   }

   /**
    * This is a callback object that receives events about
    * new pointcuts, interceptors and metadata.
    * The Aspect Management console hooks into this listener
    */
   public static AspectNotificationHandler notificationHandler = null;

   public static boolean suppressTransformationErrors = false;

   // Constructors -------------------------------------------------
   public AspectManager()
   {
      classpool = ClassPool.getDefault();
      ClassLoader cl = Thread.currentThread().getContextClassLoader();
      classpool.insertClassPath(new LoaderClassPath(cl));
      try
      {
         advised = classpool.get("org.jboss.aop.Advised");
      }
      catch (NotFoundException ex)
      {
         throw new RuntimeException(ex);
      }
   }

   /**
    * Every &lt;class-metadata&gt; tag corresponds to
    * a ClassMetaDataLoader.  The ClassMetaDataLoader knows how to take
    * arbitrary XML and apply it to SimpleMetaData.
    * <p/>
    * Given a group, return the loader for that group
    */
   public ClassMetaDataLoader findClassMetaDataLoader(String group)
   {
      ClassMetaDataLoader loader = (ClassMetaDataLoader) classMetaDataLoaders.get(group);
      if (loader == null) loader = SimpleClassMetaDataLoader.singleton;
      return loader;
   }

   /**
    * Every &lt;class-metadata&gt; tag corresponds to
    * a ClassMetaDataLoader.  The ClassMetaDataLoader knows how to take
    * arbitrary XML and apply it to SimpleMetaData.
    * <p/>
    * Add a loader for a given group
    */
   public void addClassMetaDataLoader(String group, ClassMetaDataLoader loader)
   {
      classMetaDataLoaders.put(group, loader);
   }

   /**
    * Remove a loader for a given group
    */
   public void removeClassMetaDataLoader(String group)
   {
      classMetaDataLoaders.remove(group);
   }

   public HashMap getAdvisors()
   {
      return advisors;
   }

   public Advisor getAdvisor(String name)
   {
      synchronized (advisors)
      {
         return (Advisor) advisors.get(name);
      }
   }

   public LinkedHashMap getBindings()
   {
      return bindings;
   }

   public ClassAdvisor getAdvisorIfAdvised(Class clazz)
   {
      try
      {
         ClassAdvisor advisor;
         synchronized (advisors)
         {
            advisor = (ClassAdvisor) advisors.get(clazz.getName());
         }
         if (advisor == null)
         {
            return null;
         }
         advisor.attachClass(clazz);
         if (notificationHandler != null)
         {
            notificationHandler.attachClass(clazz.getName());
         }
         return advisor;
      }
      catch (RuntimeException ex)
      {
         ex.printStackTrace();
         throw ex;
      }
   }

   /**
    * This method is called by the aspectized class when it is loaded
    * This causes all initialization of interceptors for ClassAdvisor
    *
    * @param clazz
    * @return
    */
   public ClassAdvisor getAdvisor(Class clazz)
   {
      ClassAdvisor advisor = null;
      // See if one already exists
      synchronized (advisors)
      {
         advisor = (ClassAdvisor) advisors.get(clazz.getName());
      }
      // if one does not
      if (advisor == null)
      {
         advisor = new ClassAdvisor(clazz.getName());
         synchronized (advisors)
         {
            advisors.put(clazz.getName(), advisor);
         }

         // if it is Translatable let's assume it is a JBoss class loader
         // and let's register the classloader and the classname
         // We have to do this because we may be in the JBoss app server
         // and mixing and matching precompiled (aopc) vs. load-time transformations
         if (clazz.getClassLoader() instanceof Translatable)
         {
            registerClassLoader(clazz.getClassLoader());
            registerClassName(clazz.getClassLoader(), clazz.getName());
         }
      }
      advisor.attachClass(clazz);
      if (notificationHandler != null)
      {
         notificationHandler.attachClass(clazz.getName());
      }
      return advisor;
   }

   // Public -------------------------------------------------------

   public Map getRegisteredCLs()
   {
      return registeredCLs;
   }

   public ClassPool findClassPool(ClassLoader cl)
   {
      if (!(cl instanceof Translatable))
      {
         // findClassPool has problems with boot and system classes.
         return registerClassLoader(Thread.currentThread().getContextClassLoader());
      }
      return registerClassLoader(cl);
   }

   public ClassPool registerClassLoader(ClassLoader ucl)
   {
      synchronized (registeredCLs)
      {
         // FIXME: Probably want to take this method out later
         // so that AOP framework can be independent of JMX
         // This is in here so that we can remove a UCL from the ClassPool as a
         // ClassPool.classpath
         if (registeredCLs.containsKey(ucl))
         {
            return (ClassPool) registeredCLs.get(ucl);
         }
         AOPClassPool pool = AOPClassPool.createAOPClassPool(ucl, classpool, this);
         registeredCLs.put(ucl, pool);
         return pool;
      }
   }

   protected void registerClassName(ClassLoader ucl, String classname)
   {
      HashSet classes = (HashSet) ucl2classes.get(ucl);
      if (classes == null)
      {
         classes = new HashSet();
         ucl2classes.put(ucl, classes);
      }
      classes.add(classname);
   }

   public void unregisterClassLoader(ClassLoader cl)
   {
      synchronized (registeredCLs)
      {
         AOPClassPool pool = (AOPClassPool) registeredCLs.remove(cl);
         if (pool != null) pool.close();
         HashSet classes = (HashSet) ucl2classes.remove(cl);
         if (classes != null)
         {
            Iterator it = classes.iterator();
            while (it.hasNext())
            {
               Object clazz = it.next();
               synchronized (advisors)
               {
                  advisors.remove(clazz);
               }
            }
         }
      }
   }

   public boolean isNonAdvisableClassName(String classname)
   {
      return (classname.startsWith("org.jboss.aop") ||
              classname.endsWith("$aop") ||
              classname.startsWith("javassist") ||
              classname.startsWith("org.jboss.util.") ||
              classname.startsWith("gnu.trove.") ||
              classname.startsWith("EDU.oswego.cs.dl.util.concurrent.") ||
              // System classes
              classname.startsWith("org.apache.crimson") ||
              classname.startsWith("org.apache.xalan") ||
              classname.startsWith("org.apache.xml") ||
              classname.startsWith("org.apache.xpath") ||
              classname.startsWith("org.ietf.") ||
              classname.startsWith("org.omg.") ||
              classname.startsWith("org.w3c.") ||
              classname.startsWith("org.xml.sax.") ||
              classname.startsWith("sunw.") ||
              classname.startsWith("sun.") ||
              classname.startsWith("java.") ||
              classname.startsWith("javax.") ||
              classname.startsWith("com.sun.")
              );
   }

   /**
    * This is the hook for ClassLoaders that want to instrument their classes with AOP
    * <p/>
    * This would be called during a findClass or loadClass call.  The return value
    * is used by defineClass to create the class from bytecode
    */
   public byte[] transform(ClassLoader loader,
                           String className,
                           Class classBeingRedefined,
                           ProtectionDomain protectionDomain,
                           byte[] classfileBuffer)
           throws Exception
   {
      return translate(className, loader, classfileBuffer);
   }


   /**
    * This is to be backward compatible with JBoss 3.2.3 Translator interface
    *
    * @param className
    * @param loader
    * @return
    * @throws Exception
    */
   public byte[] translate(String className, ClassLoader loader) throws Exception
   {
      return translate(className, loader, null);
   }

   /**
    * This is to be backward compatible with JBoss 3.2.3 Translator interface
    *
    * @param className
    * @param loader
    * @return
    * @throws Exception
    */
   public byte[] translate(String className, ClassLoader loader, byte[] classfileBuffer) throws Exception
   {
      if (isReEntry())
      {
         return null;
      }
      setReEntry();
      try
      {
         if (isNonAdvisableClassName(className))
         {
            return null;
         }
         AOPClassPool pool = (AOPClassPool) registerClassLoader(loader);
         CtClass clazz = null;
         try
         {
            clazz = pool.getLocally(className);
         }
         catch (NotFoundException e)
         {
            // todo Bill Burke: this scares the shit out of me, but it must be done
            // I think it will screw up hotdeployment at some time.  Then again, maybe not ;)
            ByteArrayClassPath cp = new ByteArrayClassPath(className, classfileBuffer);
            pool.insertClassPath(cp);
            clazz = pool.getLocally(className);
         }
         if (clazz.isArray())
         {
            if (verbose) System.out.println("[cannot compile] isArray: " + className);
            pool.flushClass(className);
            return null;
         }
         if (clazz.isInterface())
         {
            if (verbose) System.out.println("[cannot compile] isInterface: " + className);
            pool.flushClass(className);
            return null;
         }
         if (clazz.isFrozen())
         {
            if (verbose) System.out.println("[warning] isFrozen: " + className);
            clazz.defrost();
         }

         ClassAdvisor advisor = new ClassAdvisor(className);
         Instrumentor instrumentor = new Instrumentor(pool);

         if (!instrumentor.isTransformable(clazz))
         {
            if (verbose) System.out.println("[cannot compile] implements Untransformable: " + className);
            pool.flushClass(className);
            return null;
         }

         boolean register = attachMetaData(advisor, clazz, true);
         applyInterfaceIntroductions(advisor, clazz);
         boolean transformed = instrumentor.transform(clazz, advisor);

         if (transformed) register = true;
         if (register)
         {
            synchronized (advisors)
            {
               advisors.put(className, advisor);
            }
            registerClassName(loader, className);
         }
         if (transformed)
         {
            return clazz.toBytecode();
         }
         return null;
      }
      catch (Exception ex)
      {
         if (!(ex instanceof NotFoundException))
         {
            if (verbose)
               ex.printStackTrace();
            else
               System.err.println("[error] " + ex.getMessage() + ".. Do verbose mode if you want full stack trace.");

         }
         throw ex;
      }
      finally
      {
         clearReEntry();
      }
   }

   /**
    * Add an interceptor factory that can be referenced by name.
    */
   public void addInterceptorFactory(String name, InterceptorFactory factory)
   {
      synchronized (interceptorFactories)
      {
         interceptorFactories.put(name, factory);
      }
   }

   /**
    * Remove an interceptor factory that can be referenced by name.
    */
   public void removeInterceptorFactory(String name)
   {
      synchronized (interceptorFactories)
      {
         interceptorFactories.remove(name);
      }
   }

   public Map getInterceptorFactories()
   {
      return interceptorFactories;
   }

   /**
    * Find the interceptor factory that can be referenced by name.
    */
   public InterceptorFactory getInterceptorFactory(String name)
   {
      synchronized (interceptorFactories)
      {
         return (InterceptorFactory) interceptorFactories.get(name);
      }
   }

   /**
    * Add a referencable InterceptorStack( &lt;stack&gt; )
    */
   public void addAdviceStack(AdviceStack stack)
   {
      synchronized (interceptorStacks)
      {
         interceptorStacks.put(stack.getName(), stack);
      }
   }

   /**
    * Remove a referencable InterceptorStack( &lt;stack&gt; )
    */
   public void removeInterceptorStack(String name)
   {
      synchronized (interceptorStacks)
      {
         interceptorStacks.remove(name);
      }
   }

   /**
    * Find an interceptor stack referenced by name ( &lt;stack&gt; )
    */
   public AdviceStack getAdviceStack(String name)
   {
      synchronized (interceptorStacks)
      {
         return (AdviceStack) interceptorStacks.get(name);
      }
   }


   protected boolean attachMetaData(ClassAdvisor advisor, CtClass clazz, boolean addAdvisor) throws Exception
   {
      boolean attached = false;
      synchronized (classMetaData)
      {
         Iterator it = classMetaData.values().iterator();
         while (it.hasNext())
         {
            ClassMetaDataBinding data = (ClassMetaDataBinding) it.next();
            if (data.matches(advisor, clazz))
            {
               attached = true;
               if (addAdvisor) data.addAdvisor(advisor);
               ClassMetaDataLoader loader = data.getLoader();
               loader.bind(advisor, data, clazz.getDeclaredMethods(), clazz.getDeclaredFields(), clazz.getDeclaredConstructors());
            }
         }
      }
      return attached;
   }

   protected void attachMetaData(ClassAdvisor advisor, Class clazz)
   {
      synchronized (classMetaData)
      {
         Iterator it = classMetaData.values().iterator();
         while (it.hasNext())
         {
            ClassMetaDataBinding data = (ClassMetaDataBinding) it.next();
            if (data.matches(advisor, clazz))
            {
               data.addAdvisor(advisor);
            }
         }
      }
   }

   public ClassAdvisor getTempClassAdvisor(CtClass clazz) throws Exception
   {
      String classname = clazz.getName();
      ClassAdvisor advisor = new ClassAdvisor(classname);
      attachMetaData(advisor, clazz, false);
      return advisor;
   }

   /**
    * Find a pointcut of with a given name
    */
   public Pointcut getPointcut(String name)
   {
      synchronized (pointcuts)
      {
         return (Pointcut) pointcuts.get(name);
      }
   }

   /**
    * Remove an interceptor pointcut with a given name
    */
   public void removePointcut(String name)
   {
      synchronized (pointcuts)
      {
         pointcuts.remove(name);
      }
   }

   /**
    * Add an interceptor pointcut with a given name
    */
   public synchronized void addPointcut(Pointcut pointcut)
   {
      removePointcut(pointcut.getName());
      synchronized (pointcuts)
      {
         pointcuts.put(pointcut.getName(), pointcut);
      }
   }

   /**
    * Remove an interceptor pointcut with a given name
    */
   public void removeBinding(String name)
   {
      synchronized (bindings)
      {
         AdviceBinding binding = (AdviceBinding) bindings.remove(name);
         if (binding == null) return;
         binding.clearAdvisors();
      }
   }

   /**
    * Add an interceptor pointcut with a given name
    */
   public synchronized void addBinding(AdviceBinding binding)
   {
      removeBinding(binding.getName());
      synchronized (bindings)
      {
         bindings.put(binding.getName(), binding);
      }
      synchronized (advisors)
      {
         Iterator it = advisors.values().iterator();
         while (it.hasNext())
         {
            Advisor advisor = (Advisor) it.next();
            if (binding.getPointcut().softMatch(advisor))
            {
               if (AspectManager.verbose)
                  System.out.println("[debug] softmatch succeeded for : " + binding);
               advisor.newBindingAdded();
            }
            else
            {
               if (AspectManager.verbose)
                  System.out.println("[debug] softmatch failed for : " + binding);
            }
         }
      }
   }

   public void removeClassMetaData(String name)
   {
      synchronized (classMetaData)
      {
         ClassMetaDataBinding meta = (ClassMetaDataBinding) classMetaData.remove(name);
         if (meta == null) return;
         meta.clearAdvisors();
      }
   }

   public void addClassMetaData(ClassMetaDataBinding meta)
   {
      removeClassMetaData(meta.getName());
      synchronized (advisors)
      {
         Iterator it = advisors.values().iterator();

         while (it.hasNext())
         {
            Advisor advisor = (Advisor) it.next();
            Class clazz = ((ClassAdvisor) advisor).getClazz();

            if (meta.matches(advisor, clazz))
            {
               meta.addAdvisor(advisor);
            }
         }
      }
      synchronized (classMetaData)
      {
         classMetaData.put(meta.getName(), meta);
      }
   }

   //--- Introductions

   /**
    * Retrieve an introduction pointcut of a certain name
    */
   public InterfaceIntroduction getInterfaceIntroduction(String name)
   {
      synchronized (interfaceIntroductions)
      {
         return (InterfaceIntroduction) interfaceIntroductions.get(name);
      }
   }

   /**
    * Register an introduction pointcut
    */
   public synchronized void addInterfaceIntroduction(InterfaceIntroduction pointcut)
   {
      removeInterfaceIntroduction(pointcut.getName());
      synchronized (interfaceIntroductions)
      {
         interfaceIntroductions.put(pointcut.getName(), pointcut);
      }
   }

   /**
    * remove an introduction pointcut of a certain name
    */
   public void removeInterfaceIntroduction(String name)
   {
      synchronized (interfaceIntroductions)
      {
         InterfaceIntroduction pointcut = (InterfaceIntroduction) interfaceIntroductions.remove(name);
         if (pointcut == null) return;
         pointcut.clearAdvisors();
      }
   }

   /**
    * Register an annotation introduction
    */
   public synchronized void addAnnotationIntroduction(AnnotationIntroduction pointcut)
   {
      String name = pointcut.getOriginalAnnotationExpr() + pointcut.getOriginalExpression();
      removeInterfaceIntroduction(name);
      synchronized (annotationIntroductions)
      {
         annotationIntroductions.put(name, pointcut);
      }
   }

   /**
    * remove an annotation pointcut
    */
   public void removeAnnotationIntroduction(AnnotationIntroduction pointcut)
   {
      String name = pointcut.getOriginalAnnotationExpr() + pointcut.getOriginalExpression();
      synchronized (annotationIntroductions)
      {
         annotationIntroductions.remove(name);
      }
   }

   public Iterator getAnnotationIntoductions()
   {
      synchronized (annotationIntroductions)
      {
         return new ArrayList(annotationIntroductions.values()).iterator();
      }
   }

   protected void applyInterfaceIntroductions(ClassAdvisor advisor, Class clazz)
   {
      Iterator it = interfaceIntroductions.values().iterator();
      while (it.hasNext())
      {
         InterfaceIntroduction pointcut = (InterfaceIntroduction) it.next();
         if (pointcut.matches(advisor, clazz))
         {
            pointcut.addAdvisor(advisor);
         }
      }
   }

   protected void applyInterfaceIntroductions(ClassAdvisor advisor, CtClass clazz) throws Exception
   {
      Iterator it = interfaceIntroductions.values().iterator();
      while (it.hasNext())
      {
         InterfaceIntroduction pointcut = (InterfaceIntroduction) it.next();
         if (pointcut.matches(advisor, clazz))
         {
            pointcut.addAdvisor(advisor);
         }
      }
   }


   public Object getPerVMAspect(AspectDefinition def)
   {
      return getPerVMAspect(def.getName());
   }

   public Object getPerVMAspect(String def)
   {
      Object aspect = perVMAspects.get(def);
      if (aspect instanceof AspectDefinition)
      {
         synchronized (aspect)
         {
            AspectDefinition adef = (AspectDefinition) aspect;
            aspect = adef.getFactory().createPerVM();
            perVMAspects.put(def, aspect);
         }
      }
      return aspect;
   }

   public void addAspectDefinition(AspectDefinition def) throws Exception
   {
      removeAspectDefinition(def.getName());
      if (def.getScope() == Scope.PER_VM)
      {
         perVMAspects.put(def.getName(), def);
      }
      aspectDefinitions.put(def.getName(), def);
   }

   public void removeAspectDefinition(String name)
   {
      AspectDefinition def = (AspectDefinition) aspectDefinitions.remove(name);
      if (def != null)
      {
         def.undeploy();
         def.clearAdvisors();
         if (def.getScope() == Scope.PER_VM) perVMAspects.remove(def.getName());
      }
   }

   public Map getAspectDefinitions()
   {
      return aspectDefinitions;
   }

   public AspectDefinition getAspectDefinition(String name)
   {
      return (AspectDefinition) aspectDefinitions.get(name);
   }

   public synchronized void addTypedef(Typedef def) throws Exception
   {
      removeTypedef(def.getName());
      synchronized (typedefs)
      {
         typedefs.put(def.getName(), def);
      }
   }

   public void removeTypedef(String name)
   {
      synchronized (typedefs)
      {
         typedefs.remove(name);
      }
   }

   public Typedef getTypedef(String name)
   {
      synchronized (typedefs)
      {
         return (Typedef) typedefs.get(name);
      }
   }

   public Map getUcl2classes()
   {
      return ucl2classes;
   }

   public Map getInterfaceIntroductions()
   {
      return interfaceIntroductions;
   }

   public Map getAnnotationIntroductions()
   {
      return annotationIntroductions;
   }

   public Map getTypedefs()
   {
      return typedefs;
   }

   public Map getInterceptorStacks()
   {
      return interceptorStacks;
   }

   public Map getClassMetaDataLoaders()
   {
      return classMetaDataLoaders;
   }

   public Map getCflowStacks()
   {
      return cflowStacks;
   }

   public Map getDynamicCFlows()
   {
      return dynamicCFlows;
   }

   public Map getPerVMAspects()
   {
      return perVMAspects;
   }

   public Map getClassMetaData()
   {
      return classMetaData;
   }

   /**
    * Get the instrumented version of the class loader.
    *
    * @return The instrumented version of the class loader.
    * @throws NotFoundException      if the class loader cannot be found.
    * @throws CannotCompileException if the class cannot be compiled.
    * @throws IOException            If the class bytecodes cannot be obtained.
    */
   public byte[] getInstrumentedClassLoader()
           throws NotFoundException, IOException, CannotCompileException
   {
      final CtClass clazz = classpool.get(ClassLoader.class.getName());
      final ExprEditor defineClassEditor = new ExprEditor()
      {
         public void edit(final MethodCall methodCall) throws CannotCompileException
         {
            final String methodName = methodCall.getMethodName();
            if ("defineClass0".equals(methodName))
            {
               methodCall.replace("{"
                       + "  final byte[] newBytes = org.jboss.aop.AspectManager.instance().translate($1, $0) ;"
                       + "  if (newBytes != (byte[])null) {"
                       + "    $_ = $proceed($1, newBytes, 0, newBytes.length, $5) ;"
                       + "  } else {"
                       + "    $_ = $proceed($$);"
                       + "  }"
                       + "}");
            }

         }
      };
      clazz.instrument(defineClassEditor);
      return clazz.toBytecode();
   }

   /**
    * Are we re-entering?
    *
    * @return true if re-entry, false otherwise.
    */
   private static boolean isReEntry()
   {
      return ((Boolean) REENTRY.get()).booleanValue();
   }

   /**
    * Set the re-entry flag.
    */
   private static void setReEntry()
   {
      REENTRY.set(Boolean.TRUE);
   }

   /**
    * Clear the re-entry flag.
    */
   private static void clearReEntry()
   {
      REENTRY.set(Boolean.FALSE);
   }
}
