/***************************************
 *                                     *
 *  JBoss: The OpenSource J2EE WebOS   *
 *                                     *
 *  Distributable under LGPL license.  *
 *  See terms of license at gnu.org.   *
 *                                     *
 ***************************************/

package org.jboss.aop;

import org.jboss.aop.advice.AdviceBinding;
import org.jboss.aop.advice.AdviceFactory;
import org.jboss.aop.advice.AdviceStack;
import org.jboss.aop.advice.AspectDefinition;
import org.jboss.aop.advice.AspectFactory;
import org.jboss.aop.advice.AspectFactoryDelegator;
import org.jboss.aop.advice.DynamicCFlowDefinition;
import org.jboss.aop.advice.GenericAspectFactory;
import org.jboss.aop.advice.InterceptorFactory;
import org.jboss.aop.advice.Scope;
import org.jboss.aop.advice.ScopeUtil;
import org.jboss.aop.advice.ScopedInterceptorFactory;
import org.jboss.aop.introduction.AnnotationIntroduction;
import org.jboss.aop.introduction.InterfaceIntroduction;
import org.jboss.aop.metadata.ClassMetaDataBinding;
import org.jboss.aop.metadata.ClassMetaDataLoader;
import org.jboss.aop.pointcut.CFlow;
import org.jboss.aop.pointcut.CFlowStack;
import org.jboss.aop.pointcut.Pointcut;
import org.jboss.aop.pointcut.PointcutExpression;
import org.jboss.aop.pointcut.Typedef;
import org.jboss.aop.pointcut.TypedefExpression;
import org.jboss.aop.pointcut.ast.ASTCFlowExpression;
import org.jboss.aop.pointcut.ast.ASTStart;
import org.jboss.aop.pointcut.ast.ParseException;
import org.jboss.aop.pointcut.ast.PointcutExpressionParser;
import org.jboss.aop.pointcut.ast.TypeExpressionParser;
import org.jboss.aop.util.XmlHelper;
import org.jboss.util.xml.XmlLoadable;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.EntityResolver;
import org.xml.sax.ErrorHandler;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringReader;
import java.net.URL;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.StringTokenizer;

/**
 * @author <a href="mailto:bill@jboss.org">Bill Burke</a>
 * @version $Revision: 1.56.2.4 $
 */
public class AspectXmlLoader
{
   // Attributes ---------------------------------------------------

   static ThreadLocal counter = new ThreadLocal();
   static ThreadLocal defaultBaseName = new ThreadLocal();

   /**
    * Generates a default name if one is not provided
    */
   private static String getName(Element element, String errorHeader)
   {
      String name = element.getAttribute("name");
      if (name != null && !name.equals(""))
      {
         return name;
      }

      return getName(errorHeader);

   }

   private static String getName(String errorHeader)
   {
      Long count = (Long) counter.get();
      if (count == null)
      {
         throw new RuntimeException(errorHeader + ": must have a name");
      }

      counter.set(new Long(count.longValue() + 1));

      String baseName = (String) defaultBaseName.get();

      baseName = baseName + count.toString();
      return baseName;
   }

   public static void undeployInterceptor(Element element) throws Exception
   {
      String name = element.getAttribute("name");
      if (name != null && name.trim().equals("")) name = null;

      String clazz = element.getAttribute("class");
      if (clazz != null && clazz.trim().equals("")) clazz = null;
      String factory = element.getAttribute("factory");
      if (factory != null && factory.trim().equals("")) factory = null;
      if (name == null) name = (clazz == null) ? factory : clazz;
      AspectManager.instance().removeAspectDefinition(name);
      String factoryName = name + ".invoke";
      AspectManager.instance().removeInterceptorFactory(factoryName);
   }

   public static InterceptorFactory deployInterceptor(Element element) throws Exception
   {
      String name = element.getAttribute("name");
      if (name != null && name.trim().equals("")) name = null;

      String clazz = element.getAttribute("class");
      if (clazz != null && clazz.trim().equals("")) clazz = null;
      String factory1 = element.getAttribute("factory");
      if (factory1 != null && factory1.trim().equals("")) factory1 = null;
      if (clazz == null && factory1 == null) throw new RuntimeException("Interceptor" + " " + name + " must have a class or factory associated with it.");
      if (clazz != null && factory1 != null) throw new RuntimeException("Interceptor" + " " + name + " cannot have both the class and factory attribute set.");
      if (name == null) name = (clazz == null) ? factory1 : clazz;

      String s = element.getAttribute("scope");
      if (s != null && s.trim().equals("")) s = null;

      Scope scope = null;
      if (s != null)
      {
         scope = ScopeUtil.parse(s);
         if (scope == null) throw new RuntimeException("Illegal scope attribute value: " + s);
      }
      AspectDefinition def = (AspectDefinition) AspectManager.instance().getAspectDefinitions().get(name);

      // if interceptor definition already exists, then just check to see if scopes are the same
      // todo deprecate <interceptor> within an advice stack
      if (def != null)
      {
         if (scope == null) scope = Scope.PER_VM;
         if (scope != def.getScope()) throw new RuntimeException("multiple definitions of <interceptor> " + name + " with different scopes is illegal");
      }
      else
      {
         AspectFactory aspectFactory;
         if (clazz != null)
         {
            aspectFactory = new GenericAspectFactory(clazz, element);
         }
         else
         {
            aspectFactory = new AspectFactoryDelegator(factory1, element);
         }
         def = new AspectDefinition(name, scope, aspectFactory);
         AspectManager.instance().addAspectDefinition(def);
      }
      ScopedInterceptorFactory factory = new ScopedInterceptorFactory(def);
      AspectManager.instance().addInterceptorFactory(factory.getName(), factory);
      return factory;
   }


   public static void undeployAdvice(Element element) throws Exception
   {
      String name = element.getAttribute("name");
      String aspect = element.getAttribute("aspect");
      String factory = aspect + "." + name;
      AspectManager.instance().removeInterceptorFactory(factory);
   }

   public static InterceptorFactory deployAdvice(Element element) throws Exception
   {
      String name = element.getAttribute("name");
      String aspect = element.getAttribute("aspect");
      AspectDefinition def = AspectManager.instance().getAspectDefinition(aspect);
      if (def == null) throw new RuntimeException("advice " + name + " cannot find aspect " + aspect);
      AdviceFactory factory = new AdviceFactory(def, name);
      AspectManager.instance().addInterceptorFactory(factory.getName(), factory);
      return factory;
   }

   public static void deployBinding(Element element) throws Exception
   {
      String name = getName(element, "binding");
      String pointcut = element.getAttribute("pointcut");
      if (pointcut == null || pointcut.trim().equals(""))
      {
         throw new RuntimeException("Binding must have a pointcut element associated with it.");
      }
      String cflow = element.getAttribute("cflow");
      if (cflow != null && cflow.trim().equals(""))
      {
         cflow = null;
      }
      ASTCFlowExpression cflowExpression = null;
      if (cflow != null)
      {
         try
         {
            cflowExpression = new PointcutExpressionParser(new StringReader(cflow)).CFlowExpression();
         }
         catch (ParseException e)
         {
            throw new RuntimeException(cflow, e);  //To change body of catch statement use Options | File Templates.
         }
      }
      ArrayList interceptors = loadInterceptors(element);
      InterceptorFactory[] inters = (InterceptorFactory[]) interceptors.toArray(new InterceptorFactory[interceptors.size()]);
      Pointcut p = null;
      try
      {
         p = new PointcutExpression(getName("binding pointcut "), pointcut);
      }
      catch (ParseException ex)
      {
         throw new RuntimeException("<bind> pointcut expression failed: " + pointcut, ex);
      }
      AspectManager.instance().addPointcut(p);

      AdviceBinding binding = new AdviceBinding(name, p, cflowExpression, cflow, inters);
      AspectManager.instance().addBinding(binding);
   }

   public static void undeployBinding(Element element) throws Exception
   {
      String binding = getName(element, "binding");
      AspectManager.instance().removeBinding(binding);
      unloadInterceptors(element);
      String pointcut = getName("pointcut");
      AspectManager.instance().removePointcut(pointcut);
   }


   private static ArrayList loadInterceptors(Element element) throws Exception
   {
      ArrayList interceptors = new ArrayList();
      NodeList children2 = element.getChildNodes();
      for (int j = 0; j < children2.getLength(); j++)
      {
         if (children2.item(j).getNodeType() == Node.ELEMENT_NODE)
         {
            Element interceptorElement = (Element) children2.item(j);
            String tag2 = interceptorElement.getTagName();
            if (tag2.equals("interceptor"))
            {
               InterceptorFactory factory = deployInterceptor(interceptorElement);
               interceptors.add(factory);
            }
            else if (tag2.equals("interceptor-ref"))
            {
               String iname = interceptorElement.getAttribute("name");
               if (iname == null) throw new RuntimeException("interceptor-ref has null name attribute");
               InterceptorFactory factory = AspectManager.instance().getInterceptorFactory(iname);
               if (factory == null) throw new RuntimeException("unable to resolve interceptor-ref: " + iname);
               interceptors.add(factory);
            }
            else if (tag2.equals("stack-ref"))
            {
               String name = interceptorElement.getAttribute("name");
               AdviceStack stack = AspectManager.instance().getAdviceStack(name);
               if (stack == null) throw new Exception("there is no <stack> defined for name: " + name);
               interceptors.addAll(stack.getInterceptorFactories());
            }
            else if (tag2.equals("advice"))
            {
               InterceptorFactory factory = deployAdvice(interceptorElement);
               interceptors.add(factory);
            }
         }
      }
      return interceptors;
   }

   private static void unloadInterceptors(Element element) throws Exception
   {
      NodeList children2 = element.getChildNodes();
      for (int j = 0; j < children2.getLength(); j++)
      {
         if (children2.item(j).getNodeType() == Node.ELEMENT_NODE)
         {
            Element interceptorElement = (Element) children2.item(j);
            String tag2 = interceptorElement.getTagName();
            if (tag2.equals("interceptor"))
            {
               undeployInterceptor(interceptorElement);
            }
            if (tag2.equals("advice"))
            {
               undeployAdvice(interceptorElement);
            }
         }
      }
   }

   public static void undeployAspect(Element pointcut) throws Exception
   {
      String clazz = pointcut.getAttribute("class");
      if (clazz == null || clazz.trim().equals("")) clazz = null;
      String name = pointcut.getAttribute("name");
      if (name == null || name.trim().equals("")) name = clazz;
      if (name == null) return;
      AspectManager.instance().removeAspectDefinition(name);
   }

   public static AspectDefinition deployAspect(Element element, String type) throws Exception
   {
      String name = element.getAttribute("name");
      if (name != null && name.trim().equals("")) name = null;

      String clazz = element.getAttribute("class");
      if (clazz != null && clazz.trim().equals("")) clazz = null;
      String factory = element.getAttribute("factory");
      if (factory != null && factory.trim().equals("")) factory = null;
      if (clazz == null && factory == null) throw new RuntimeException(type + " " + name + " must have a class or factory associated with it.");
      if (clazz != null && factory != null) throw new RuntimeException(type + " " + name + " cannot have both the class and factory attribute set.");
      if (name == null) name = (clazz == null) ? factory : clazz;

      String s = element.getAttribute("scope");
      if (s != null && s.trim().equals("")) s = null;

      Scope scope = null;
      if (s != null)
      {
         scope = ScopeUtil.parse(s);
         if (scope == null) throw new RuntimeException("Illegal scope attribute value: " + s);
      }
      AspectFactory aspectFactory;
      if (clazz != null)
      {
         aspectFactory = new GenericAspectFactory(clazz, element);
      }
      else
      {
         aspectFactory = new AspectFactoryDelegator(factory, element);
      }
      AspectDefinition def = new AspectDefinition(name, scope, aspectFactory);
      AspectManager.instance().addAspectDefinition(def);
      return def;
   }

   public static void undeployCFlowStack(Element pointcut) throws Exception
   {
      String name = pointcut.getAttribute("name");
      AspectManager.instance().removeCFlowStack(name);
   }

   public static void deployCFlowStack(Element pointcut) throws Exception
   {
      String name = pointcut.getAttribute("name");
      if (name != null && name.trim().equals("")) name = null;
      if (name == null) throw new RuntimeException("name required for a cflow-stack");
      CFlowStack stack = new CFlowStack(name);
      NodeList children2 = pointcut.getChildNodes();
      for (int j = 0; j < children2.getLength(); j++)
      {
         if (children2.item(j).getNodeType() == Node.ELEMENT_NODE)
         {
            Element cflowElement = (Element) children2.item(j);
            String tag2 = cflowElement.getTagName();
            String expr = cflowElement.getAttribute("expr");
            if (expr != null && expr.trim().equals(""))
            {
               throw new RuntimeException(tag2 + " requires an expr attribute to be defined");
            }
            if (tag2.equals("called"))
            {
               stack.addCFlow(new CFlow(expr, false));
            }
            if (tag2.equals("not-called"))
            {
               stack.addCFlow(new CFlow(expr, true));
            }
         }
      }

      AspectManager.instance().addCFlowStack(stack);
   }


   public static void undeployInterceptorStack(Element element) throws Exception
   {
      String name = element.getAttribute("name");
      unloadInterceptors(element);
      AspectManager.instance().removeInterceptorStack(name);
   }

   public static void deployInterceptorStack(Element element) throws Exception
   {
      ArrayList interceptors = loadInterceptors(element);
      String name = element.getAttribute("name");
      AdviceStack stack = new AdviceStack(name, interceptors);
      AspectManager.instance().addAdviceStack(stack);
   }

   private static ClassMetaDataBinding loadMetaData(Element element)
           throws Exception
   {
      String classExpr = element.getAttribute("class");
      String tag = element.getAttribute("tag");
      String name = getName(element, "metadata: " + tag + " " + classExpr);
      ClassMetaDataLoader loader = AspectManager.instance().findClassMetaDataLoader(tag);
      return loader.importMetaData(element, name, tag, classExpr);
   }

   public static void undeployMetaDataLoader(Element element) throws Exception
   {
      String tag = element.getAttribute("tag");
      AspectManager.instance().removeClassMetaDataLoader(tag);
   }

   public static void deployMetaDataLoader(Element element) throws Exception
   {
      String tag = element.getAttribute("tag");
      String classname = element.getAttribute("class");
      Class clazz = Thread.currentThread().getContextClassLoader().loadClass(classname);
      ClassMetaDataLoader loader = (ClassMetaDataLoader) clazz.newInstance();
      if (loader instanceof XmlLoadable)
      {
         ((XmlLoadable) loader).importXml(element);
      }
      AspectManager.instance().addClassMetaDataLoader(tag, loader);
   }

   public static void deployClassMetaData(Element element) throws Exception
   {
      ClassMetaDataBinding data = loadMetaData(element);
      AspectManager.instance().addClassMetaData(data);
   }

   public static void undeployClassMetaData(Element element) throws Exception
   {
      String classExpr = element.getAttribute("class");
      String tag = element.getAttribute("tag");
      String name = getName(element, "metadata: " + tag + " " + classExpr);
      AspectManager.instance().removeClassMetaData(name);
   }

   public static void undeployPointcut(Element pointcut) throws Exception
   {
      String name = pointcut.getAttribute("name");
      AspectManager.instance().removePointcut(name);
   }

   public static void undeployPluggablePointcut(Element pointcut) throws Exception
   {
      String name = pointcut.getAttribute("name");
      AspectManager.instance().removePointcut(name);
   }

   public static void deployPluggablePointcut(Element pointcut) throws Exception
   {
      String name = pointcut.getAttribute("name");
      if (name == null || name.trim().equals(""))
      {
         throw new RuntimeException("pluggable pointcut declaration must have a name associated with it");
      }
      String clazz = pointcut.getAttribute("class");
      if (clazz != null && clazz.trim().equals(""))
      {
         throw new RuntimeException("pluggable pointcut declaration must have an expr associated with it");
      }

      Pointcut p = null;

      Class pClass = null;
      try
      {
         pClass = Thread.currentThread().getContextClassLoader().loadClass(clazz);
      }
      catch (ClassNotFoundException e)
      {
         throw new RuntimeException("pluggable pointcut class not found: " + clazz);
      }
      p = (Pointcut) pClass.newInstance();
      if (p instanceof XmlLoadable)
      {
         ((XmlLoadable) p).importXml(pointcut);
      }
      AspectManager.instance().addPointcut(p);
   }

   public static void undeployDynamicCFlow(Element pointcut) throws Exception
   {
      String name = pointcut.getAttribute("name");
      AspectManager.instance().removeDynamicCFlow(name);
   }

   public static void deployDynamicCFlow(Element pointcut) throws Exception
   {
      String name = pointcut.getAttribute("name");
      if (name == null || name.trim().equals(""))
      {
         throw new RuntimeException("dynamic cflow declaration must have a name associated with it");
      }
      String clazz = pointcut.getAttribute("class");
      if (clazz != null && clazz.trim().equals(""))
      {
         throw new RuntimeException("dynamic cflow declaration must have an expr associated with it");
      }

      AspectManager.instance().addDynamicCFlow(name, new DynamicCFlowDefinition(pointcut, clazz, name));
   }

   public static void deployPointcut(Element pointcut) throws Exception
   {
      String name = pointcut.getAttribute("name");
      if (name == null || name.trim().equals(""))
      {
         throw new RuntimeException("pointcut declaration must have a name associated with it");
      }
      String expr = pointcut.getAttribute("expr");
      if (expr == null || expr.trim().equals(""))
      {
         throw new RuntimeException("pointcut declaration must have an expr associated with it");
      }
      Pointcut p = null;
      try
      {
         p = new PointcutExpression(name, expr);
      }
      catch (ParseException ex)
      {
         throw new RuntimeException("<pointcut name='" + name + "' expr='" + expr + "'/> failed", ex);
      }
      AspectManager.instance().addPointcut(p);
   }

   public static void undeployPrepare(Element pointcut) throws Exception
   {
      String name = getName(pointcut, "prepare");
      AspectManager.instance().removePointcut(name);
   }

   public static void deployPrepare(Element pointcut) throws Exception
   {
      String name = getName(pointcut, "prepare");
      String expr = pointcut.getAttribute("expr");
      if (expr != null && expr.trim().equals(""))
      {
         throw new RuntimeException("pointcut declaration must have an expr associated with it");
      }
      Pointcut p = null;
      try
      {
         p = new PointcutExpression(name, expr);
      }
      catch (ParseException ex)
      {
         throw new RuntimeException("<prepare> failed: " + expr, ex);
      }
      AspectManager.instance().addPointcut(p);
   }

   public static void deployAnnotationIntroduction(Element pointcut) throws Exception
   {
      AspectManager.instance().addAnnotationIntroduction(loadAnnotationIntroduction(pointcut));
   }

   public static void undeployAnnotationIntroduction(Element pointcut) throws Exception
   {
      AspectManager.instance().removeAnnotationIntroduction(loadAnnotationIntroduction(pointcut));
   }

   public static AnnotationIntroduction loadAnnotationIntroduction(Element pointcut) throws Exception
   {
      String expr = pointcut.getAttribute("expr");
      if (expr != null && expr.trim().equals(""))
      {
         throw new RuntimeException("annotation introduction must have an expr attribute");
      }

      String invisible = pointcut.getAttribute("invisible");
      if (invisible != null && expr.trim().equals(""))
      {
         throw new RuntimeException("annotation introduction must have an invisible attribute");
      }

      boolean isInvisble = new Boolean(invisible).booleanValue();

      String annotation = XmlHelper.getElementContent(pointcut);

      return AnnotationIntroduction.createComplexAnnotationIntroduction(expr, annotation, isInvisble);
   }

   public static void undeployIntroductionPointcut(Element pointcut) throws Exception
   {
      String name = getName(pointcut, "introduction-pointcut");

      AspectManager.instance().removeInterfaceIntroduction(name);
   }

   public static void deployIntroductionPointcut(Element pointcut) throws Exception
   {
      String name = getName(pointcut, "introduction");
      String classExpr = pointcut.getAttribute("class");
      if (classExpr != null && classExpr.trim().equals(""))
      {
         classExpr = null;
      }

      String ast = pointcut.getAttribute("expr");
      if (ast != null && ast.trim().equals(""))
      {
         ast = null;
      }


      if (classExpr == null && ast == null)
      {
         throw new RuntimeException("A class nor a expr attribute is defined for this <introduction>");
      }

      if (classExpr != null && ast != null)
      {
         throw new RuntimeException("You cannot define both a class and expr attribute in the same <introduction>");
      }

      String intfs = XmlHelper.getOptionalChildContent(pointcut, "interfaces");
      String[] ifaces = null;
      if (intfs != null)
      {
         StringTokenizer tokenizer = new StringTokenizer(intfs, ",");
         ArrayList interfaces = new ArrayList();
         while (tokenizer.hasMoreTokens())
         {
            String intf = tokenizer.nextToken().trim();
            if (!intf.equals("")) interfaces.add(intf);
         }
         ifaces = (String[]) interfaces.toArray(new String[interfaces.size()]);
      }

      InterfaceIntroduction pcut = null;
      if (classExpr != null)
      {
         pcut = new InterfaceIntroduction(name, classExpr, ifaces);
      }
      else
      {
         ASTStart start = new TypeExpressionParser(new StringReader(ast)).Start();
         pcut = new InterfaceIntroduction(name, start, ifaces);
      }
      Iterator it = XmlHelper.getChildrenByTagName(pointcut, "mixin");
      while (it.hasNext())
      {
         Element mixin = (Element) it.next();
         if (mixin != null)
         {
            String construction = XmlHelper.getOptionalChildContent(mixin, "construction");
            String classname = XmlHelper.getUniqueChildContent(mixin, "class");
            String isTransientString = mixin.getAttribute("transient");
            boolean isTransient = true;
            if (isTransientString == null || isTransientString.trim().equals(""))
            {
               isTransient = true;
            }
            else
            {
               isTransient = new Boolean(isTransientString).booleanValue();
            }

            intfs = XmlHelper.getUniqueChildContent(mixin, "interfaces");
            StringTokenizer tokenizer = new StringTokenizer(intfs, ",");
            ArrayList interfaces = new ArrayList();
            while (tokenizer.hasMoreTokens())
            {
               String intf = tokenizer.nextToken().trim();
               if (!intf.equals("")) interfaces.add(intf);
            }
            ifaces = (String[]) interfaces.toArray(new String[interfaces.size()]);
            pcut.getMixins().add(new InterfaceIntroduction.Mixin(classname, ifaces, construction, isTransient));
         }
      }
      AspectManager.instance().addInterfaceIntroduction(pcut);
   }

   public static void deployTypedef(Element pointcut) throws Exception
   {
      String name = pointcut.getAttribute("name");
      if (name == null || name.trim().equals(""))
      {
         throw new RuntimeException("typedef declaration must have a name associated with it");
      }
      String expr = pointcut.getAttribute("expr");
      if (expr == null || expr.trim().equals(""))
      {
         throw new RuntimeException("typedef declaration must have an expr associated with it");
      }

      try
      {
         Typedef typedef = new TypedefExpression(name, expr);
         AspectManager.instance().addTypedef(typedef);
      }
      catch (ParseException ex)
      {
         throw new RuntimeException("<typedef name='" + name + "' expr='" + expr + "'/> failed", ex);
      }
   }

   public static void undeployTypedef(Element pointcut) throws Exception
   {
      String name = pointcut.getAttribute("name");
      AspectManager.instance().removeTypedef(name);
   }


   private static void setupDefaultName(URL url) throws Exception
   {
      if (url == null) return;
      counter.set(new Long(0));
      defaultBaseName.set(url.toString());
   }


   public static void deployXML(Document doc, URL url) throws Exception
   {
      setupDefaultName(url);
      NodeList children = doc.getDocumentElement().getChildNodes();
      try
      {
         for (int i = 0; i < children.getLength(); i++)
         {
            if (children.item(i).getNodeType() == Node.ELEMENT_NODE)
            {
               Element element = (Element) children.item(i);
               String tag = element.getTagName();
               if (AspectManager.verbose)
               {
                  System.out.println("[deploy] <" + tag + "> " + url.toString());
               }
               if (tag.equals("interceptor"))
               {
                  deployInterceptor(element);
               }
               else if (tag.equals("introduction"))
               {
                  deployIntroductionPointcut(element);
               }
               else if (tag.equals("metadata-loader"))
               {
                  deployMetaDataLoader(element);
               }
               else if (tag.equals("metadata"))
               {
                  deployClassMetaData(element);
               }
               else if (tag.equals("stack"))
               {
                  deployInterceptorStack(element);
               }
               else if (tag.equals("aspect"))
               {
                  deployAspect(element, "Aspect");
               }
               else if (tag.equals("pointcut"))
               {
                  deployPointcut(element);
               }
               else if (tag.equals("pluggable-pointcut"))
               {
                  deployPluggablePointcut(element);
               }
               else if (tag.equals("bind"))
               {
                  deployBinding(element);
               }
               else if (tag.equals("prepare"))
               {
                  deployPrepare(element);
               }
               else if (tag.equals("cflow-stack"))
               {
                  deployCFlowStack(element);
               }
               else if (tag.equals("dynamic-cflow"))
               {
                  deployDynamicCFlow(element);
               }
               else if (tag.equals("annotation-introduction"))
               {
                  deployAnnotationIntroduction(element);
               }
               else if (tag.equals("typedef"))
               {
                  deployTypedef(element);
               }
               else
               {
                  throw new IllegalArgumentException("Unknown AOP tag: " + tag);
               }
            }
         }
      }
      catch (Exception e)
      {
         System.err.println("[error] " + e.getMessage() + " " + url);
         if (AspectManager.verbose) e.printStackTrace();
         throw new RuntimeException(e);
      }
   }

   public static void undeployXML(Document doc, URL url) throws Exception
   {
      setupDefaultName(url);
      NodeList children = doc.getDocumentElement().getChildNodes();
      for (int i = 0; i < children.getLength(); i++)
      {
         if (children.item(i).getNodeType() == Node.ELEMENT_NODE)
         {
            Element element = (Element) children.item(i);
            String tag = element.getTagName();
            if (AspectManager.verbose)
            {
               System.out.println("[undeploy] <" + tag + "> " + url.toString());
            }
            if (tag.equals("interceptor"))
            {
               undeployInterceptor(element);
            }
            else if (tag.equals("introduction"))
            {
               undeployIntroductionPointcut(element);
            }
            else if (tag.equals("metadata-loader"))
            {
               undeployMetaDataLoader(element);
            }
            else if (tag.equals("metadata"))
            {
               undeployClassMetaData(element);
            }
            else if (tag.equals("stack"))
            {
               undeployInterceptorStack(element);
            }
            else if (tag.equals("aspect"))
            {
               undeployAspect(element);
            }
            else if (tag.equals("pointcut"))
            {
               undeployPointcut(element);
            }
            else if (tag.equals("bind"))
            {
               undeployBinding(element);
            }
            else if (tag.equals("prepare"))
            {
               undeployPrepare(element);
            }
            else if (tag.equals("cflow-stack"))
            {
               undeployCFlowStack(element);
            }
            else if (tag.equals("pluggable-pointcut"))
            {
               undeployPluggablePointcut(element);
            }
            else if (tag.equals("dynamic-cflow"))
            {
               undeployDynamicCFlow(element);
            }
            else if (tag.equals("typedef"))
            {
               undeployTypedef(element);
            }
            else if (tag.equals("annotation-introduction"))
            {
               undeployAnnotationIntroduction(element);
            }
         }
      }
   }


   public static void deployXML(URL url) throws Exception
   {
      deployXML(loadURL(url), url);
   }

   public static void undeployXML(URL url) throws Exception
   {
      undeployXML(loadURL(url), url);
   }

   private static class Resolver implements EntityResolver
   {
      public InputSource resolveEntity(String publicId, String systemId) throws SAXException, IOException
      {
         if (systemId.endsWith("jboss-aop_1_0.dtd"))
         {
            try
            {
               URL url = AspectXmlLoader.class.getResource("/jboss-aop_1_0.dtd");
               InputStream is = url.openStream();
               return new InputSource(is);
            }
            catch (IOException e)
            {
               e.printStackTrace();
               return null;
            }
         }
         return null;
      }
   }

   private static class LocalErrorHandler implements ErrorHandler
   {
      public void warning(SAXParseException exception) throws SAXException
      {

      }

      public void error(SAXParseException exception) throws SAXException
      {

      }

      public void fatalError(SAXParseException exception) throws SAXException
      {

      }
   }

   private static Document loadURL(URL configURL) throws Exception
   {
      InputStream is = configURL != null ? configURL.openStream() : null;
      if (is == null)
         throw new IOException("Failed to obtain InputStream from url: " + configURL);

      DocumentBuilderFactory docBuilderFactory = null;
      docBuilderFactory = DocumentBuilderFactory.newInstance();
      docBuilderFactory.setValidating(false);
      InputSource source = new InputSource(is);
      URL url = AspectXmlLoader.class.getResource("/jboss-aop_1_0.dtd");
      source.setSystemId(url.toString());
      DocumentBuilder docBuilder = docBuilderFactory.newDocumentBuilder();
      docBuilder.setEntityResolver(new Resolver());
      //docBuilder.setErrorHandler(new LocalErrorHandler());
      Document doc = docBuilder.parse(source);
      return doc;
   }

}
