/***************************************
 *                                     *
 *  JBoss: The OpenSource J2EE WebOS   *
 *                                     *
 *  Distributable under LGPL license.  *
 *  See terms of license at gnu.org.   *
 *                                     *
 ***************************************/

package org.jboss.aop;

import gnu.trove.TLongObjectHashMap;
import org.jboss.aop.advice.AdviceBinding;
import org.jboss.aop.advice.AspectDefinition;
import org.jboss.aop.advice.CFlowInterceptor;
import org.jboss.aop.advice.Interceptor;
import org.jboss.aop.advice.InterceptorFactory;
import org.jboss.aop.annotation.AnnotationElement;
import org.jboss.aop.instrument.ConstructorExecutionTransformer;
import org.jboss.aop.instrument.MethodExecutionTransformer;
import org.jboss.aop.introduction.InterfaceIntroduction;
import org.jboss.aop.joinpoint.ConstructorCalledByConstructorInvocation;
import org.jboss.aop.joinpoint.ConstructorCalledByConstructorJoinpoint;
import org.jboss.aop.joinpoint.ConstructorCalledByMethodInvocation;
import org.jboss.aop.joinpoint.ConstructorCalledByMethodJoinpoint;
import org.jboss.aop.joinpoint.ConstructorInvocation;
import org.jboss.aop.joinpoint.ConstructorJoinpoint;
import org.jboss.aop.joinpoint.FieldJoinpoint;
import org.jboss.aop.joinpoint.FieldReadInvocation;
import org.jboss.aop.joinpoint.FieldWriteInvocation;
import org.jboss.aop.joinpoint.Invocation;
import org.jboss.aop.joinpoint.InvocationResponse;
import org.jboss.aop.joinpoint.Joinpoint;
import org.jboss.aop.joinpoint.MethodCalledByConstructorInvocation;
import org.jboss.aop.joinpoint.MethodCalledByConstructorJoinpoint;
import org.jboss.aop.joinpoint.MethodCalledByMethodInvocation;
import org.jboss.aop.joinpoint.MethodCalledByMethodJoinpoint;
import org.jboss.aop.joinpoint.MethodInvocation;
import org.jboss.aop.joinpoint.MethodJoinpoint;
import org.jboss.aop.metadata.ClassMetaDataBinding;
import org.jboss.aop.metadata.ClassMetaDataLoader;
import org.jboss.aop.util.ConstructorComparator;
import org.jboss.aop.util.FieldComparator;
import org.jboss.aop.util.MethodHashing;

import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

/**
 * Advises a class and provides access to the class's aspect chain.
 * Each advisable class has an associated <code>Advisor</code> instance.
 * References methods using <code>int</code> IDs rather than the actual
 * instances for
 * optimal performance. Provides ability to invoke methods on an advised
 * object without advice (see
 * <code>Advisor.invokeWithoutAdvisement()</code>).
 *
 * @author <a href="mailto:bill@jboss.org">Bill Burke</a>
 * @version $Revision: 1.92.2.1 $
 */
public class ClassAdvisor extends Advisor
{

   /**
    * Suffix added to unadvised methods.
    */
   public static final String NOT_TRANSFORMABLE_SUFFIX = "$aop";

   protected Class clazz = null;

   // The method signatures are sorted at transformation and load time to
   // make sure the tables line up.
   private TLongObjectHashMap methodInterceptors = new TLongObjectHashMap();
   private TLongObjectHashMap advisedMethods = new TLongObjectHashMap();
   private TLongObjectHashMap unadvisedMethods = new TLongObjectHashMap();


   // caller pointcut support for methods calling methods only
   protected TLongObjectHashMap methodCalledByMethodBindings = new TLongObjectHashMap();
   protected HashMap backrefMethodCalledByMethodBindings = new HashMap();
   protected TLongObjectHashMap methodCalledByMethodInterceptors = new TLongObjectHashMap();

   // constructor caller pointcut support for methods calling methods only
   protected TLongObjectHashMap conCalledByMethodBindings = new TLongObjectHashMap();
   protected HashMap backrefConCalledByMethodBindings = new HashMap();
   protected TLongObjectHashMap conCalledByMethodInterceptors = new TLongObjectHashMap();

   // caller pointcut support for constructors calling methods
   protected HashMap[] methodCalledByConBindings;
   protected HashMap[] methodCalledByConInterceptors;
   protected HashMap backrefMethodCalledByConstructorBindings = new HashMap();

   // caller pointcut support for constructors calling methods
   protected HashMap[] conCalledByConBindings;
   protected HashMap[] conCalledByConInterceptors;
   protected HashMap backrefConCalledByConstructorBindings = new HashMap();

   // Used by instrumentor to access separate interceptor chains for read and write access
   private Interceptor[][] fieldReadInterceptors;
   private Interceptor[][] fieldWriteInterceptors;
   private Field[] advisedFields;
   private HashMap staticFieldAspects = new HashMap();

   private Interceptor[][] constructorInterceptors;
   private Constructor[] constructors;

   protected boolean initialized = false;


   /**
    * This method is to support PER_JOINPOINT scoping of Aspects for static fields
    * Fields are special in that a get and set do not create separate aspect instances.
    *
    * @param joinpoint
    * @param def
    * @return
    */
   public Object getFieldAspect(FieldJoinpoint joinpoint, AspectDefinition def)
   {
      Object aspect = staticFieldAspects.get(joinpoint);
      if (aspect == null)
      {
         aspect = def.getFactory().createPerJoinpoint(this, joinpoint);
         staticFieldAspects.put(joinpoint, aspect);
      }
      return aspect;
   }

   public Object resolveAnnotation(Class annotation)
   {
      Object value = super.resolveAnnotation(annotation);
      if (clazz == null) return null;
      if (value == null) value = AnnotationElement.getAnyAnnotation(clazz, annotation);
      return value;
   }

   public boolean hasAnnotation(String annotation)
   {
      if (super.hasAnnotation(annotation)) return true;
      if (clazz == null) return false;
      try
      {
         return AnnotationElement.isAnyAnnotationPresent(clazz, annotation);
      }
      catch (Exception e)
      {
         throw new RuntimeException(e);  //To change body of catch statement use Options | File Templates.
      }
   }

   public Class getClazz()
   {
      return clazz;
   }

   public Field[] getAdvisedFields()
   {
      return advisedFields;
   }

   public TLongObjectHashMap getAdvisedMethods()
   {
      return advisedMethods;
   }

   public TLongObjectHashMap getUnadvisedMethods()
   {
      return unadvisedMethods;
   }

   public Class getAdvisedClass()
   {
      return clazz;
   }

   public ClassAdvisor(String classname)
   {
      super(classname);
   }

   public Constructor[] getConstructors()
   {
      return constructors;
   }

   public TLongObjectHashMap getMethodCalledByMethodInterceptors()
   {
      return methodCalledByMethodInterceptors;
   }

   public HashMap[] getMethodCalledByConInterceptors()
   {
      return methodCalledByConInterceptors;
   }

   public HashMap[] getConCalledByConInterceptors()
   {
      return conCalledByConInterceptors;
   }

   public TLongObjectHashMap getConCalledByMethodInterceptors()
   {
      return conCalledByMethodInterceptors;
   }

   public TLongObjectHashMap getMethodCalledByMethodBindings()
   {
      return methodCalledByMethodBindings;
   }

   public Interceptor[][] getConstructorInterceptors()
   {
      return constructorInterceptors;
   }

   public Interceptor[][] getFieldReadInterceptors()
   {
      return fieldReadInterceptors;
   }

   public Interceptor[][] getFieldWriteInterceptors()
   {
      return fieldWriteInterceptors;
   }

   public TLongObjectHashMap getMethodInterceptors()
   {
      return methodInterceptors;
   }

   /**
    * Constructs a new helper.
    */
   public synchronized void attachClass(Class clazz)
   {
      try
      {
         if (initialized) return;
         AspectManager.instance().attachMetaData(this, clazz);
         interfaceIntroductions.clear();
         // metadata should always come before creation of interceptor chain
         // so that the interceptor factories have access to metadata.
         // and so that metadata joinpoints can be checked
         //
         // Also metadata needs to be applied before applyIntroductionPointcuts because
         // an annotation may be triggered by XML metadata as well as
         // after populateMixinMethods so that proper metadata is applied to added methods
         rebindClassMetaData();
         AspectManager.instance().applyInterfaceIntroductions(this, clazz);
         this.clazz = clazz;
         createFieldTable();
         createMethodTables();
         createConstructorTables();
         populateMixinMethods();
         // metadata should always come before creation of interceptor chain
         // so that the interceptor factories have access to metadata.
         // and so that metadata joinpoints can be checked
         //
         // Also metadata needs to be applied before applyIntroductionPointcuts because
         // an annotation may be triggered by XML metadata as well as
         // after populateMixinMethods so that proper metadata is applied to added methods
         rebindClassMetaData();
         createInterceptorChains();
         initialized = true;
      }
      catch (Exception e)
      {
         throw new RuntimeException(e);
      }
   }

   /**
    * Get method from clazz .If method not found,get the method
    * from the clazz's parent.
    */
   static private Method getMethod(Class clazz, Method method) throws NoSuchMethodException
   {

      if ((clazz == null) || (clazz.equals(Object.class))) throw new NoSuchMethodException(method.getName());
      try
      {
         String wrappedName = ClassAdvisor.notAdvisedMethodName(clazz.getName(), method.getName());
         return clazz.getMethod(wrappedName, method.getParameterTypes());
      }
      catch (NoSuchMethodException e)
      {
         return getMethod(clazz.getSuperclass(), method);
      }
   }
   
   /** 
    * Get a constructor's index in the class. Returns -1 if not there 
    */
   public int getConstructorIndex(Constructor constructor)
   {
      for (int i = 0; i < constructors.length; i++)
      {
         if (constructor.equals(constructors[i]))
         {
            return i;
         }
      }
      
      return -1;
   }

   /** 
    * Get a field's index in the class. Returns -1 if not there 
    */
   public int getFieldIndex(Field field)
   {
      for (int i = 0; i < advisedFields.length; i++)
      {
         if (field.equals(advisedFields[i]))
         {
            return i;
         }
      }
      
      return -1;
   }
   
   
   /**
    * Put mixin methods from mixin class into unadvisedMethods map so that
    * they can be correctly invoked upon.
    */
   protected void populateMixinMethods() throws Exception
   {
      ArrayList pointcuts = getInterfaceIntroductions();
      Iterator it = pointcuts.iterator();
      while (it.hasNext())
      {
         InterfaceIntroduction pointcut = (InterfaceIntroduction) it.next();
         ArrayList mixins = pointcut.getMixins();
         for (int i = 0; i < mixins.size(); i++)
         {
            InterfaceIntroduction.Mixin mixin = (InterfaceIntroduction.Mixin) mixins.get(i);
            Class mixinClass = Thread.currentThread().getContextClassLoader().loadClass(mixin.getClassName());
            String[] interfaces = mixin.getInterfaces();
            for (int j = 0; j < interfaces.length; j++)
            {
               Class intf = Thread.currentThread().getContextClassLoader().loadClass(interfaces[j]);
               Method[] methods = intf.getMethods();
               for (int k = 0; k < methods.length; k++)
               {
                  /*
Method method = mixinClass.getMethod(methods[k].getName(), methods[k].getParameterTypes());
long hash = MethodHashing.methodHash(method);
unadvisedMethods.put(hash, method);
System.out.println(method + ":" + hash);
*/
                  //Put wrapped method in the class itself into the unadvisedMethods map
                  //   String wrappedName = ClassAdvisor.notAdvisedMethodName(clazz.getName(), methods[k].getName());
                  //   Method method = clazz.getMethod(wrappedName, methods[k].getParameterTypes());
                  Method method = getMethod(clazz, methods[k]);
                  long hash = MethodHashing.methodHash(method);
                  unadvisedMethods.put(hash, method);
               }
            }
         }
      }
   }


   public synchronized void removeAdviceBinding(AdviceBinding binding)
   {
      removeCallerPointcut(binding); // if binding is a caller remove references to it
      super.removeAdviceBinding(binding);
   }

   protected void createInterceptorChain(InterceptorFactory[] factories, ArrayList newinterceptors, Joinpoint joinpoint)
   {
      for (int i = 0; i < factories.length; i++)
      {
         if (factories[i].isDeployed()) newinterceptors.add(factories[i].create(this, joinpoint));
      }
   }

   protected void resolveMethodPointcut(TLongObjectHashMap newMethodInterceptors, AdviceBinding binding)
   {
      long[] keys = advisedMethods.keys();
      for (int i = 0; i < keys.length; i++)
      {
         Method method = (Method) advisedMethods.get(keys[i]);
         if (binding.getPointcut().matchesExecution(this, method))
         {
            adviceBindings.add(binding);
            if (AspectManager.verbose)
            {
               /*
               RepositoryClassLoader loader = (RepositoryClassLoader)clazz.getClassLoader();
               try
               {
                  System.err.println("method matched binding " + binding.getPointcut().getExpr() + " " + method.toString() + " " + loader.getObjectName());
               }
               catch (MalformedObjectNameException e)
               {
                  throw new RuntimeException(e);
               }
               */
               System.err.println("method matched binding " + binding.getPointcut().getExpr() + " " + method.toString());

            }
            binding.addAdvisor(this);
            MethodInfo info = (MethodInfo) newMethodInterceptors.get(keys[i]);
            ArrayList curr = info.interceptorChain;
            if (binding.getCFlow() != null)
            {
               ArrayList cflowChain = new ArrayList();
               createInterceptorChain(binding.getInterceptorFactories(), cflowChain, new MethodJoinpoint(method));
               Interceptor[] cflowInterceptors = (Interceptor[]) cflowChain.toArray(new Interceptor[cflowChain.size()]);
               curr.add(new CFlowInterceptor(binding.getCFlowString(), binding.getCFlow(), cflowInterceptors));
            }
            else
            {
               createInterceptorChain(binding.getInterceptorFactories(), curr, new MethodJoinpoint(method));
            }
         }
      }
   }

   protected void resolveFieldPointcut(ArrayList newFieldInterceptors, AdviceBinding binding, boolean write)
   {
      for (int i = 0; i < advisedFields.length; i++)
      {
         Field field = advisedFields[i];
         if ((!write && binding.getPointcut().matchesGet(this, field))
                 || (write && binding.getPointcut().matchesSet(this, field)))
         {
            if (AspectManager.verbose) System.err.println("field matched binding " + binding.getName());
            adviceBindings.add(binding);
            binding.addAdvisor(this);
            ArrayList curr = (ArrayList) newFieldInterceptors.get(i);
            if (binding.getCFlow() != null)
            {
               ArrayList cflowChain = new ArrayList();
               createInterceptorChain(binding.getInterceptorFactories(), cflowChain, new FieldJoinpoint(field));
               Interceptor[] cflowInterceptors = (Interceptor[]) cflowChain.toArray(new Interceptor[cflowChain.size()]);
               curr.add(new CFlowInterceptor(binding.getCFlowString(), binding.getCFlow(), cflowInterceptors));
            }
            else
            {
               createInterceptorChain(binding.getInterceptorFactories(), curr, new FieldJoinpoint(field));
            }
         }
      }
   }

   protected void resolveConstructorPointcut(ArrayList newConstructorInterceptors, AdviceBinding binding)
   {
      for (int i = 0; i < constructors.length; i++)
      {
         Constructor constructor = constructors[i];
         if (binding.getPointcut().matchesExecution(this, constructor))
         {
            if (AspectManager.verbose) System.err.println(constructor + " matched binding " + binding.getName() + " " + binding.getPointcut().getExpr());
            adviceBindings.add(binding);
            binding.addAdvisor(this);
            ArrayList curr = (ArrayList) newConstructorInterceptors.get(i);
            if (binding.getCFlow() != null)
            {
               ArrayList cflowChain = new ArrayList();
               createInterceptorChain(binding.getInterceptorFactories(), cflowChain, new ConstructorJoinpoint(constructor));
               Interceptor[] cflowInterceptors = (Interceptor[]) cflowChain.toArray(new Interceptor[cflowChain.size()]);
               curr.add(new CFlowInterceptor(binding.getCFlowString(), binding.getCFlow(), cflowInterceptors));
            }
            else
            {
               createInterceptorChain(binding.getInterceptorFactories(), curr, new ConstructorJoinpoint(constructor));
            }
         }
      }

   }

   protected TLongObjectHashMap initializeMethodChain()
   {
      TLongObjectHashMap newInterceptors = new TLongObjectHashMap();
      long[] keys = advisedMethods.keys();
      for (int i = 0; i < keys.length; i++)
      {
         MethodInfo info = new MethodInfo();
         Method amethod = (Method) advisedMethods.get(keys[i]);
         info.advisedMethod = amethod;
         Method umethod = (Method) unadvisedMethods.get(keys[i]);

         info.unadvisedMethod = umethod;
         info.hash = keys[i];
         info.advisor = this;
         newInterceptors.put(keys[i], info);
         try
         {
            Field infoField = clazz.getDeclaredField(MethodExecutionTransformer.getMethodInfoFieldName(amethod.getName(), keys[i]));
            infoField.setAccessible(true);
            infoField.set(null, info);
         }
         catch (NoSuchFieldException e)
         {
            // ignore, method may not be advised.
         }
         catch (IllegalAccessException e)
         {
            throw new RuntimeException(e);  //To change body of catch statement use Options | File Templates.
         }
      }
      return newInterceptors;
   }

   protected ArrayList initializeFieldChain()
   {
      ArrayList chain = new ArrayList(advisedFields.length);
      for (int i = 0; i < advisedFields.length; i++)
      {
         ArrayList list = new ArrayList();
         chain.add(list);
      }
      return chain;
   }

   protected ArrayList initializeConstructorChain()
   {
      ArrayList chain = new ArrayList(constructors.length);
      for (int i = 0; i < constructors.length; i++)
      {
         ArrayList list = new ArrayList();
         chain.add(list);
      }
      return chain;
   }

   protected void finalizeMethodChain(TLongObjectHashMap newMethodInterceptors)
   {
      long[] keys = newMethodInterceptors.keys();
      for (int i = 0; i < keys.length; i++)
      {
         MethodInfo info = (MethodInfo) newMethodInterceptors.get(keys[i]);
         ArrayList list = info.interceptorChain;
         info.interceptors = (Interceptor[]) list.toArray(new Interceptor[list.size()]);
      }
   }

   protected void finalizeFieldChain(ArrayList newFieldInterceptors)
   {
      for (int i = 0; i < newFieldInterceptors.size(); i++)
      {
         ArrayList list = (ArrayList) newFieldInterceptors.get(i);
         Interceptor[] chain = (Interceptor[]) list.toArray(new Interceptor[list.size()]);
         newFieldInterceptors.set(i, chain);
      }
   }

   protected void finalizeConstructorChain(ArrayList newFieldInterceptors)
   {
      for (int i = 0; i < newFieldInterceptors.size(); i++)
      {
         ArrayList list = (ArrayList) newFieldInterceptors.get(i);
         Interceptor[] chain = (Interceptor[]) list.toArray(new Interceptor[list.size()]);
         newFieldInterceptors.set(i, chain);
      }
   }

   protected void createInterceptorChains() throws Exception
   {
      TLongObjectHashMap newMethodInterceptors = initializeMethodChain();
      ArrayList newFieldReadInterceptors = initializeFieldChain();
      ArrayList newFieldWriteInterceptors = initializeFieldChain();
      ArrayList newConstructorInterceptors = initializeConstructorChain();
      Iterator it = AspectManager.instance().bindings.values().iterator();
      while (it.hasNext())
      {
         AdviceBinding binding = (AdviceBinding) it.next();
         if (AspectManager.verbose) System.out.println("iterate binding " + binding.getName());
         resolveMethodPointcut(newMethodInterceptors, binding);
         resolveFieldPointcut(newFieldReadInterceptors, binding, false);
         resolveFieldPointcut(newFieldWriteInterceptors, binding, true);
         resolveConstructorPointcut(newConstructorInterceptors, binding);
      }
      finalizeMethodChain(newMethodInterceptors);
      finalizeFieldChain(newFieldReadInterceptors);
      finalizeFieldChain(newFieldWriteInterceptors);
      finalizeConstructorChain(newConstructorInterceptors);
      methodInterceptors = newMethodInterceptors;
      fieldReadInterceptors = (Interceptor[][]) newFieldReadInterceptors.toArray(new Interceptor[0][0]);
      fieldWriteInterceptors = (Interceptor[][]) newFieldWriteInterceptors.toArray(new Interceptor[0][0]);
      constructorInterceptors = (Interceptor[][]) newConstructorInterceptors.toArray(new Interceptor[0][0]);
      doesHaveAspects = adviceBindings.size() > 0;
   }


   protected CallerMethodInfo initializeCallerInterceptorsMap(long callingMethodHash, String calledClass, long calledMethodHash, Method calledMethod) throws Exception
   {
      HashMap calledClassesMap = (HashMap) methodCalledByMethodInterceptors.get(callingMethodHash);
      if (calledClassesMap == null)
      {
         calledClassesMap = new HashMap();
         methodCalledByMethodInterceptors.put(callingMethodHash, calledClassesMap);
      }
      TLongObjectHashMap calledMethodsMap = (TLongObjectHashMap) calledClassesMap.get(calledClass);
      if (calledMethodsMap == null)
      {
         calledMethodsMap = new TLongObjectHashMap();
         calledClassesMap.put(calledClass, calledMethodsMap);
      }
      //The standard MethodCalledByXXXXInvocation class calls by reflection and needs access
      calledMethod.setAccessible(true);
      CallerMethodInfo info = new CallerMethodInfo(calledMethod, null);
      calledMethodsMap.put(calledMethodHash, info);
      return info;
   }

   protected CallerConstructorInfo initializeConCalledByMethodInterceptorsMap(long callingMethodHash, String calledClass, long calledConHash, Constructor calledCon) throws Exception
   {
      HashMap calledClassesMap = (HashMap) conCalledByMethodInterceptors.get(callingMethodHash);
      if (calledClassesMap == null)
      {
         calledClassesMap = new HashMap();
         conCalledByMethodInterceptors.put(callingMethodHash, calledClassesMap);
      }
      TLongObjectHashMap calledMethodsMap = (TLongObjectHashMap) calledClassesMap.get(calledClass);
      if (calledMethodsMap == null)
      {
         calledMethodsMap = new TLongObjectHashMap();
         calledClassesMap.put(calledClass, calledMethodsMap);
      }
      CallerConstructorInfo info = null;
      info = createCallerConstructorInfo(calledClass, calledCon);
      calledMethodsMap.put(calledConHash, info);
      return info;
   }

   private CallerConstructorInfo createCallerConstructorInfo(String calledClass, Constructor calledCon)
   {
      //The standard ConstructorCalledByXXXXInvocation class calls by reflection and needs access
      calledCon.setAccessible(true);

      CallerConstructorInfo info;
      try
      {
         int index = calledClass.lastIndexOf('.');
         String baseClassName = calledClass.substring(index + 1);
         Method wrapper = calledCon.getDeclaringClass().getDeclaredMethod(ConstructorExecutionTransformer.constructorFactory(baseClassName), calledCon.getParameterTypes());
         info = new CallerConstructorInfo(calledCon, wrapper, null);
      }
      catch (NoSuchMethodException e)
      {
         info = new CallerConstructorInfo(calledCon, null);
      }
      return info;
   }

   protected CallerMethodInfo initializeConstructorCallerInterceptorsMap(int callingIndex, String calledClass, long calledMethodHash, Method calledMethod) throws Exception
   {
      HashMap calledClassesMap = methodCalledByConInterceptors[callingIndex];
      if (calledClassesMap == null)
      {
         calledClassesMap = new HashMap();
         methodCalledByConInterceptors[callingIndex] = calledClassesMap;
      }
      TLongObjectHashMap calledMethodsMap = (TLongObjectHashMap) calledClassesMap.get(calledClass);
      if (calledMethodsMap == null)
      {
         calledMethodsMap = new TLongObjectHashMap();
         calledClassesMap.put(calledClass, calledMethodsMap);
      }

      //The standard MethodCalledByXXXXInvocation class calls by reflection and needs access
      calledMethod.setAccessible(true);
      CallerMethodInfo info = new CallerMethodInfo(calledMethod, null);
      calledMethodsMap.put(calledMethodHash, info);
      return info;
   }

   protected CallerConstructorInfo initializeConCalledByConInterceptorsMap(int callingIndex, String calledClass, long calledConHash, Constructor calledCon) throws Exception
   {
      HashMap calledClassesMap = conCalledByConInterceptors[callingIndex];
      if (calledClassesMap == null)
      {
         calledClassesMap = new HashMap();
         conCalledByConInterceptors[callingIndex] = calledClassesMap;
      }
      TLongObjectHashMap calledMethodsMap = (TLongObjectHashMap) calledClassesMap.get(calledClass);
      if (calledMethodsMap == null)
      {
         calledMethodsMap = new TLongObjectHashMap();
         calledClassesMap.put(calledClass, calledMethodsMap);
      }
      CallerConstructorInfo info = createCallerConstructorInfo(calledClass, calledCon);
      calledMethodsMap.put(calledConHash, info);
      return info;
   }

   protected void rebuildCallerInterceptors() throws Exception
   {
      long[] callingKeys = methodCalledByMethodInterceptors.keys();
      for (int i = 0; i < callingKeys.length; i++)
      {
         long callingHash = callingKeys[i];
         HashMap calledClasses = (HashMap) methodCalledByMethodInterceptors.get(callingHash);
         Iterator classesIterator = calledClasses.entrySet().iterator();
         while (classesIterator.hasNext())
         {
            Map.Entry entry = (Map.Entry) classesIterator.next();
            String cname = (String) entry.getKey();
            TLongObjectHashMap calledMethods = (TLongObjectHashMap) entry.getValue();
            long[] calledKeys = calledMethods.keys();
            for (int j = 0; j < calledKeys.length; j++)
            {
               long calledHash = calledKeys[j];
               ArrayList bindings = getCallerBindings(callingHash, cname, calledHash);
               Method calling = MethodHashing.findMethodByHash(clazz, callingHash);
               bindCallerInterceptorChain(bindings, callingHash, cname, calledHash, calling);
            }
         }
      }
      for (int i = 0; i < methodCalledByConInterceptors.length; i++)
      {
         HashMap calledClasses = methodCalledByConInterceptors[i];
         if (calledClasses == null) continue;
         Iterator classesIterator = calledClasses.entrySet().iterator();
         while (classesIterator.hasNext())
         {
            Map.Entry entry = (Map.Entry) classesIterator.next();
            String cname = (String) entry.getKey();
            TLongObjectHashMap calledMethods = (TLongObjectHashMap) entry.getValue();
            long[] calledKeys = calledMethods.keys();
            for (int j = 0; j < calledKeys.length; j++)
            {
               long calledHash = calledKeys[j];
               ArrayList bindings = getConstructorCallerBindings(i, cname, calledHash);
               bindConstructorCallerInterceptorChain(bindings, i, cname, calledHash);
            }
         }
      }
      callingKeys = conCalledByMethodInterceptors.keys();
      for (int i = 0; i < callingKeys.length; i++)
      {
         long callingHash = callingKeys[i];
         HashMap calledClasses = (HashMap) conCalledByMethodInterceptors.get(callingHash);
         Iterator classesIterator = calledClasses.entrySet().iterator();
         while (classesIterator.hasNext())
         {
            Map.Entry entry = (Map.Entry) classesIterator.next();
            String cname = (String) entry.getKey();
            TLongObjectHashMap calledMethods = (TLongObjectHashMap) entry.getValue();
            long[] calledKeys = calledMethods.keys();
            for (int j = 0; j < calledKeys.length; j++)
            {
               long calledHash = calledKeys[j];
               ArrayList bindings = getConCalledByMethodBindings(callingHash, cname, calledHash);
               bindConCalledByMethodInterceptorChain(bindings, callingHash, cname, calledHash);
            }
         }
      }
      for (int i = 0; i < conCalledByConInterceptors.length; i++)
      {
         HashMap calledClasses = conCalledByConInterceptors[i];
         if (calledClasses == null) continue;
         Iterator classesIterator = calledClasses.entrySet().iterator();
         while (classesIterator.hasNext())
         {
            Map.Entry entry = (Map.Entry) classesIterator.next();
            String cname = (String) entry.getKey();
            TLongObjectHashMap calledMethods = (TLongObjectHashMap) entry.getValue();
            long[] calledKeys = calledMethods.keys();
            for (int j = 0; j < calledKeys.length; j++)
            {
               long calledHash = calledKeys[j];
               ArrayList bindings = getConCalledByConBindings(i, cname, calledHash);
               bindConCalledByConInterceptorChain(bindings, i, cname, calledHash);
            }
         }
      }
   }

   private ArrayList getCallerBindings(long callingHash, String cname, long calledHash)
   {
      HashMap calledClasses = (HashMap) methodCalledByMethodBindings.get(callingHash);
      TLongObjectHashMap calledMethods = (TLongObjectHashMap) calledClasses.get(cname);
      return (ArrayList) calledMethods.get(calledHash);
   }

   private ArrayList getConCalledByMethodBindings(long callingHash, String cname, long calledHash)
   {
      HashMap calledClasses = (HashMap) conCalledByMethodBindings.get(callingHash);
      TLongObjectHashMap calledCons = (TLongObjectHashMap) calledClasses.get(cname);
      return (ArrayList) calledCons.get(calledHash);
   }

   private ArrayList getConstructorCallerBindings(int callingIndex, String cname, long calledHash)
   {
      HashMap calledClasses = methodCalledByConBindings[callingIndex];
      TLongObjectHashMap calledMethods = (TLongObjectHashMap) calledClasses.get(cname);
      return (ArrayList) calledMethods.get(calledHash);
   }

   private ArrayList getConCalledByConBindings(int callingIndex, String cname, long calledHash)
   {
      HashMap calledClasses = conCalledByConBindings[callingIndex];
      TLongObjectHashMap calledMethods = (TLongObjectHashMap) calledClasses.get(cname);
      return (ArrayList) calledMethods.get(calledHash);
   }

   private void bindCallerInterceptorChain(ArrayList bindings, long callingHash, String cname, long calledHash, Method calling)
   {
      ArrayList interceptors = new ArrayList();
      Iterator it = bindings.iterator();
      CallerMethodInfo info = getCallerMethodInfo(callingHash, cname, calledHash);
      while (it.hasNext())
      {
         AdviceBinding binding = (AdviceBinding) it.next();
         if (binding.getCFlow() != null)
         {
            ArrayList cflowChain = new ArrayList();
            createInterceptorChain(binding.getInterceptorFactories(), cflowChain, new MethodCalledByMethodJoinpoint(calling, info.method));
            Interceptor[] cflowInterceptors = (Interceptor[]) cflowChain.toArray(new Interceptor[cflowChain.size()]);
            interceptors.add(new CFlowInterceptor(binding.getCFlowString(), binding.getCFlow(), cflowInterceptors));
         }
         else
         {
            createInterceptorChain(binding.getInterceptorFactories(), interceptors, new MethodCalledByMethodJoinpoint(calling, info.method));
         }
      }
      Interceptor[] chain = (Interceptor[]) interceptors.toArray(new Interceptor[interceptors.size()]);
      info.interceptors = (chain.length == 0) ? null : chain;
   }

   private void bindConCalledByMethodInterceptorChain(ArrayList bindings, long callingHash, String cname, long calledHash) throws Exception
   {
      ArrayList interceptors = new ArrayList();
      Iterator it = bindings.iterator();
      CallerConstructorInfo info = getConCalledByMethod(callingHash, cname, calledHash);
      Method calling = MethodHashing.findMethodByHash(clazz, callingHash);
      while (it.hasNext())
      {
         AdviceBinding binding = (AdviceBinding) it.next();
         if (binding.getCFlow() != null)
         {
            ArrayList cflowChain = new ArrayList();
            createInterceptorChain(binding.getInterceptorFactories(), cflowChain, new ConstructorCalledByMethodJoinpoint(calling, info.constructor));
            Interceptor[] cflowInterceptors = (Interceptor[]) cflowChain.toArray(new Interceptor[cflowChain.size()]);
            interceptors.add(new CFlowInterceptor(binding.getCFlowString(), binding.getCFlow(), cflowInterceptors));
         }
         else
         {
            createInterceptorChain(binding.getInterceptorFactories(), interceptors, new ConstructorCalledByMethodJoinpoint(calling, info.constructor));
         }
      }
      Interceptor[] chain = (Interceptor[]) interceptors.toArray(new Interceptor[interceptors.size()]);
      info.interceptors = (chain.length == 0) ? null : chain;
   }

   private void bindConCalledByConInterceptorChain(ArrayList bindings, int callingIndex, String cname, long calledHash)
   {
      ArrayList interceptors = new ArrayList();
      Iterator it = bindings.iterator();
      CallerConstructorInfo info = getConCalledByCon(callingIndex, cname, calledHash);
      while (it.hasNext())
      {
         AdviceBinding binding = (AdviceBinding) it.next();
         if (binding.getCFlow() != null)
         {
            ArrayList cflowChain = new ArrayList();
            createInterceptorChain(binding.getInterceptorFactories(), cflowChain, new ConstructorCalledByConstructorJoinpoint(constructors[callingIndex], info.constructor));
            Interceptor[] cflowInterceptors = (Interceptor[]) cflowChain.toArray(new Interceptor[cflowChain.size()]);
            interceptors.add(new CFlowInterceptor(binding.getCFlowString(), binding.getCFlow(), cflowInterceptors));
         }
         else
         {
            createInterceptorChain(binding.getInterceptorFactories(), interceptors, new ConstructorCalledByConstructorJoinpoint(constructors[callingIndex], info.constructor));
         }
      }
      Interceptor[] chain = (Interceptor[]) interceptors.toArray(new Interceptor[interceptors.size()]);
      info.interceptors = (chain.length == 0) ? null : chain;
   }

   private void bindConstructorCallerInterceptorChain(ArrayList bindings, int callingIndex, String cname, long calledHash)
   {
      ArrayList interceptors = new ArrayList();
      Iterator it = bindings.iterator();
      CallerMethodInfo info = getConstructorCallerMethodInfo(callingIndex, cname, calledHash);
      while (it.hasNext())
      {
         AdviceBinding binding = (AdviceBinding) it.next();
         if (binding.getCFlow() != null)
         {
            ArrayList cflowChain = new ArrayList();
            createInterceptorChain(binding.getInterceptorFactories(), cflowChain, new MethodCalledByConstructorJoinpoint(constructors[callingIndex], info.method));
            Interceptor[] cflowInterceptors = (Interceptor[]) cflowChain.toArray(new Interceptor[cflowChain.size()]);
            interceptors.add(new CFlowInterceptor(binding.getCFlowString(), binding.getCFlow(), cflowInterceptors));
         }
         else
         {
            createInterceptorChain(binding.getInterceptorFactories(), interceptors, new MethodCalledByConstructorJoinpoint(constructors[callingIndex], info.method));
         }
      }
      Interceptor[] chain = (Interceptor[]) interceptors.toArray(new Interceptor[interceptors.size()]);
      info.interceptors = (chain.length == 0) ? null : chain;
   }

   protected void rebuildInterceptors()
   {
      if (initialized)
      {
         try
         {
            adviceBindings.clear();
            createInterceptorChains();
            rebuildCallerInterceptors();
         }
         catch (Exception ex)
         {
            throw new RuntimeException(ex);
         }
      }
   }

   protected void bindClassMetaData(ClassMetaDataBinding data)
   {
      try
      {
         ClassMetaDataLoader loader = data.getLoader();
         Object[] objs = advisedMethods.getValues();
         Method[] methods = new Method[objs.length];
         Field[] fields = advisedFields;
         // set to empty array because advisedFields may not have been initialized yet
         if (fields == null) fields = new Field[0];
         Constructor[] cons = constructors;
         // set to empty array because constructors may not have been initialized yet
         if (cons == null) cons = new Constructor[0];
         for (int i = 0; i < objs.length; i++) methods[i] = (Method) objs[i];
         loader.bind(this, data, methods, fields, cons);
      }
      catch (Exception ex)
      {
         // REVISIT:  Need to know how errors affects deployment
         ex.printStackTrace();
      }
   }

   protected void rebindClassMetaData()
   {
      defaultMetaData.clear();
      methodMetaData.clear();
      fieldMetaData.clear();
      constructorMetaData.clear();
      classMetaData.clear();

      for (int i = 0; i < classMetaDataBindings.size(); i++)
      {
         ClassMetaDataBinding data = (ClassMetaDataBinding) classMetaDataBindings.get(i);
         bindClassMetaData(data);
      }
   }


   public synchronized void addClassMetaData(ClassMetaDataBinding data)
   {
      classMetaDataBindings.add(data);
      if (this.clazz == null) return;  // don't bind till later.

      bindClassMetaData(data);
      // Recalculate interceptorPointcuts because of MetaDataInterceptorPointcuts
      adviceBindings.clear();
      doesHaveAspects = false;
      rebuildInterceptors();
   }

   public synchronized void removeClassMetaData(ClassMetaDataBinding data)
   {
      if (classMetaDataBindings.remove(data))
      {
         if (this.clazz == null) return; // not bound yet
         rebindClassMetaData();
         // Recalculate interceptorPointcuts because of MetaDataInterceptorPointcuts
         adviceBindings.clear();
         doesHaveAspects = false;
         rebuildInterceptors();
      }
   }

   public void initializeEmptyCallerChain(long callingMethodHash, String calledClass, long calledMethodHash) throws Exception
   {
      HashMap callingMethod = (HashMap) methodCalledByMethodBindings.get(callingMethodHash);
      if (callingMethod == null)
      {
         callingMethod = new HashMap();
         methodCalledByMethodBindings.put(callingMethodHash, callingMethod);
      }
      TLongObjectHashMap classMap = (TLongObjectHashMap) callingMethod.get(calledClass);
      if (classMap == null)
      {
         classMap = new TLongObjectHashMap();
         callingMethod.put(calledClass, classMap);
      }
      ArrayList bindings = (ArrayList) classMap.get(calledMethodHash);
      if (bindings == null)
      {
         bindings = new ArrayList();
         classMap.put(calledMethodHash, bindings);
      }
   }

   public void initializeConCalledByMethodEmptyChain(long callingMethodHash, String calledClass, long calledConHash) throws Exception
   {
      HashMap callingMethod = (HashMap) conCalledByMethodBindings.get(callingMethodHash);
      if (callingMethod == null)
      {
         callingMethod = new HashMap();
         conCalledByMethodBindings.put(callingMethodHash, callingMethod);
      }
      TLongObjectHashMap classMap = (TLongObjectHashMap) callingMethod.get(calledClass);
      if (classMap == null)
      {
         classMap = new TLongObjectHashMap();
         callingMethod.put(calledClass, classMap);
      }
      ArrayList bindings = (ArrayList) classMap.get(calledConHash);
      if (bindings == null)
      {
         bindings = new ArrayList();
         classMap.put(calledConHash, bindings);
      }
   }

   public void initializeEmptyConstructorCallerChain(int callingIndex, String calledClass, long calledMethodHash) throws Exception
   {
      HashMap callingCon = methodCalledByConBindings[callingIndex];
      if (callingCon == null)
      {
         callingCon = new HashMap();
         methodCalledByConBindings[callingIndex] = callingCon;
      }
      TLongObjectHashMap classMap = (TLongObjectHashMap) callingCon.get(calledClass);
      if (classMap == null)
      {
         classMap = new TLongObjectHashMap();
         callingCon.put(calledClass, classMap);
      }
      ArrayList bindings = (ArrayList) classMap.get(calledMethodHash);
      if (bindings == null)
      {
         bindings = new ArrayList();
         classMap.put(calledMethodHash, bindings);
      }
   }

   public void initializeConCalledByConEmptyChain(int callingIndex, String calledClass, long calledConHash) throws Exception
   {
      HashMap callingCon = conCalledByConBindings[callingIndex];
      if (callingCon == null)
      {
         callingCon = new HashMap();
         conCalledByConBindings[callingIndex] = callingCon;
      }
      TLongObjectHashMap classMap = (TLongObjectHashMap) callingCon.get(calledClass);
      if (classMap == null)
      {
         classMap = new TLongObjectHashMap();
         callingCon.put(calledClass, classMap);
      }
      ArrayList bindings = (ArrayList) classMap.get(calledConHash);
      if (bindings == null)
      {
         bindings = new ArrayList();
         classMap.put(calledConHash, bindings);
      }
   }

   public void addMethodCalledByMethodPointcut(long callingMethodHash, String calledClass, long calledMethodHash, AdviceBinding binding) throws Exception
   {
      if (AspectManager.verbose) System.err.println("method call matched binding " + binding.getPointcut().getExpr());
      adviceBindings.add(binding);
      binding.addAdvisor(this);
      HashMap callingMethod = (HashMap) methodCalledByMethodBindings.get(callingMethodHash);
      if (callingMethod == null)
      {
         callingMethod = new HashMap();
         methodCalledByMethodBindings.put(callingMethodHash, callingMethod);
      }
      TLongObjectHashMap classMap = (TLongObjectHashMap) callingMethod.get(calledClass);
      if (classMap == null)
      {
         classMap = new TLongObjectHashMap();
         callingMethod.put(calledClass, classMap);
      }
      ArrayList bindings = (ArrayList) classMap.get(calledMethodHash);
      boolean createdBindings = false;
      if (bindings == null)
      {
         bindings = new ArrayList();
         classMap.put(calledMethodHash, bindings);
         createdBindings = true;
      }
      if (!bindings.contains(binding)) bindings.add(binding);

      // this is so that we can undeploy a caller
      ArrayList backrefs = (ArrayList) backrefMethodCalledByMethodBindings.get(binding.getName());
      if (backrefs == null)
      {
         backrefs = new ArrayList();
         backrefMethodCalledByMethodBindings.put(binding.getName(), backrefs);
         backrefs.add(bindings);
      }
      else if (createdBindings) backrefs.add(bindings);
   }

   public void addConstructorCalledByMethodPointcut(long callingMethodHash, String calledClass, long calledMethodHash, AdviceBinding binding) throws Exception
   {
      if (AspectManager.verbose) System.err.println("method call matched binding " + binding.getPointcut().getExpr());
      adviceBindings.add(binding);
      binding.addAdvisor(this);
      HashMap callingMethod = (HashMap) conCalledByMethodBindings.get(callingMethodHash);
      if (callingMethod == null)
      {
         callingMethod = new HashMap();
         conCalledByMethodBindings.put(callingMethodHash, callingMethod);
      }
      TLongObjectHashMap classMap = (TLongObjectHashMap) callingMethod.get(calledClass);
      if (classMap == null)
      {
         classMap = new TLongObjectHashMap();
         callingMethod.put(calledClass, classMap);
      }
      ArrayList bindings = (ArrayList) classMap.get(calledMethodHash);
      boolean createdBindings = false;
      if (bindings == null)
      {
         bindings = new ArrayList();
         classMap.put(calledMethodHash, bindings);
         createdBindings = true;
      }
      if (!bindings.contains(binding)) bindings.add(binding);

      // this is so that we can undeploy a caller
      ArrayList backrefs = (ArrayList) backrefConCalledByMethodBindings.get(binding.getName());
      if (backrefs == null)
      {
         backrefs = new ArrayList();
         backrefConCalledByMethodBindings.put(binding.getName(), backrefs);
         backrefs.add(bindings);
      }
      else if (createdBindings) backrefs.add(bindings);
   }

   public void addConstructorCallerPointcut(int callingIndex, String calledClass, long calledMethodHash, AdviceBinding binding) throws Exception
   {
      if (AspectManager.verbose) System.err.println("constructor call matched binding " + binding.getPointcut().getExpr());
      adviceBindings.add(binding);
      binding.addAdvisor(this);
      HashMap callingCon = methodCalledByConBindings[callingIndex];
      if (callingCon == null)
      {
         callingCon = new HashMap();
         methodCalledByConBindings[callingIndex] = callingCon;
      }
      TLongObjectHashMap classMap = (TLongObjectHashMap) callingCon.get(calledClass);
      if (classMap == null)
      {
         classMap = new TLongObjectHashMap();
         callingCon.put(calledClass, classMap);
      }
      ArrayList bindings = (ArrayList) classMap.get(calledMethodHash);
      boolean createdBindings = false;
      if (bindings == null)
      {
         bindings = new ArrayList();
         classMap.put(calledMethodHash, bindings);
         createdBindings = true;
      }
      if (!bindings.contains(binding)) bindings.add(binding);

      // this is so that we can undeploy a caller
      ArrayList backrefs = (ArrayList) backrefMethodCalledByConstructorBindings.get(binding.getName());
      if (backrefs == null)
      {
         backrefs = new ArrayList();
         backrefMethodCalledByConstructorBindings.put(binding.getName(), backrefs);
         backrefs.add(bindings);
      }
      else if (createdBindings) backrefs.add(bindings);
   }

   public void addConstructorCalledByConPointcut(int callingIndex, String calledClass, long calledConHash, AdviceBinding binding) throws Exception
   {
      if (AspectManager.verbose) System.err.println("constructor call matched binding " + binding.getPointcut().getExpr());
      adviceBindings.add(binding);
      binding.addAdvisor(this);
      HashMap callingCon = conCalledByConBindings[callingIndex];
      if (callingCon == null)
      {
         callingCon = new HashMap();
         conCalledByConBindings[callingIndex] = callingCon;
      }
      TLongObjectHashMap classMap = (TLongObjectHashMap) callingCon.get(calledClass);
      if (classMap == null)
      {
         classMap = new TLongObjectHashMap();
         callingCon.put(calledClass, classMap);
      }
      ArrayList bindings = (ArrayList) classMap.get(calledConHash);
      boolean createdBindings = false;
      if (bindings == null)
      {
         bindings = new ArrayList();
         classMap.put(calledConHash, bindings);
         createdBindings = true;
      }
      if (!bindings.contains(binding)) bindings.add(binding);

      // this is so that we can undeploy a caller
      ArrayList backrefs = (ArrayList) backrefConCalledByConstructorBindings.get(binding.getName());
      if (backrefs == null)
      {
         backrefs = new ArrayList();
         backrefConCalledByConstructorBindings.put(binding.getName(), backrefs);
         backrefs.add(bindings);
      }
      else if (createdBindings) backrefs.add(bindings);
   }

   public void removeCallerPointcut(AdviceBinding binding)
   {
      ArrayList backrefs = (ArrayList) backrefMethodCalledByMethodBindings.get(binding.getName());
      if (backrefs == null) return;
      for (int i = 0; i < backrefs.size(); i++)
      {
         ArrayList ref = (ArrayList) backrefs.get(i);
         ref.remove(binding);
      }
   }

   /**
    * Generates internal, unadvised version of a method name.
    */
   public static String notAdvisedMethodName(String className,
                                             String methodName)
   {
      return className.replace('.', '$') + "$" + methodName +
              NOT_TRANSFORMABLE_SUFFIX;
   }

   /**
    * Is this the name of a private, unadvised thing?
    */
   public static boolean isWithoutAdvisement(String name)
   {
      return name.endsWith(NOT_TRANSFORMABLE_SUFFIX);
   }

   /**
    * Is the field advisable?
    */
   public static boolean isAdvisable(Field field)
   {
      // note: this should match the implementation in the instrumentor.
      int modifiers = field.getModifiers();
      return
              !Modifier.isFinal(modifiers) &&
              !isWithoutAdvisement(field.getName()) &&
              !field.getName().startsWith("_") &&
              (field.getName().indexOf('$') == -1);
   }


   /**
    * Is the method advisable?
    */
   public static boolean isAdvisable(Method method)
   {
      // note: this should match the implementation in the instrumentor.
      int modifiers = method.getModifiers();
      return
              !isWithoutAdvisement(method.getName()) &&
              !Modifier.isAbstract(modifiers) &&
              !Modifier.isNative(modifiers) &&
              !method.getName().startsWith("_");
   }


   private void populateFieldTable(ArrayList fields, Class superclass)
           throws Exception
   {
      if (superclass == null) return;
      if (superclass.equals(Object.class)) return;

      populateFieldTable(fields, superclass.getSuperclass());

      // if (!isAdvised(superclass)) return;

      ArrayList temp = new ArrayList();
      Field[] declaredFields = superclass.getDeclaredFields();
      for (int i = 0; i < declaredFields.length; i++)
      {
         if (isAdvisable(declaredFields[i]))
         {
            // Need to do this because notadvisable fields maybe private or protected
            declaredFields[i].setAccessible(true);
            temp.add(declaredFields[i]);
         }
      }
      Collections.sort(temp, FieldComparator.INSTANCE);
      fields.addAll(temp);
   }

   /**
    * Gets advised methods.
    */
   private void createFieldTable() throws Exception
   {
      ArrayList fields = new ArrayList();

      populateFieldTable(fields, clazz);

      advisedFields = (Field[]) fields.toArray(new Field[fields.size()]);

   }

   private void addDeclaredMethods(Class superclass) throws Exception
   {
      Method[] declaredMethods = superclass.getDeclaredMethods();
      for (int i = 0; i < declaredMethods.length; i++)
      {
         if (isAdvisable(declaredMethods[i]))
         {
            long hash = MethodHashing.methodHash(declaredMethods[i]);
            advisedMethods.put(hash, declaredMethods[i]);
            try
            {
               Method m = declaredMethods[i];
               Method un = superclass.getDeclaredMethod(notAdvisedMethodName(superclass.getName(),
                       m.getName()),
                       m.getParameterTypes());
               un.setAccessible(true);
               unadvisedMethods.put(hash, un);
            }
            catch (NoSuchMethodException ignored)
            {
            }
         }
      }
   }


   /**
    * Create a HashMap of method hash and Method
    * Superclasses get added first so subclasses will override with
    * correct overriden method
    */
   private void populateMethodTables(Class superclass)
           throws Exception
   {
      if (superclass == null) return;
      if (superclass.equals(Object.class)) return;

      populateMethodTables(superclass.getSuperclass());

      ClassAdvisor superAdvisor = AspectManager.instance().getAdvisorIfAdvised(superclass);
      if (superAdvisor != null)
      {
         TLongObjectHashMap superHash = superAdvisor.getUnadvisedMethods();
         long[] keys = superHash.keys();
         for (int i = 0; i < keys.length; i++)
         {
            unadvisedMethods.put(keys[i], superHash.get(keys[i]));
         }
      }
      addDeclaredMethods(superclass);
   }

   private void createMethodTables()
           throws Exception
   {
      populateMethodTables(clazz.getSuperclass());
      addDeclaredMethods(clazz);
   }

   private void createConstructorTables() throws Exception
   {
      constructors = clazz.getDeclaredConstructors();
      methodCalledByConBindings = new HashMap[constructors.length];
      methodCalledByConInterceptors = new HashMap[constructors.length];

      conCalledByConBindings = new HashMap[constructors.length];
      conCalledByConInterceptors = new HashMap[constructors.length];
      for (int i = 0; i < constructors.length; i++)
      {
         constructors[i].setAccessible(true);
      }
      Arrays.sort(constructors, ConstructorComparator.INSTANCE);
   }

   public CallerMethodInfo resolveCallerMethodInfo(long callingMethodHash, String calledClass, long calledMethodHash)
   {
      try
      {
         Method callingMethod = MethodHashing.findMethodByHash(clazz, callingMethodHash);
         if (callingMethod == null) throw new RuntimeException("Unable to figure out calling method of a caller pointcut");
         Class called = Thread.currentThread().getContextClassLoader().loadClass(calledClass);
         Method calledMethod = MethodHashing.findMethodByHash(called, calledMethodHash);
         if (calledMethod == null) throw new RuntimeException("Unable to figure out calledmethod of a caller pointcut");

         Iterator it = AspectManager.instance().bindings.values().iterator();
         boolean matched = false;
         while (it.hasNext())
         {
            AdviceBinding binding = (AdviceBinding) it.next();
            if (binding.getPointcut().matchesCall(this, callingMethod, called, calledMethod))
            {
               addMethodCalledByMethodPointcut(callingMethodHash, calledClass, calledMethodHash, binding);
               matched = true;
            }
         }
         if (!matched) initializeEmptyCallerChain(callingMethodHash, calledClass, calledMethodHash);
         CallerMethodInfo info = initializeCallerInterceptorsMap(callingMethodHash, calledClass, calledMethodHash, calledMethod);
         ArrayList bindings = getCallerBindings(callingMethodHash, calledClass, calledMethodHash);
         bindCallerInterceptorChain(bindings, callingMethodHash, calledClass, calledMethodHash, callingMethod);
         return info;
      }
      catch (Exception x)
      {
         throw new RuntimeException(x);
      }
   }

   public CallerConstructorInfo resolveCallerConstructorInfo(long callingMethodHash, String calledClass, long calledConHash)
   {
      try
      {
         Method callingMethod = MethodHashing.findMethodByHash(clazz, callingMethodHash);
         if (callingMethod == null) throw new RuntimeException("Unable to figure out calling method of a constructor caller pointcut");
         Class called = Thread.currentThread().getContextClassLoader().loadClass(calledClass);
         Constructor calledCon = MethodHashing.findConstructorByHash(called, calledConHash);
         if (calledCon == null) throw new RuntimeException("Unable to figure out calledcon of a constructor caller pointcut");

         Iterator it = AspectManager.instance().bindings.values().iterator();
         boolean matched = false;
         while (it.hasNext())
         {
            AdviceBinding binding = (AdviceBinding) it.next();
            if (binding.getPointcut().matchesCall(this, callingMethod, called, calledCon))
            {
               addConstructorCalledByMethodPointcut(callingMethodHash, calledClass, calledConHash, binding);
               matched = true;
            }
         }
         if (!matched) initializeConCalledByMethodEmptyChain(callingMethodHash, calledClass, calledConHash);
         CallerConstructorInfo info = initializeConCalledByMethodInterceptorsMap(callingMethodHash, calledClass, calledConHash, calledCon);
         ArrayList bindings = getConCalledByMethodBindings(callingMethodHash, calledClass, calledConHash);
         bindConCalledByMethodInterceptorChain(bindings, callingMethodHash, calledClass, calledConHash);
         return info;
      }
      catch (Exception x)
      {
         throw new RuntimeException(x);
      }
   }

   public CallerMethodInfo resolveConstructorCallerMethodInfo(int callingIndex, String calledClass, long calledMethodHash)
   {
      try
      {
         Constructor callingConstructor = constructors[callingIndex];
         if (callingConstructor == null) throw new RuntimeException("Unable to figure out calling method of a caller pointcut");
         Class called = Thread.currentThread().getContextClassLoader().loadClass(calledClass);
         Method calledMethod = MethodHashing.findMethodByHash(called, calledMethodHash);
         if (calledMethod == null) throw new RuntimeException("Unable to figure out calledmethod of a caller pointcut");

         Iterator it = AspectManager.instance().bindings.values().iterator();
         boolean matched = false;
         while (it.hasNext())
         {
            AdviceBinding binding = (AdviceBinding) it.next();
            if (binding.getPointcut().matchesCall(this, callingConstructor, called, calledMethod))
            {
               addConstructorCallerPointcut(callingIndex, calledClass, calledMethodHash, binding);
               matched = true;
            }
         }
         if (!matched) initializeEmptyConstructorCallerChain(callingIndex, calledClass, calledMethodHash);
         CallerMethodInfo info = initializeConstructorCallerInterceptorsMap(callingIndex, calledClass, calledMethodHash, calledMethod);
         ArrayList bindings = getConstructorCallerBindings(callingIndex, calledClass, calledMethodHash);
         bindConstructorCallerInterceptorChain(bindings, callingIndex, calledClass, calledMethodHash);
         return info;
      }
      catch (Exception x)
      {
         throw new RuntimeException(x);
      }
   }

   public CallerConstructorInfo resolveConstructorCallerConstructorInfo(int callingIndex, String calledClass, long calledConHash)
   {
      try
      {
         Constructor callingConstructor = constructors[callingIndex];
         if (callingConstructor == null) throw new RuntimeException("Unable to figure out calling method of a caller pointcut");
         Class called = Thread.currentThread().getContextClassLoader().loadClass(calledClass);
         Constructor calledCon = MethodHashing.findConstructorByHash(called, calledConHash);
         if (calledCon == null) throw new RuntimeException("Unable to figure out calledcon of a caller pointcut");

         Iterator it = AspectManager.instance().bindings.values().iterator();
         boolean matched = false;
         while (it.hasNext())
         {
            AdviceBinding binding = (AdviceBinding) it.next();
            if (binding.getPointcut().matchesCall(this, callingConstructor, called, calledCon))
            {
               addConstructorCalledByConPointcut(callingIndex, calledClass, calledConHash, binding);
               matched = true;
            }
         }
         if (!matched) initializeConCalledByConEmptyChain(callingIndex, calledClass, calledConHash);
         CallerConstructorInfo info = initializeConCalledByConInterceptorsMap(callingIndex, calledClass, calledConHash, calledCon);
         ArrayList bindings = getConCalledByConBindings(callingIndex, calledClass, calledConHash);
         bindConCalledByConInterceptorChain(bindings, callingIndex, calledClass, calledConHash);
         return info;
      }
      catch (Exception x)
      {
         throw new RuntimeException(x);
      }
   }
   /////////////////////////
   // Invoking

   /**
    * Invokes target object without applying interceptors.
    */
   public Object invokeWithoutAdvisement(Object target, long methodHash,
                                         Object[] arguments) throws Throwable
   {
      try
      {
         Method method = (Method) unadvisedMethods.get(methodHash);
         return method.invoke(target, arguments);
      }
      catch (InvocationTargetException e)
      {
         throw e.getTargetException();
      }
   }

   public Object invokeNewWithoutAdvisement(Object[] arguments, Constructor constructor) throws Throwable
   {
      try
      {
         return constructor.newInstance(arguments);
      }
      catch (InstantiationException in)
      {
         throw new RuntimeException("failed to call constructor", in);
      }
      catch (IllegalAccessException ill)
      {
         throw new RuntimeException("illegal access", ill);
      }
      catch (InvocationTargetException ite)
      {
         throw ite.getCause();
      }
   }


   public Object invokeMethod(long methodHash, Object[] arguments) throws Throwable
   {
      return invokeMethod(null, methodHash, arguments);
   }

   public Object invokeMethod(Object target, long methodHash, Object[] arguments) throws Throwable
   {
      InstanceAdvisor advisor = null;
      if (target != null)
      {
         InstanceAdvised advised = (InstanceAdvised) target;
         advisor = advised._getInstanceAdvisor();
      }
      MethodInfo info = (MethodInfo) methodInterceptors.get(methodHash);
      return invokeMethod((ClassInstanceAdvisor) advisor, target, methodHash, arguments, info);
   }


   public Object invokeMethod(InstanceAdvisor instanceAdvisor, Object target, long methodHash, Object[] arguments)
           throws Throwable
   {
      MethodInfo info = (MethodInfo) methodInterceptors.get(methodHash);
      if (info == null)
      {
         System.out.println("info is null for hash: " + methodHash + " of " + clazz.getName());
      }
      return invokeMethod((ClassInstanceAdvisor) instanceAdvisor, target, methodHash, arguments, info);
   }

   public Object invokeMethod(ClassInstanceAdvisor instanceAdvisor, Object target, long methodHash, Object[] arguments, MethodInfo info)
           throws Throwable
   {
      Interceptor[] aspects = info.interceptors;
      if (instanceAdvisor != null && (instanceAdvisor.appendedInterceptors != null || instanceAdvisor.insertedInterceptors != null))
      {
         aspects = instanceAdvisor.getInterceptors(aspects);
      }
      MethodInvocation invocation = new MethodInvocation(info, aspects);

      invocation.setArguments(arguments);
      invocation.setTargetObject(target);
      return invocation.invokeNext();
   }

   public Object invokeCaller(long callingMethodHash, Object target, Object[] args, CallerMethodInfo info)
           throws Throwable
   {
      MethodCalledByMethodInvocation invocation = new MethodCalledByMethodInvocation(info.interceptors, info.method, args, clazz, callingMethodHash);
      invocation.setTargetObject(target);
      return invocation.invokeNext();
   }

   public Object invokeConCalledByMethod(long callingMethodHash, Object[] args, CallerConstructorInfo info)
           throws Throwable
   {
      ConstructorCalledByMethodInvocation invocation = new ConstructorCalledByMethodInvocation(info.interceptors, info, callingMethodHash, clazz, args);
      return invocation.invokeNext();
   }

   public Object invokeConstructorCaller(int callingIndex, Object target, Object[] args, CallerMethodInfo info)
           throws Throwable
   {
      MethodCalledByConstructorInvocation invocation = new MethodCalledByConstructorInvocation(info.method, args, constructors[callingIndex], info.interceptors);
      invocation.setTargetObject(target);
      return invocation.invokeNext();
   }

   public Object invokeConCalledByCon(int callingIndex, Object[] args, CallerConstructorInfo info)
           throws Throwable
   {
      ConstructorCalledByConstructorInvocation invocation = new ConstructorCalledByConstructorInvocation(constructors[callingIndex], info, args);
      return invocation.invokeNext();
   }

   private CallerMethodInfo getCallerMethodInfo(long callingMethodHash, String calledClass, long calledMethodHash)
   {
      HashMap calledClasses = (HashMap) methodCalledByMethodInterceptors.get(callingMethodHash);
      TLongObjectHashMap calledMethods = (TLongObjectHashMap) calledClasses.get(calledClass);
      CallerMethodInfo info = (CallerMethodInfo) calledMethods.get(calledMethodHash);
      return info;
   }

   private CallerConstructorInfo getConCalledByMethod(long callingMethodHash, String calledClass, long calledConHash)
   {
      HashMap calledClasses = (HashMap) conCalledByMethodInterceptors.get(callingMethodHash);
      TLongObjectHashMap calledMethods = (TLongObjectHashMap) calledClasses.get(calledClass);
      CallerConstructorInfo info = (CallerConstructorInfo) calledMethods.get(calledConHash);
      return info;
   }

   private CallerMethodInfo getConstructorCallerMethodInfo(int callingIndex, String calledClass, long calledMethodHash)
   {
      HashMap calledClasses = methodCalledByConInterceptors[callingIndex];
      TLongObjectHashMap calledMethods = (TLongObjectHashMap) calledClasses.get(calledClass);
      CallerMethodInfo info = (CallerMethodInfo) calledMethods.get(calledMethodHash);
      return info;
   }

   private CallerConstructorInfo getConCalledByCon(int callingIndex, String calledClass, long calledConHash)
   {
      HashMap calledClasses = conCalledByConInterceptors[callingIndex];
      TLongObjectHashMap calledMethods = (TLongObjectHashMap) calledClasses.get(calledClass);
      CallerConstructorInfo info = (CallerConstructorInfo) calledMethods.get(calledConHash);
      return info;
   }

   public InvocationResponse dynamicInvoke(Object target, Invocation invocation)
           throws Throwable
   {
      // Only need to set Method because fields will already have been set.
      if (invocation instanceof MethodInvocation)
      {
         InstanceAdvised advised = (InstanceAdvised) target;
         Interceptor[] aspects = null;
         MethodInvocation methodInvocation = (MethodInvocation) invocation;
         long hash = methodInvocation.getMethodHash();
         MethodInfo info = (MethodInfo) methodInterceptors.get(hash);
         aspects = info.interceptors;
         if (advised != null)
         {
            aspects = advised._getInstanceAdvisor().getInterceptors(aspects);
         }
         MethodInvocation nextInvocation = new MethodInvocation(info, aspects);
         nextInvocation.setTargetObject(target);
         nextInvocation.setArguments(methodInvocation.getArguments());
         nextInvocation.setAdvisor(this);
         InvocationResponse response = new InvocationResponse(nextInvocation.invokeNext());
         response.setContextInfo(nextInvocation.getResponseContextInfo());
         return response;
      }
      throw new RuntimeException("dynamic field invocations not supported yet!");
   }


   public Object invokeNew(Object[] args, int idx) throws Throwable
   {
      Interceptor[] cInterceptors = constructorInterceptors[idx];
      ConstructorInvocation invocation = new ConstructorInvocation(cInterceptors);

      invocation.setAdvisor(this);
      invocation.setArguments(args);
      invocation.setConstructor(constructors[idx]);
      return invocation.invokeNext();
   }

   /**
    * Invokes interceptor chain.
    * This is the beginning
    */
   public Object invokeRead(Object target, int index)
           throws Throwable
   {
      Interceptor[] aspects = fieldReadInterceptors[index];
      FieldReadInvocation invocation;
      if (target != null)
      {
         InstanceAdvised advised = (InstanceAdvised) target;
         ClassInstanceAdvisor advisor = (ClassInstanceAdvisor) advised._getInstanceAdvisor();
         if (advisor != null && (advisor.appendedInterceptors != null || advisor.insertedInterceptors != null))
         {
            aspects = advisor.getInterceptors(aspects);
         }
      }
      invocation = new FieldReadInvocation(advisedFields[index], index, aspects);
      invocation.setAdvisor(this);
      invocation.setTargetObject(target);
      return invocation.invokeNext();
   }

   /**
    * Invokes interceptor chain.
    * This is the beginning
    */
   public Object invokeWrite(Object target, int index, Object value)
           throws Throwable
   {
      Interceptor[] aspects = fieldWriteInterceptors[index];
      FieldWriteInvocation invocation;
      if (target != null)
      {
         InstanceAdvised advised = (InstanceAdvised) target;
         ClassInstanceAdvisor advisor = (ClassInstanceAdvisor) advised._getInstanceAdvisor();
         if (advisor != null && (advisor.appendedInterceptors != null || advisor.insertedInterceptors != null))
         {
            aspects = advised._getInstanceAdvisor().getInterceptors(aspects);
         }
      }
      invocation = new FieldWriteInvocation(advisedFields[index], index, value, aspects);
      invocation.setAdvisor(this);
      invocation.setTargetObject(target);
      return invocation.invokeNext();
   }

   /**
    * Helper method
    * This is the end of the chain
    */
   public Object invoke(Invocation invocation) throws Throwable
   {
      if (invocation instanceof FieldWriteInvocation)
      {
         FieldWriteInvocation fieldInvocation = (FieldWriteInvocation) invocation;
         Object target = fieldInvocation.getTargetObject();
         Object val = fieldInvocation.getValue();
         Field field = fieldInvocation.getField();
         field.set(target, val);
         return null;
      }
      else if (invocation instanceof FieldReadInvocation)
      {
         FieldReadInvocation fieldInvocation = (FieldReadInvocation) invocation;
         Object target = fieldInvocation.getTargetObject();
         Field field = fieldInvocation.getField();
         return field.get(target);
      }
      else if (invocation instanceof MethodInvocation)
      {
         MethodInvocation methodInvocation = (MethodInvocation) invocation;
         return invokeWithoutAdvisement(methodInvocation.getTargetObject(),
                 methodInvocation.getMethodHash(),
                 methodInvocation.getArguments());
      }
      else if (invocation instanceof ConstructorInvocation)
      {
         ConstructorInvocation cInvocation = (ConstructorInvocation) invocation;
         Object[] arguments = cInvocation.getArguments();
         Constructor constructor = cInvocation.getConstructor();
         return invokeNewWithoutAdvisement(arguments, constructor);
      }
      throw new IllegalStateException("Unknown Invocation type: " + invocation.getClass().getName());
   }

}
