/***************************************
 *                                     *
 *  JBoss: The OpenSource J2EE WebOS   *
 *                                     *
 *  Distributable under LGPL license.  *
 *  See terms of license at gnu.org.   *
 *                                     *
 ***************************************/

package org.jboss.aop;

import EDU.oswego.cs.dl.util.concurrent.ConcurrentReaderHashMap;
import org.jboss.aop.advice.AdviceStack;
import org.jboss.aop.advice.AspectDefinition;
import org.jboss.aop.advice.Interceptor;
import org.jboss.aop.advice.InterceptorFactory;
import org.jboss.aop.joinpoint.Joinpoint;
import org.jboss.aop.metadata.SimpleMetaData;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.WeakHashMap;

/**
 * Holds an object instance's metadata and attached interceptors
 *
 * @author <a href="mailto:bill@jboss.org">Bill Burke</a>
 * @version $Revision: 1.20 $
 */
public class ClassInstanceAdvisor implements InstanceAdvisor, java.io.Serializable
{
   protected ArrayList insertedInterceptors = null;
   protected ArrayList appendedInterceptors = null;
   protected SimpleMetaData metadata;
   protected Object instance;
   protected transient Advised advised;
   protected transient Advisor advisor;
   public boolean hasInstanceAspects = false;
   /** aspects is a weak hash map of AspectDefinitions so that perinstance advices can be undeployed/redeployed */
   protected transient WeakHashMap aspects;
   protected transient WeakHashMap joinpointAspects;

   public ClassInstanceAdvisor()
   {
   }

   public ClassInstanceAdvisor(Object obj)
   {
      this.instance = obj;
      if (instance instanceof Advised)
      {
         advised = (Advised) instance;
         advisor = advised._getAdvisor();
         if (advisor instanceof ClassAdvisor)
         {
            initializeAspects();
            initializeJoinpointAspects();
         }
      }
      else
         advised = null;
   }

   protected synchronized void initializeAspects()
   {
      if (advisor == null) return;
      if (aspects != null) return; // doublecheck I know, but I don't want to do synchronization if not needed
      ClassAdvisor cadvisor = (ClassAdvisor) advisor;
      Set defs = cadvisor.getPerInstanceAspectDefinitions();
      if (defs.size() > 0)
      {
         aspects = new WeakHashMap();
         Iterator it = defs.iterator();
         while (it.hasNext())
         {
            AspectDefinition def = (AspectDefinition) it.next();
            Object aspect = def.getFactory().createPerInstance(advisor, this);
            aspects.put(def, aspect);
         }
      }

   }

   private synchronized void initializeJoinpointAspects()
   {
      if (advisor == null) return;
      if (joinpointAspects != null) return; // doublecheck I know, but I don't want to do synchronization if not needed
      ClassAdvisor cadvisor = (ClassAdvisor) advisor;
      Map jpAspects = cadvisor.getPerInstanceJoinpointAspectDefinitions();
      if (jpAspects.size() > 0)
      {
         joinpointAspects = new WeakHashMap();
         Iterator it = jpAspects.keySet().iterator();
         while (it.hasNext())
         {
            AspectDefinition def = (AspectDefinition) it.next();
            ConcurrentReaderHashMap joins = new ConcurrentReaderHashMap();
            joinpointAspects.put(def, joins);
            Set joinpoints = (Set)jpAspects.get(def);
            Iterator jps = joinpoints.iterator();
            while (jps.hasNext())
            {
               Object joinpoint = jps.next();
               joins.put(joinpoint, def.getFactory().createPerJoinpoint(advisor, this, (Joinpoint)joinpoint));
            }
         }
      }
   }

   public Object getPerInstanceAspect(String def)
   {
      Iterator it = aspects.keySet().iterator();
      while (it.hasNext())
      {
         AspectDefinition d = (AspectDefinition) it.next();
         if (d.getName().equals(def)) return aspects.get(d);
      }
      return null;
   }

   public Object getPerInstanceAspect(AspectDefinition def)
   {
      // aspects is a weak hash map of AspectDefinitions so that perinstance advices can be undeployed/redeployed
      if (aspects == null)
      {
         initializeAspects();
         return aspects.get(def);
      }
      Object aspect = aspects.get(def);
      if (aspect == null)
      {
         synchronized (this) // doublecheck, but I don't want to synchronize everywhere and dynamic aspects are rare
         {
            aspect = aspects.get(def);
            if (aspect != null) return aspect;
            if (advisor != null && advisor instanceof ClassAdvisor)
            {
               ClassAdvisor cadvisor = (ClassAdvisor) advisor;
               cadvisor.getPerInstanceAspectDefinitions().add(def);
               aspect = def.getFactory().createPerInstance(advisor, this);
               WeakHashMap copy = new WeakHashMap(aspects);
               copy.put(def, aspect);
               aspects = copy;
            }
         }
      }
      return aspect;
   }

   public Object getPerInstanceJoinpointAspect(Joinpoint joinpoint, AspectDefinition def)
   {
      // aspects is a weak hash map of AspectDefinitions so that perinstance advices can be undeployed/redeployed
      if (joinpointAspects == null)
      {
         initializeJoinpointAspects();
         return getJoinpointAspect(def, joinpoint);
      }
      Object aspect = getJoinpointAspect(def, joinpoint);
      if (aspect == null)
      {
         synchronized (this) // doublecheck, but I don't want to synchronize everywhere and dynamic aspects are rare
         {
            aspect = getJoinpointAspect(def, joinpoint);
            if (aspect != null) return aspect;
            if (advisor != null && advisor instanceof ClassAdvisor)
            {
               ClassAdvisor cadvisor = (ClassAdvisor) advisor;
               cadvisor.addPerInstanceJoinpointAspect(joinpoint, def);
               aspect = def.getFactory().createPerJoinpoint(advisor, this, (Joinpoint)joinpoint);
               WeakHashMap copy = new WeakHashMap(joinpointAspects);
               Map map = (Map)copy.get(def);
               if (map == null)
               {
                  map = new ConcurrentReaderHashMap();
               }
               map.put(joinpoint, aspect);
               joinpointAspects = copy;
            }
         }
      }
      return aspect;
   }

   private Object getJoinpointAspect(AspectDefinition def, Joinpoint joinpoint)
   {
      if (joinpointAspects == null) return null;
      Map map = (Map)joinpointAspects.get(def);
      Object aspect = map.get(joinpoint);
      return aspect;
   }

   public SimpleMetaData getMetaData()
   {
      if (metadata == null)
      {
         synchronized (this) // doublecheck :(
         {
            if (metadata == null)
            {
               metadata = new SimpleMetaData();
            }
         }
      }
      return metadata;
   }

   public Interceptor[] getInterceptors()
   {
      ArrayList newlist = new ArrayList();
      if (insertedInterceptors != null) newlist.addAll(insertedInterceptors);
      if (appendedInterceptors != null) newlist.addAll(appendedInterceptors);
      return (Interceptor[]) newlist.toArray(new Interceptor[newlist.size()]);
   }

   /**
    * Called by the advisor
    */
   public Interceptor[] getInterceptors(Interceptor[] advisorChain)
   {
      if (insertedInterceptors == null && appendedInterceptors == null) return advisorChain;
      ArrayList newlist = new ArrayList();
      if (insertedInterceptors != null) newlist.addAll(insertedInterceptors);
      newlist.addAll(Arrays.asList(advisorChain));
      if (appendedInterceptors != null) newlist.addAll(appendedInterceptors);
      return (Interceptor[]) newlist.toArray(new Interceptor[newlist.size()]);
   }

   public void insertInterceptor(Interceptor interceptor)
   {
      ArrayList newList = new ArrayList();
      if (insertedInterceptors != null)
      {
         newList.addAll(insertedInterceptors);
      }
      newList.add(interceptor);
      insertedInterceptors = newList;
      hasInstanceAspects = true;
   }

   public void appendInterceptor(Interceptor interceptor)
   {
      ArrayList newList = new ArrayList();
      if (appendedInterceptors != null)
      {
         newList.addAll(appendedInterceptors);
      }
      newList.add(interceptor);
      appendedInterceptors = newList;
      hasInstanceAspects = true;
   }


   /**
    * This will not remove interceptor pointcuts!  You will have to do this through AspectManager
    */
   public void removeInterceptor(String name)
   {
      if (insertedInterceptors != null)
      {
         for (int i = 0; i < insertedInterceptors.size(); i++)
         {
            Interceptor interceptor = (Interceptor) insertedInterceptors.get(i);
            if (interceptor.getName().equals(name))
            {
               ArrayList newList = new ArrayList();
               newList.addAll(insertedInterceptors);
               newList.remove(i);
               insertedInterceptors = newList;
            }
         }
      }
      if (appendedInterceptors != null)
      {
         for (int i = 0; i < appendedInterceptors.size(); i++)
         {
            Interceptor interceptor = (Interceptor) appendedInterceptors.get(i);
            if (interceptor.getName().equals(name))
            {
               ArrayList newList = new ArrayList();
               newList.addAll(appendedInterceptors);
               newList.remove(i);
               appendedInterceptors = newList;
            }
         }
      }
      hasInstanceAspects = ((insertedInterceptors != null && insertedInterceptors.size() > 0)
              || (appendedInterceptors != null && appendedInterceptors.size() > 0));
   }

   public final boolean hasAspects()
   {
      return hasInstanceAspects;
   }

   public void insertInterceptorStack(String stackName)
   {
      AdviceStack stack = AspectManager.instance().getAdviceStack(stackName);
      if (stack == null) throw new RuntimeException("Stack " + stackName + " not found.");

      ClassAdvisor classAdvisor = null;
      if (instance instanceof Advised)
      {
         Advised advised = (Advised) instance;
         classAdvisor = ((ClassAdvisor) advised._getAdvisor());
      }
      Iterator it = stack.getInterceptorFactories().iterator();
      while (it.hasNext())
      {
         InterceptorFactory factory = (InterceptorFactory) it.next();
         if (!factory.isDeployed()) continue;
         Interceptor interceptor = factory.create(classAdvisor, null);
         insertInterceptor(interceptor);
      }
   }

   public void appendInterceptorStack(String stackName)
   {
      AdviceStack stack = AspectManager.instance().getAdviceStack(stackName);
      if (stack == null) throw new RuntimeException("Stack " + stackName + " not found.");

      ClassAdvisor classAdvisor = null;
      if (instance instanceof Advised)
      {
         Advised advised = (Advised) instance;
         classAdvisor = ((ClassAdvisor) advised._getAdvisor());
      }
      Iterator it = stack.getInterceptorFactories().iterator();
      while (it.hasNext())
      {
         InterceptorFactory factory = (InterceptorFactory) it.next();
         if (!factory.isDeployed()) continue;
         Interceptor interceptor = factory.create(classAdvisor, null);
         appendInterceptor(interceptor);
      }
   }

   public void removeInterceptorStack(String stackName)
   {
      AdviceStack stack = AspectManager.instance().getAdviceStack(stackName);
      if (stack == null) throw new RuntimeException("Stack " + stackName + " not found.");

      ClassAdvisor classAdvisor = null;
      if (instance instanceof Advised)
      {
         Advised advised = (Advised) instance;
         classAdvisor = ((ClassAdvisor) advised._getAdvisor());
      }
      Iterator it = stack.getInterceptorFactories().iterator();
      while (it.hasNext())
      {
         InterceptorFactory factory = (InterceptorFactory) it.next();
         if (!factory.isDeployed()) continue;
         Interceptor interceptor = factory.create(classAdvisor, null);
         removeInterceptor(interceptor.getName());
      }
   }


}
