/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.aop.advice;

import org.jboss.aop.Advisor;
import org.jboss.aop.AspectManager;
import org.jboss.aop.pointcut.Pointcut;
import org.jboss.aop.pointcut.PointcutExpression;
import org.jboss.aop.pointcut.ast.ASTCFlowExpression;
import org.jboss.aop.pointcut.ast.ParseException;
import org.jboss.aop.pointcut.ast.PointcutExpressionParser;

import java.io.StringReader;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;

/**
 * Comment
 *
 * @author <a href="mailto:bill@jboss.org">Bill Burke</a>
 * @version $Revision: 1.9 $
 *
 **/
public class AdviceBinding
{
   private static volatile long counter = 0;

   protected String name;
   protected Pointcut pointcut;
   protected ASTCFlowExpression cflow;
   protected String cflowString;

   protected HashSet advisors = new HashSet();
   protected InterceptorFactory[] interceptorFactories;

   public AdviceBinding(String name, Pointcut p, ASTCFlowExpression cflow, String cflowString, InterceptorFactory[] factories) throws ParseException
   {
      this.name = name;
      interceptorFactories = factories;
      this.cflow = cflow;

      pointcut = p;
      this.cflowString = cflowString;
   }

   /**
    * This constructor is used for creation of AdviceBinding programmatically
    *
    * @param pointcutExpression
    * @param cflow
    * @throws ParseException
    */
   public AdviceBinding(String pointcutExpression, String cflow) throws ParseException
   {
      this(Long.toString(System.currentTimeMillis()) + ":" + Long.toString(counter++), pointcutExpression, cflow);
   }

   /**
    * This constructor is used for creation of AdviceBinding programmatically
    *
    * @param pointcutExpression
    * @param cflow
    * @throws ParseException
    */
   public AdviceBinding(String name, String pointcutExpression, String cflow) throws ParseException
   {
      this.name = name;
      pointcut = new PointcutExpression(Long.toString(System.currentTimeMillis()) + ":" + Long.toString(counter++), pointcutExpression);
      if (cflow != null)
      {
         cflowString = cflow;
         this.cflow = new PointcutExpressionParser(new StringReader(cflowString)).CFlowExpression();
      }
      interceptorFactories = new InterceptorFactory[0];
   }


   /**
    * Add an interceptor to chain.  This is an actual class
    * that implements Interceptor.  A GenericInterceptorFactory will
    * be created to wrap the class.
    * @param clazz
    */
   public void addInterceptor(Class clazz)
   {
      List list = Arrays.asList(interceptorFactories);
      list = new ArrayList(list);
      list.add(new GenericInterceptorFactory(clazz));
      interceptorFactories = (InterceptorFactory[])list.toArray(new InterceptorFactory[list.size()]);
   }

   public String getName()
   {
      return name;
   }

   public InterceptorFactory[] getInterceptorFactories()
   {
      return interceptorFactories;
   }

   public void addAdvisor(Advisor advisor)
   {
      if (AspectManager.verbose) System.out.println("[debug] added advisor: " + advisor.getName() + " from binding: " + name);
      // Don't hold a direct reference to an advisor because of undeploy and redeploy
      advisors.add(advisor.getName());
   }

   public boolean hasAdvisors()
   {
      return advisors.size() > 0;
   }

   public void clearAdvisors()
   {
      for (Iterator it = advisors.iterator(); it.hasNext();)
      {
         String advisorName = (String) it.next();
         Advisor advisor = AspectManager.instance().getAdvisor(advisorName);
         if (advisor != null)
            advisor.removeAdviceBinding(this);
      }
      advisors.clear();
   }

   public boolean equals(Object obj)
   {
      if (obj == this) return true;
      if (!(obj instanceof AdviceBinding)) return false;
      return ((AdviceBinding) obj).getName().equals(name);
   }

   public int hashCode()
   {
      return name.hashCode();
   }

   public Pointcut getPointcut()
   {
      return pointcut;
   }

   public ASTCFlowExpression getCFlow()
   {
      return cflow;
   }

   public String getCFlowString()
   {
      return cflowString;
   }
}
