/***************************************
 *                                     *
 *  JBoss: The OpenSource J2EE WebOS   *
 *                                     *
 *  Distributable under LGPL license.  *
 *  See terms of license at gnu.org.   *
 *                                     *
 ***************************************/

package org.jboss.aop.instrument;

import javassist.CannotCompileException;
import javassist.CtBehavior;
import javassist.CtClass;
import javassist.CtConstructor;
import javassist.CtField;
import javassist.CtMethod;
import javassist.CtNewMethod;
import javassist.Modifier;
import javassist.NotFoundException;
import javassist.expr.ExprEditor;
import javassist.expr.MethodCall;
import javassist.expr.NewExpr;
import org.jboss.aop.AOPClassPool;
import org.jboss.aop.AspectManager;
import org.jboss.aop.ClassAdvisor;
import org.jboss.aop.instrument.FieldAccessTransformer.NotOptimizedTransformer;
import org.jboss.aop.instrument.FieldAccessTransformer.OptimizedTransformer;
import org.jboss.aop.pointcut.Pointcut;
import org.jboss.aop.util.JavassistMethodHashing;

import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

/**
 * Instruments Caller pointcuts
 * 
 * @author <a href="mailto:kabirkhan@bigfoot.com">Kabir Khan</a>
 */

public class CallerTransformer
{
   Instrumentor instrumentor;
   boolean optimize;
   NotOptimizedTransformer notOptimizedTransformer;
   OptimizedTransformer optimizedTransformer;

   // Static --------------------------------------------------------
		 
   // Constructors --------------------------------------------------
   public CallerTransformer(Instrumentor instrumentor)
   {
      this.instrumentor = instrumentor;
      this.optimize = AspectManager.optimize;
   }
   
   public boolean applyCallerPointcuts(CtClass clazz, ClassAdvisor advisor) throws CannotCompileException
   {
      CallerExprEditor expr = new CallerExprEditor(advisor, clazz);
      CtMethod[] methods = clazz.getDeclaredMethods();
      for (int i = 0; i < methods.length; i++)
      {
         if (Instrumentor.isAdvisable(methods[i]))
         {
            methods[i].instrument(expr);
         }
      }

      CtConstructor[] cons = clazz.getDeclaredConstructors();
      for (int i = 0; i < cons.length; i++)
      {
         cons[i].instrument(expr);
      }

      return expr.appliedCallerBinding;
   }
   
   class CallerExprEditor extends ExprEditor
   {
      CtClass callingClass;
      ClassAdvisor advisor;
      List constructors;
      public boolean appliedCallerBinding = false;

      HashMap callerInfos = new HashMap();
      int invocationCounter = 0;

      public CallerExprEditor(ClassAdvisor advisor, CtClass callingClass)
      {
         this.advisor = advisor;
         this.callingClass = callingClass;
         this.constructors = instrumentor.getConstructors(callingClass);
      }
      
      private String getHashString(long hash)
      {
         if (hash < 0)
         {
            return  "_N_" + (-1 * hash);
         }
         else
         {
            return "_" + hash;
         }
      }

      private String getUniqueInvocationfieldname(long callingHash, String classname, long calledHash)
      {
         classname = classname.replace('.', '_');
         classname = classname.replace('/', '_');
            
         return getHashString(callingHash) + classname + getHashString(calledHash);
      }

      private String getUniqueInvocationClassname(String classname){
         return classname + "_" + (++invocationCounter) + "_";
      }
      
      private String getConstructorInfoName(long callingHash, String classname, long calledHash)
      {
         return "aop$constructorCall_" + getUniqueInvocationfieldname(callingHash, classname, calledHash);
      }
      
      private String getOptimizedConCalledByMethodInvocationClassName(long callingHash, String classname, long calledHash)
      {
         return getUniqueInvocationClassname(classname) + "ConByMInvocation";
      }
            
      private String getConstructorConstructorInfoName(long callingIndex, String classname, long calledHash)
      {
         return "aop$constructorCall_con_" + getUniqueInvocationfieldname(callingIndex, classname, calledHash);
      }

      private String getOptimizedConCalledByConInvocationClassName(long callingIndex, String classname, long calledHash)
      {
         return getUniqueInvocationClassname(classname) + "ConByConInvocation";
      }

      private String getCallerInfoName(long callingHash, String classname, long calledHash)
      {
         return "aop$methodCall_" + getUniqueInvocationfieldname(callingHash, classname, calledHash);
      }

      private String getOptimizedMethodCalledByMethodClassName(long callingHash, String classname, long calledHash)
      {
         return getUniqueInvocationClassname(classname) + "MByMInvocation";
     }

      private String getConstructorCallerInfoName(int index, String classname, long calledHash)
      {
         return "aop$methodCall_" + getUniqueInvocationfieldname(index, classname, calledHash);
      }

      private String getOptimizedMethodCalledByConstructorClassName(int callingIndex, String classname, long calledHash)
      {
         return getUniqueInvocationClassname(classname) + "MByConInvocation";
      }

      private void addCallerMethodInfoField(CtClass clazz, String fieldName, long callingHash, String classname, long calledHash) throws Exception
      {
         CtClass type = instrumentor.forName("org.jboss.aop.CallerMethodInfo");
         CtField field = new CtField(type, fieldName, clazz);
         field.setModifiers(
                 Modifier.PRIVATE | Modifier.STATIC);
         CtField.Initializer initializer = CtField.Initializer.byExpr(
                 "aop$classAdvisor$aop.resolveCallerMethodInfo(" + callingHash + "L, \"" + classname + "\", " + calledHash + "L)");
         clazz.addField(field, initializer);
      }

      private void addCallerConstructorInfoField(CtClass clazz, String fieldName, long callingHash, String classname, long calledHash) throws Exception
      {
         CtClass type = instrumentor.forName("org.jboss.aop.CallerConstructorInfo");
         CtField field = new CtField(type, fieldName, clazz);
         field.setModifiers(
                 Modifier.PRIVATE | Modifier.STATIC);
         CtField.Initializer initializer = CtField.Initializer.byExpr(
                 "aop$classAdvisor$aop.resolveCallerConstructorInfo(" + callingHash + "L, \"" + classname + "\", " + calledHash + "L)");
         clazz.addField(field, initializer);
      }

      private void addConstructorCallerConstructorInfoField(CtClass clazz, String fieldName, int callingIndex, String classname, long calledHash) throws Exception
      {
         CtClass type = instrumentor.forName("org.jboss.aop.CallerConstructorInfo");
         CtField field = new CtField(type, fieldName, clazz);
         field.setModifiers(
                 Modifier.PRIVATE | Modifier.STATIC);
         CtField.Initializer initializer = CtField.Initializer.byExpr(
                 "aop$classAdvisor$aop.resolveConstructorCallerConstructorInfo(" + callingIndex + ", \"" + classname + "\", " + calledHash + "L)");
         clazz.addField(field, initializer);
      }

      private void addConstructorCallerMethodInfoField(CtClass clazz, String fieldName, int callingIndex, String classname, long calledHash) throws Exception
      {
         CtClass type = instrumentor.forName("org.jboss.aop.CallerMethodInfo");
         CtField field = new CtField(type, fieldName, clazz);
         field.setModifiers(
                 Modifier.PRIVATE | Modifier.STATIC);
         CtField.Initializer initializer = CtField.Initializer.byExpr(
                 "aop$classAdvisor$aop.resolveConstructorCallerMethodInfo(" + callingIndex + ", \"" + classname + "\", " + calledHash + "L)");
         clazz.addField(field, initializer);
      }

      public void edit(MethodCall call) throws CannotCompileException
      {
         try
         {
            String classname = call.getClassName();
            String methodName = call.getMethodName();
            if (ClassAdvisor.isWithoutAdvisement(methodName)
                    || methodName.startsWith("_")
                    || classname.startsWith("org.jboss.aop")
                    || call.getMethodName().equals("class$") // todo not sure why this is part of the method call
            )
               return;

            CtBehavior behavior = call.where();

            boolean hasPointcut = false;

            // todo shouldn't iterate every time.  must be a better way
            Iterator it = AspectManager.instance().getPointcuts().values().iterator();
            while (it.hasNext())
            {
               Pointcut p = (Pointcut) it.next();
               if (p.matchesCall(advisor, call))
               {
                  hasPointcut = true;
                  break;
               }
            }
            if (hasPointcut)
            {
               if (behavior instanceof CtMethod)
                  modifyMethod(call, classname);
               else if (behavior instanceof CtConstructor) modifyConstructor(call, classname);
            }
         }
         catch (Exception ex)
         {
            System.err.println("error getting:" + call.getClassName() + ". '" + call.getMethodName() + "'");
            ex.printStackTrace();
            throw new CannotCompileException(ex);
         }
      }

      protected void modifyConstructor(MethodCall call, String classname) throws Exception, NotFoundException
      {
         instrumentor.setupBasics(callingClass);
         
         CtConstructor con = (CtConstructor) call.where();
         int callingIndex = constructors.indexOf(con);
         long calledHash = JavassistMethodHashing.methodHash(call.getMethod());
         String callerInfoField = getConstructorCallerInfoName(callingIndex, classname, calledHash);
         CtMethod calledMethod = call.getMethod();

         if (callerInfos.get(callerInfoField) == null)
         {
            //Create MethodCalledByConstructorInvocation class
            String invocationClassName = optimize ?
                  		createOptimizedMethodCalledByConInvocationClass(calledMethod, callingIndex, calledHash)
                  		: "whatever";
            callerInfos.put(callerInfoField, invocationClassName);
            addConstructorCallerMethodInfoField(callingClass, callerInfoField, callingIndex, classname, calledHash);
         }
         
         if (optimize)
         {
            String invocationClassName = (String)callerInfos.get(callerInfoField);
            String typedTargetObject = javassist.Modifier.isStatic(calledMethod.getModifiers()) ? "" : "invocation.typedTargetObject=$0;";
                                  
            String replaced =
               "if (" + callerInfoField + ".interceptors != (org.jboss.aop.advice.Interceptor[])null) { " +
               invocationClassName + " invocation = new "  
               + invocationClassName + "    (" + callerInfoField + ".method, $args, " + Instrumentor.HELPER_FIELD_NAME + ".getConstructors()["+ callingIndex + "]," + callerInfoField + ".interceptors);" +
               typedTargetObject +
               MethodExecutionTransformer.setArguments(calledMethod.getParameterTypes()) +
               "invocation.setTargetObject($0);" + 
               "$_ = ($r)invocation.invokeNext();" +
               "} else { " +
               "$_ = $proceed($$); " +
               "}";
            call.replace(replaced);
         }
         else
         {
            String replaced =
                    "if (" + callerInfoField + ".interceptors != (org.jboss.aop.advice.Interceptor[])null) { " +
                    "$_ = ($r)aop$classAdvisor$aop.invokeConstructorCaller("
                    + callingIndex + ", $0, $args, " + callerInfoField + "); " +
                    "} else { " +
                    "$_ = $proceed($$); " +
                    "}";
            call.replace(replaced);
         }

         appliedCallerBinding = true;
      }

      protected void modifyMethod(MethodCall call, String classname) throws Exception, NotFoundException
      {
         instrumentor.setupBasics(callingClass);
         long callingHash = JavassistMethodHashing.methodHash((CtMethod) call.where());
         long calledHash = JavassistMethodHashing.methodHash(call.getMethod());
         String callerInfoField = getCallerInfoName(callingHash, classname, calledHash);
         CtMethod calledMethod = call.getMethod();
         
         if (callerInfos.get(callerInfoField) == null)
         {
            String invocationClassName = optimize ? 
                        createOptimizedMethodCalledByMethodInvocationClass(calledMethod, callingHash, calledHash)
                        : "whatever";
            callerInfos.put(callerInfoField, invocationClassName);
            addCallerMethodInfoField(callingClass, callerInfoField, callingHash, classname, calledHash);
         }

         if (optimize)
         {
            String invocationClassName = (String)callerInfos.get(callerInfoField);
            String typedTargetObject = javassist.Modifier.isStatic(calledMethod.getModifiers()) ? "" : "invocation.typedTargetObject=$0;";

            String replaced =
               "if (" + callerInfoField + ".interceptors != (org.jboss.aop.advice.Interceptor[])null) { " +
               invocationClassName + " invocation = new "  
               + invocationClassName + "    (" + callerInfoField + ".interceptors, " + callerInfoField + ".method, $args, " + Instrumentor.HELPER_FIELD_NAME + ".getClazz(), " + callingHash + "L);" +               
               MethodExecutionTransformer.setArguments(calledMethod.getParameterTypes()) +
               typedTargetObject +
               "invocation.setTargetObject($0);" + 
               "$_ = ($r)invocation.invokeNext();" +
               "} else { " +
               "$_ = $proceed($$); " +
               "}";
            
            call.replace(replaced);
         }
         else
         {
            String replaced =
                    "if (" + callerInfoField + ".interceptors != (org.jboss.aop.advice.Interceptor[])null) { " +
                    "$_ = ($r)aop$classAdvisor$aop.invokeCaller("
                    + callingHash + "L, $0, $args, " + callerInfoField + "); " +
                    "} else { " +
                    "$_ = $proceed($$); " +
                    "}";
            call.replace(replaced);
         }
         appliedCallerBinding = true;
      }

      protected void modifyMethod(NewExpr call, String classname) throws Exception, NotFoundException
      {
         instrumentor.setupBasics(callingClass);
         long callingHash = JavassistMethodHashing.methodHash((CtMethod)call.where());
         long calledHash = JavassistMethodHashing.constructorHash(call.getConstructor());
         String callerInfoField = getConstructorInfoName(callingHash, classname, calledHash);
         CtConstructor calledConstructor = call.getConstructor();

         if (callerInfos.get(callerInfoField) == null)
         {
            String invocationClassName = optimize ? 
                     createOptimizedConCalledByMethodInvocationClass(calledConstructor, callingHash, calledHash)
                     : "whatever";
            callerInfos.put(callerInfoField, invocationClassName);
            addCallerConstructorInfoField(callingClass, callerInfoField, callingHash, classname, calledHash);
         }

         //If target constructor is advisable we do not want to create an optimized invocation class
         //since the interceptor chain for its execution pointcut does not get invoked. 
         ClassAdvisor adv = AspectManager.instance().getTempClassAdvisor(calledConstructor.getDeclaringClass());
         boolean isAdvisableTargetCon = ConstructorExecutionTransformer.isAdvisableConstructor(calledConstructor, adv);
         if (optimize && !isAdvisableTargetCon)
         {
            String invocationClassName = (String)callerInfos.get(callerInfoField);
            String replaced =
               "if (" + callerInfoField + ".interceptors != (org.jboss.aop.advice.Interceptor[])null) { " +
               invocationClassName + " invocation = new "  
               + invocationClassName + "    (" + callerInfoField + ".interceptors, " + callerInfoField + ", " + callingHash + "L," + Instrumentor.HELPER_FIELD_NAME + ".getClazz(), $args);" +               
               MethodExecutionTransformer.setArguments(calledConstructor.getParameterTypes()) +
               "$_ = ($r)invocation.invokeNext();" +
               "} else { " +
               "$_ = $proceed($$); " +
               "}";
            call.replace(replaced);
         }
         else
         {
          
            String replaced =
                    "if (" + callerInfoField + ".interceptors != (org.jboss.aop.advice.Interceptor[])null) { " +
                    "$_ = ($r)aop$classAdvisor$aop.invokeConCalledByMethod("
                    + callingHash + "L, $args, " + callerInfoField + "); " +
                    "} else { " +
                    "$_ = $proceed($$); " +
                    "}";

            call.replace(replaced);
         }
         appliedCallerBinding = true;
      }


      protected void modifyConstructor(NewExpr call, String classname) throws Exception, NotFoundException
      {
         instrumentor.setupBasics(callingClass);
         CtConstructor con = (CtConstructor) call.where();
         int callingIndex = constructors.indexOf(con);
         long calledHash = JavassistMethodHashing.constructorHash(call.getConstructor());
         String callerInfoField = getConstructorConstructorInfoName(callingIndex, classname, calledHash);
         CtConstructor calledConstructor = call.getConstructor();
         if (callerInfos.get(callerInfoField) == null)
         {
            //Create ConstructorCalledByConstructorInvocation class
            String invocationClassName = optimize ?
                     createOptimizedConCalledByConInvocationClass(calledConstructor, callingIndex, calledHash)
                     : "whatever";

            callerInfos.put(callerInfoField, invocationClassName);
            addConstructorCallerConstructorInfoField(callingClass, callerInfoField, callingIndex, classname, calledHash);
         }
         
         //If target constructor is advisable we do not want to create an optimized invocation class
         //since the interceptor chain for its execution pointcut does not get invoked. 
         ClassAdvisor adv = AspectManager.instance().getTempClassAdvisor(calledConstructor.getDeclaringClass());
         boolean isAdvisableTargetCon = ConstructorExecutionTransformer.isAdvisableConstructor(calledConstructor, adv); 
         if (optimize && !isAdvisableTargetCon)
         {
            String invocationClassName = (String)callerInfos.get(callerInfoField);
            
            String replaced =
               "if (" + callerInfoField + ".interceptors != (org.jboss.aop.advice.Interceptor[])null) { " +
               invocationClassName + " invocation = new " + invocationClassName + "    (aop$classAdvisor$aop.getConstructors()[" + callingIndex + "], " + callerInfoField + ", $args);" +
               MethodExecutionTransformer.setArguments(calledConstructor.getParameterTypes()) +
               "$_ = ($r)invocation.invokeNext();" +
               "} else { " +
               "$_ = $proceed($$); " +
               "}";
            
            call.replace(replaced);
            
         }
         else
         {
            String replaced =
               "if (" + callerInfoField + ".interceptors != (org.jboss.aop.advice.Interceptor[])null) { " +
               "$_ = ($r)aop$classAdvisor$aop.invokeConCalledByCon("
               + callingIndex + ", $args, " + callerInfoField + "); " +
               "} else { " +
               "$_ = $proceed($$); " +
               "}";

            call.replace(replaced);
         }
         appliedCallerBinding = true;
      }

      public void edit(NewExpr call) throws CannotCompileException
      {
         try
         {
            String classname = call.getClassName();

            boolean hasPointcut = false;

            // todo shouldn't iterate every time.  must be a better way
            Iterator it = AspectManager.instance().getPointcuts().values().iterator();
            while (it.hasNext())
            {
               Pointcut p = (Pointcut) it.next();
               if (p.matchesCall(advisor, call))
               {
                  hasPointcut = true;
                  break;
               }
            }
            if (hasPointcut)
            {
               CtBehavior behavior = call.where();
               if (behavior instanceof CtMethod)
                  modifyMethod(call, classname);
               else if (behavior instanceof CtConstructor) modifyConstructor(call, classname);
            }
         }
         catch (Exception ex)
         {
            System.out.println(ex.getMessage());
            ex.printStackTrace();
            throw new CannotCompileException(ex);
         }
      }
 
      private String createOptimizedConCalledByConInvocationClass(CtConstructor con, int callingIndex, long calledHash) throws Exception
      {
         //TODO: Merge this method with createOptimizedConCalledByMethodInvocationClass()
         AOPClassPool pool = (AOPClassPool) instrumentor.getClassPool();
         CtClass conInvocation = pool.get("org.jboss.aop.joinpoint.ConstructorCalledByConstructorInvocation");

         ////////////////
         //Create the class
         String className = getOptimizedConCalledByConInvocationClassName(callingIndex, callingClass.getName(), calledHash);
         boolean makeInnerClass = Modifier.isPrivate(con.getModifiers());
         CtClass invocation = TransformerCommon.makeInvocationClass(pool, makeInnerClass, callingClass, className, conInvocation); 

         CtClass[] params = con.getParameterTypes();
         TransformerCommon.addArgumentFieldsToInvocation(invocation, params);

         /////////
         //Create invokeNext() body
         CtMethod in = conInvocation.getDeclaredMethod("invokeNext");
         CtMethod invokeNext = CtNewMethod.make(in.getReturnType(),
                  "invokeNext", in.getParameterTypes(), in.getExceptionTypes(),
                  null, invocation);
         invokeNext.setModifiers(in.getModifiers());

         String code = TransformerCommon.STANDARD_INTERCEPTOR_CODE;

         code += "return new " + con.getDeclaringClass().getName() + "(";
         for (int i = 0 ; i < params.length ; i++)
         {
            if (i > 0) code += ", ";
            code += "arg" + i;
         }
         code += ");  ";
         code += "}";
         try
         {
            invokeNext.setBody(code);
         }
         catch (CannotCompileException e)
         {
            System.out.println(code);
            throw e;
         }
         invocation.addMethod(invokeNext);
         TransformerCommon.addGetArguments(pool, invocation, con.getParameterTypes());

         ////////////
         //Create copy() method
         CtMethod copyTemplate = conInvocation.getDeclaredMethod("copy");

         CtMethod copy = CtNewMethod.make(copyTemplate.getReturnType(), "copy",
                  copyTemplate.getParameterTypes(), copyTemplate
                           .getExceptionTypes(), null, invocation);
         copy.setModifiers(copyTemplate.getModifiers());

         String copyCode = "{ "
                  + "   "
                  + invocation.getName()
                  + " wrapper = new "
                  + invocation.getName()
                  + "(this.calling, this.info, this.arguments);"
                  + "   wrapper.metadata = super.metadata; "
                  + "   wrapper.currentInterceptor = super.currentInterceptor; "
                  + "   wrapper.instanceResolver = super.instanceResolver; "
                  + "   wrapper.interceptors = super.interceptors; "
                  + "   wrapper.advisor = super.advisor; "
                  + "   wrapper.responseContextInfo = super.responseContextInfo; "
                  + "   wrapper.targetObject = super.targetObject; ";

         for (int i = 0 ; i < params.length ; i++)
         {
            copyCode += "   wrapper.arg" + i + " = this.arg" + i + "; ";
         }
         copyCode += "   return wrapper; }";

         try
         {
            copy.setBody(copyCode);
         }
         catch (CannotCompileException e)
         {
            System.out.println(copyCode);
            throw e;
         }
         invocation.addMethod(copy);

         TransformerCommon.compileOrLoadClass(callingClass, invocation);

         //Return fully qualified name of class (may be an inner class)
         return invocation.getName();
      }

      private String createOptimizedConCalledByMethodInvocationClass(CtConstructor con, long callingHash, long calledHash) throws Exception
      {
         AOPClassPool pool = (AOPClassPool) instrumentor.getClassPool();
         CtClass conInvocation = pool.get("org.jboss.aop.joinpoint.ConstructorCalledByMethodInvocation");

         ////////////////
         //Create the class
         String className = getOptimizedConCalledByMethodInvocationClassName(callingHash, callingClass.getName(), calledHash);
         boolean makeInnerClass = Modifier.isPrivate(con.getModifiers());
         CtClass invocation = TransformerCommon.makeInvocationClass(pool, makeInnerClass, callingClass, className, conInvocation); 
         
         CtClass[] params = con.getParameterTypes();
         TransformerCommon.addArgumentFieldsToInvocation(invocation, params);

         /////////
         //Create invokeNext() body
         CtMethod in = conInvocation.getDeclaredMethod("invokeNext");
         CtMethod invokeNext = CtNewMethod.make(in.getReturnType(),
                  "invokeNext", in.getParameterTypes(), in.getExceptionTypes(),
                  null, invocation);
         invokeNext.setModifiers(in.getModifiers());

         String code = TransformerCommon.STANDARD_INTERCEPTOR_CODE;

         code += "return new " + con.getDeclaringClass().getName() + "(";
         for (int i = 0 ; i < params.length ; i++)
         {
            if (i > 0) code += ", ";
            code += "arg" + i;
         }
         code += ");  ";
         code += "}";
         try
         {
            invokeNext.setBody(code);
         }
         catch (CannotCompileException e)
         {
            System.out.println(code);
            throw e;
         }
         invocation.addMethod(invokeNext);
         TransformerCommon.addGetArguments(pool, invocation, con.getParameterTypes());

         ////////////
         //Create copy() method
         CtMethod copyTemplate = conInvocation.getDeclaredMethod("copy");

         CtMethod copy = CtNewMethod.make(copyTemplate.getReturnType(), "copy",
                  copyTemplate.getParameterTypes(), copyTemplate
                           .getExceptionTypes(), null, invocation);
         copy.setModifiers(copyTemplate.getModifiers());

         String copyCode = "{ "
                  + "   "
                  + invocation.getName()
                  + " wrapper = new "
                  + invocation.getName()
                  + "(this.interceptors, this.info, this.callingMethodHash, this.callingClass, this.arguments);"
                  + "   wrapper.metadata = this.metadata; "
                  + "   wrapper.currentInterceptor = this.currentInterceptor; "
                  + "   wrapper.instanceResolver = this.instanceResolver; "
                  + "   wrapper.targetObject = this.targetObject; "
                  + "   wrapper.advisor = this.advisor; "
                  + "   wrapper.responseContextInfo = this.responseContextInfo; ";      
         
         for (int i = 0 ; i < params.length ; i++)
         {
            copyCode += "   wrapper.arg" + i + " = this.arg" + i + "; ";
         }
         copyCode += "   return wrapper; }";

         try
         {
            copy.setBody(copyCode);
         }
         catch (CannotCompileException e)
         {
            System.out.println(copyCode);
            throw e;
         }
         invocation.addMethod(copy);

         TransformerCommon.compileOrLoadClass(callingClass, invocation);
         
         //Return fully qualified name of class (may be an inner class)
         return invocation.getName();
      }

      private String createOptimizedMethodCalledByConInvocationClass(CtMethod method, int callingIndex, long calledHash) throws Exception
      {
         AOPClassPool pool = (AOPClassPool) instrumentor.getClassPool();
         CtClass methodInvocation = pool.get("org.jboss.aop.joinpoint.MethodCalledByConstructorInvocation");

         ////////////////
         //Create the class
         String className = getOptimizedMethodCalledByConstructorClassName(callingIndex, callingClass.getName(), calledHash);
         boolean makeInnerClass = Modifier.isPrivate(method.getModifiers());
         CtClass invocation = TransformerCommon.makeInvocationClass(pool, makeInnerClass, callingClass, className, methodInvocation); 

         CtClass[] params = method.getParameterTypes();
         TransformerCommon.addArgumentFieldsToInvocation(invocation, params);

         /////////
         //Create invokeNext() body
         boolean isStatic = javassist.Modifier.isStatic(method.getModifiers());
         if (!isStatic)
         {
            CtField target = new CtField(method.getDeclaringClass(), "typedTargetObject", invocation);
            target.setModifiers(Modifier.PUBLIC);
            invocation.addField(target);
         }

         
         CtMethod in = methodInvocation.getDeclaredMethod("invokeNext");
         CtMethod invokeNext = CtNewMethod.make(in.getReturnType(), "invokeNext", in.getParameterTypes(), in.getExceptionTypes(), null, invocation);
         invokeNext.setModifiers(in.getModifiers());

         String code = TransformerCommon.STANDARD_INTERCEPTOR_CODE;

         String returnStr = (method.getReturnType().equals(CtClass.voidType)) ? "" : "return ($w)";
         if (isStatic)
         {
            code +=
                    "   " + returnStr + " " + method.getDeclaringClass().getName() + ".";
         }
         else
         {
            code +=
                    "   " + returnStr + " typedTargetObject.";
         }
         code += method.getName() + "(";
         for (int i = 0; i < params.length; i++)
         {
            if (i > 0) code += ", ";
            code += "arg" + i;
         }
         code += ");  ";
         if (method.getReturnType().equals(CtClass.voidType))
         {
            code += " return null; ";
         }
         code += "}";
         try
         {
            invokeNext.setBody(code);
         }
         catch (CannotCompileException e)
         {
            System.out.println(code);
            throw e;
         }
         invocation.addMethod(invokeNext);
         TransformerCommon.addGetArguments(pool, invocation, method.getParameterTypes());

         ////////////
         //Create copy() method
         CtMethod copyTemplate = methodInvocation.getDeclaredMethod("copy");

         CtMethod copy = CtNewMethod.make(copyTemplate.getReturnType(), "copy",
                  copyTemplate.getParameterTypes(), copyTemplate
                           .getExceptionTypes(), null, invocation);
         copy.setModifiers(copyTemplate.getModifiers());

         String copyCode = "{ "
                  + "   "
                  + invocation.getName()
                  + " wrapper = new "
                  + invocation.getName()
                  + "(this.calledMethod, this.arguments, this.calling, this.interceptors);"
                  + "   wrapper.metadata = this.metadata; "
                  + "   wrapper.currentInterceptor = this.currentInterceptor; "
                  + "   wrapper.instanceResolver = this.instanceResolver; "
                  + "   wrapper.targetObject = this.targetObject; "
                  + "   wrapper.advisor = this.advisor; "
                  + "   wrapper.responseContextInfo = this.responseContextInfo; ";
         
         if (!isStatic)
         {
            copyCode += "wrapper.typedTargetObject = typedTargetObject;";
         }
         
         for (int i = 0 ; i < params.length ; i++)
         {
            copyCode += "   wrapper.arg" + i + " = this.arg" + i + "; ";
         }
         copyCode += "   return wrapper; }";
         
         try
         {
            copy.setBody(copyCode);
         }
         catch (CannotCompileException e)
         {
            System.out.println(copyCode);
            throw e;
         }
         invocation.addMethod(copy);

         TransformerCommon.compileOrLoadClass(callingClass, invocation);


         //Return fully qualified name of class (may be an inner class)
         return invocation.getName();      
      }

      private String createOptimizedMethodCalledByMethodInvocationClass(CtMethod method, long callingHash, long calledHash) throws Exception
      {
         AOPClassPool pool = (AOPClassPool) instrumentor.getClassPool();
         CtClass methodInvocation = pool.get("org.jboss.aop.joinpoint.MethodCalledByMethodInvocation");

         ////////////////
         //Create the class
         String className = getOptimizedMethodCalledByMethodClassName(callingHash, callingClass.getName(), calledHash);
         boolean makeInnerClass = Modifier.isPrivate(method.getModifiers());
         CtClass invocation = TransformerCommon.makeInvocationClass(pool, makeInnerClass, callingClass, className, methodInvocation); 

         CtClass[] params = method.getParameterTypes();
         TransformerCommon.addArgumentFieldsToInvocation(invocation, params);

         /////////
         //Create invokeNext() body
         boolean isStatic = javassist.Modifier.isStatic(method.getModifiers());
         if (!isStatic)
         {
            CtField target = new CtField(method.getDeclaringClass(), "typedTargetObject", invocation);
            target.setModifiers(Modifier.PUBLIC);
            invocation.addField(target);
         }

         
         CtMethod in = methodInvocation.getDeclaredMethod("invokeNext");
         CtMethod invokeNext = CtNewMethod.make(in.getReturnType(), "invokeNext", in.getParameterTypes(), in.getExceptionTypes(), null, invocation);
         invokeNext.setModifiers(in.getModifiers());

         String code = TransformerCommon.STANDARD_INTERCEPTOR_CODE;

         String returnStr = (method.getReturnType().equals(CtClass.voidType)) ? "" : "return ($w)";
         if (isStatic)
         {
            code +=
                    "   " + returnStr + " " + method.getDeclaringClass().getName() + ".";
         }
         else
         {
            code +=
                    "   " + returnStr + " typedTargetObject.";
         }
         code += method.getName() + "(";
         for (int i = 0; i < params.length; i++)
         {
            if (i > 0) code += ", ";
            code += "arg" + i;
         }
         code += ");  ";
         if (method.getReturnType().equals(CtClass.voidType))
         {
            code += " return null; ";
         }
         code += "}";
         try
         {
            invokeNext.setBody(code);
         }
         catch (CannotCompileException e)
         {
            System.out.println(code);
            throw e;
         }
         invocation.addMethod(invokeNext);
         TransformerCommon.addGetArguments(pool, invocation, method.getParameterTypes());

         ////////////
         //Create copy() method
         CtMethod copyTemplate = methodInvocation.getDeclaredMethod("copy");

         CtMethod copy = CtNewMethod.make(copyTemplate.getReturnType(), "copy",
                  copyTemplate.getParameterTypes(), copyTemplate
                           .getExceptionTypes(), null, invocation);
         copy.setModifiers(copyTemplate.getModifiers());

         String copyCode = "{ "
                  + "   "
                  + invocation.getName()
                  + " wrapper = new "
                  + invocation.getName()
                  + "(this.interceptors, this.calledMethod, this.arguments, this.callingClass, this.callingMethodHash);"
                  + "   wrapper.metadata = this.metadata; "
                  + "   wrapper.currentInterceptor = this.currentInterceptor; "
                  + "   wrapper.instanceResolver = this.instanceResolver; "
                  + "   wrapper.targetObject = this.targetObject; "
                  + "   wrapper.advisor = this.advisor; "
                  + "   wrapper.callingMethod = this.callingMethod; "
                  + "   wrapper.responseContextInfo = this.responseContextInfo; ";
         
         if (!isStatic)
         {
            copyCode += "wrapper.typedTargetObject = typedTargetObject;";
         }
         
         for (int i = 0 ; i < params.length ; i++)
         {
            copyCode += "   wrapper.arg" + i + " = this.arg" + i + "; ";
         }
         copyCode += "   return wrapper; }";
         
         try
         {
            copy.setBody(copyCode);
         }
         catch (CannotCompileException e)
         {
            System.out.println(copyCode);
            throw e;
         }
         invocation.addMethod(copy);

         TransformerCommon.compileOrLoadClass(callingClass, invocation);

         //Return fully qualified name of class (may be an inner class)
         return invocation.getName();      
      }
   }
}
