/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.aop.instrument;

import javassist.CannotCompileException;
import javassist.ClassPool;
import javassist.CtClass;
import javassist.CtConstructor;
import javassist.CtField;
import javassist.CtMethod;
import javassist.CtNewConstructor;
import javassist.CtNewMethod;
import javassist.Modifier;
import javassist.NotFoundException;
import org.jboss.aop.AOPClassPool;
import org.jboss.aop.AspectManager;
import org.jboss.aop.ClassAdvisor;
import org.jboss.aop.pointcut.Pointcut;
import org.jboss.aop.standalone.Compiler;

import java.io.File;
import java.io.FileOutputStream;
import java.net.URI;
import java.net.URL;
import java.util.Iterator;
import java.util.List;

/**
 * Comment
 *
 * @author <a href="mailto:bill@jboss.org">Bill Burke</a>
 * @version $Revision: 1.9.2.1 $
 */
public class ConstructorExecutionTransformer
{
   private Instrumentor instrumentor;

   public ConstructorExecutionTransformer(Instrumentor instrumentor)
   {
      this.instrumentor = instrumentor;
   }

   public static String constructorFactory(String className)
   {
      return className + "_new_" + ClassAdvisor.NOT_TRANSFORMABLE_SUFFIX;
   }

   public boolean replaceConstructorAccess(ClassAdvisor sourceAdvisor, CtClass source) throws NotFoundException
   {
      if (!isAdvisableConstructor(source, sourceAdvisor)) return false;
      instrumentor.converter.replaceNew(source, source, constructorFactory(source.getSimpleName()));
      return true;
   }

   public void buildConstructorWrappers(CtClass clazz, ClassAdvisor advisor)
           throws Exception
   {
      // NOTE: we have to do this for every constructor
      // because of the way CodeConverter works.
      if (!isAdvisableConstructor(clazz, advisor)) return;
      instrumentor.setupBasics(clazz);
      List constructors = instrumentor.getConstructors(clazz);

      Iterator it = constructors.iterator();
      for (int index = 0; it.hasNext(); index++)
      {

         CtConstructor constructor = (CtConstructor) it.next();
         int mod = Modifier.STATIC;
         if ((constructor.getModifiers() & Modifier.PUBLIC) != 0)
         {
            mod |= Modifier.PUBLIC;
         }
         else if ((constructor.getModifiers() & Modifier.PROTECTED) != 0)
         {
            mod |= Modifier.PROTECTED;
         }
         else if ((constructor.getModifiers() & Modifier.PRIVATE) != 0)
         {
            mod |= Modifier.PRIVATE;
         }
         else
         {
            mod |= Modifier.PUBLIC;
         }

         CtClass[] exceptions = constructor.getExceptionTypes();
         String name = clazz.getSimpleName();
         CtClass type = constructor.getDeclaringClass();

         if (AspectManager.optimize)
         {
            createOptimizedWrapper(type, name, constructor, exceptions, clazz, mod, advisor, index);
         }
         else
         {
            createWrapper(type, name, constructor, exceptions, clazz, mod, advisor, index);

         }

      }
   }

   private void createWrapper(CtClass type, String name, CtConstructor constructor, CtClass[] exceptions, CtClass clazz, int mod, ClassAdvisor advisor, int index) throws CannotCompileException, NotFoundException
   {
      CtMethod wmethod = CtNewMethod.make(type, constructorFactory(name), constructor.getParameterTypes(), exceptions, null, clazz);
      wmethod.setModifiers(mod);

      String code = null;
      if (isAdvisableConstructor(constructor, advisor))
      {
         String args = "(Object[])null";
         if (constructor.getParameterTypes().length > 0) args = "$args";
         code =
                 "{ " +
                 "    if (" + Instrumentor.HELPER_FIELD_NAME + ".doesHaveAspects) " +
                 "    { " +
                 "       return ($r)" + Instrumentor.HELPER_FIELD_NAME + ".invokeNew(" + args + ", (int)" + (index) + "); " +
                 "    } " +
                 "    return new " + clazz.getName() + "($$); " +
                 "}";
      }
      else
      {
         code =
                 "{ " +
                 "    return new " + clazz.getName() + "($$); " +
                 "}";
      }

      wmethod.setBody(code);
      clazz.addMethod(wmethod);
   }

   private void createOptimizedWrapper(CtClass type, String name, CtConstructor constructor, CtClass[] exceptions, CtClass clazz, int mod, ClassAdvisor advisor, int index) throws Exception
   {
      String invocationClass = createOptimizedInvocationClass(type, constructor, index);
      CtMethod wmethod = CtNewMethod.make(type, constructorFactory(name), constructor.getParameterTypes(), exceptions, null, clazz);
      wmethod.setModifiers(mod);

      String code = null;
      if (isAdvisableConstructor(constructor, advisor))
      {
         code =
                 "{ " +
                 "    if (" + Instrumentor.HELPER_FIELD_NAME + ".doesHaveAspects) " +
                 "    { " +
                 "       org.jboss.aop.advice.Interceptor[] interceptors = " + Instrumentor.HELPER_FIELD_NAME + ".getConstructorInterceptors()[" + index + "]; " +
                 "       " + invocationClass + " invocation = new " + invocationClass + "(interceptors); " +
                 "       invocation.setConstructor(" + Instrumentor.HELPER_FIELD_NAME + ".getConstructors()[" + index + "]); " +
                 MethodExecutionTransformer.setArguments(constructor.getParameterTypes()) +
                 "       invocation.setAdvisor(" + Instrumentor.HELPER_FIELD_NAME + "); " +
                 "       return ($r)invocation.invokeNext(); " +
                 "    } " +
                 "    return new " + clazz.getName() + "($$); " +
                 "}";
      }
      else
      {
         code =
                 "{ " +
                 "    return new " + clazz.getName() + "($$); " +
                 "}";
      }

      try
      {
         wmethod.setBody(code);
      }
      catch (CannotCompileException e)
      {
         System.out.println(code);
         throw new RuntimeException(e);  //To change body of catch statement use Options | File Templates.
      }
      clazz.addMethod(wmethod);
   }

   /**
    * returns true if there is at least one constructor that has an execution
    *
    * @param clazz
    * @param advisor
    * @return
    * @throws javassist.NotFoundException
    */
   public static boolean isAdvisableConstructor(CtClass clazz, ClassAdvisor advisor) throws NotFoundException
   {
      CtConstructor[] constructors = clazz.getDeclaredConstructors();
      for (int i = 0; i < constructors.length; i++)
      {
         if (isAdvisableConstructor(constructors[i], advisor)) return true;
      }
      return false;
   }

   public static boolean isAdvisableConstructor(CtConstructor con, ClassAdvisor advisor) throws NotFoundException
   {
      Iterator pointcuts = AspectManager.instance().getPointcuts().values().iterator();
      while (pointcuts.hasNext())
      {
         Pointcut pointcut = (Pointcut) pointcuts.next();
         if (pointcut.matchesExecution(advisor, con))
         {
            return true;
         }
      }
      return false;
   }

   protected String createOptimizedInvocationClass(CtClass clazz, CtConstructor con, int index) throws Exception
   {
      AOPClassPool pool = (AOPClassPool) instrumentor.getClassPool();
      CtClass conInvocation = pool.get("org.jboss.aop.joinpoint.ConstructorInvocation");
      CtClass untransformable = pool.get("org.jboss.aop.instrument.Untransformable");

      String className = con.getDeclaringClass().getName() + index + "OptimizedConstructorInvocation";
      boolean makeInnerClass = !Modifier.isPublic(con.getModifiers());

      CtClass invocation;
      if (makeInnerClass)
      {
         //Strip away the package from classname
         String innerClassName = className.substring(className.lastIndexOf('.') + 1);
         
         //Only static nested classes are supported
         boolean classStatic = true;
         invocation = clazz.makeNestedClass(innerClassName, classStatic);
         invocation.setSuperclass(conInvocation);
      }
      else
      {
         invocation = pool.makeClass(className, conInvocation);
      }

      invocation.addInterface(untransformable);
      CtConstructor template = conInvocation.getDeclaredConstructors()[0];
      CtConstructor icon = CtNewConstructor.make(template.getParameterTypes(), template.getExceptionTypes(), invocation);
      invocation.addConstructor(icon);

      CtClass[] params = con.getParameterTypes();
      for (int i = 0; i < params.length; i++)
      {
         CtField field = new CtField(params[i], "arg" + i, invocation);
         field.setModifiers(Modifier.PUBLIC);
         invocation.addField(field);
      }

      CtMethod in = conInvocation.getDeclaredMethod("invokeNext");
      CtMethod invokeNext = CtNewMethod.make(in.getReturnType(), "invokeNext", in.getParameterTypes(), in.getExceptionTypes(), null, invocation);
      invokeNext.setModifiers(in.getModifiers());

      String code =
              "{ " +
              "   if (currentInterceptor < interceptors.length) " +
              "   { " +
              "      try   " +
              "      { " +
              "         return interceptors[currentInterceptor++].invoke(this); " +
              "      }  " +
              "      catch (Throwable t)  " +
              "      {  " +
              "         currentInterceptor--;  " +
              "         throw t; " +
              "      } " +
              "   } ";

      code += "return new " + con.getDeclaringClass().getName() + "(";
      for (int i = 0; i < params.length; i++)
      {
         if (i > 0) code += ", ";
         code += "arg" + i;
      }
      code += ");  ";
      code += "}";
      try
      {
         invokeNext.setBody(code);
      }
      catch (CannotCompileException e)
      {
         System.out.println(code);
         throw e;
      }
      invocation.addMethod(invokeNext);
      TransformerCommon.addGetArguments(pool, invocation, con.getParameterTypes());
      addCopy(pool, invocation, con.getParameterTypes());
      // If compile time
      if (Compiler.loader != null)
      {
         File file;
         URL url = Compiler.loader.getResource(con.getDeclaringClass().getName().replace('.', '/') + ".class");
         String path = url.toString();
         path = path.substring(0, path.lastIndexOf('/') + 1);
         path = path + invocation.getSimpleName() + ".class";
         URI newUrl = new URI(path);
         file = new File(newUrl);
         FileOutputStream fp = new FileOutputStream(file);
         fp.write(invocation.toBytecode());
         fp.close();
      }
      else // if load time
      {
         invocation.toClass();
      }

      //Return fully qualified name of class (may be an inner class)
      return invocation.getName();
   }

   private void addCopy(ClassPool pool, CtClass invocation, CtClass[] params) throws Exception
   {
      CtClass methodInvocation = pool.get("org.jboss.aop.joinpoint.ConstructorInvocation");
      CtMethod template = methodInvocation.getDeclaredMethod("copy");

      CtMethod copy = CtNewMethod.make(template.getReturnType(), "copy", template.getParameterTypes(), template.getExceptionTypes(), null, invocation);
      copy.setModifiers(template.getModifiers());

      String code =
              "{ " +
              "   " + invocation.getName() + " wrapper = new " + invocation.getName() + "(this.interceptors); " +
              "   wrapper.constructor = this.constructor; " +
              "   wrapper.arguments = this.arguments; " +
              "   wrapper.metadata = this.metadata; " +
              "   wrapper.currentInterceptor = this.currentInterceptor; ";
      for (int i = 0; i < params.length; i++)
      {
         code += "   wrapper.arg" + i + " = this.arg" + i + "; ";
      }
      code += "   return wrapper; }";

      copy.setBody(code);
      invocation.addMethod(copy);

   }

}
