/* *************************************
 *                                     *
 *  JBoss: The OpenSource J2EE WebOS   *
 *                                     *
 *  Distributable under LGPL license.  *
 *  See terms of license at gnu.org.   *
 *                                     *
 ************************************* */
package org.jboss.aop.instrument;

import javassist.CannotCompileException;
import javassist.ClassPool;
import javassist.CodeConverter;
import javassist.CtClass;
import javassist.CtField;
import javassist.CtMethod;
import javassist.CtNewMethod;
import javassist.Modifier;
import javassist.NotFoundException;
import javassist.expr.ExprEditor;
import javassist.expr.FieldAccess;
import org.jboss.aop.AOPClassPool;
import org.jboss.aop.AspectManager;
import org.jboss.aop.ClassAdvisor;
import org.jboss.aop.pointcut.Pointcut;

import java.util.Iterator;
import java.util.List;

/**
 * @author <a href="mailto:kabirkhan@bigfoot.com">Kabir Khan</a>
 *
 */
public class FieldAccessTransformer
{
   // Constants -----------------------------------------------------
   
   // Attributes ----------------------------------------------------
   Instrumentor instrumentor;
   boolean optimize;
   NotOptimizedTransformer notOptimizedTransformer;
   OptimizedTransformer optimizedTransformer;
      
   // Static --------------------------------------------------------
       
   // Constructors --------------------------------------------------
   public FieldAccessTransformer(Instrumentor instrumentor)
   {
      this.instrumentor = instrumentor;
      this.optimize = AspectManager.optimize;
      if (optimize)
      {
         optimizedTransformer = new OptimizedTransformer();
      }
      else
      {
         notOptimizedTransformer = new NotOptimizedTransformer();
      }
   }

   // Public --------------------------------------------------------
   public void buildFieldWrappers(CtClass clazz, ClassAdvisor advisor) throws NotFoundException, CannotCompileException
   {
      if (optimize)
      {
         optimizedTransformer.buildFieldWrappers(clazz, advisor);
      }
      else
      {
         notOptimizedTransformer.buildFieldWrappers(clazz, advisor);
      }
   }

   public boolean replaceFieldAccess(List fields, CtClass clazz, ClassAdvisor fieldsAdvisor) throws NotFoundException
   {
      CodeConverter converter = instrumentor.getCodeConverter();
      boolean converted = false;
      Iterator it = fields.iterator();
      while (it.hasNext())
      {
         CtField field = (CtField) it.next();
         if (isAdvisableGetField(field, fieldsAdvisor))
         {
            converted = true;
            converter.replaceFieldRead(field, clazz, fieldRead(field.getName()));
         }
         if (isAdvisableSetField(field, fieldsAdvisor))
         {
            converted = true;
            converter.replaceFieldWrite(field, clazz, fieldWrite(field.getName()));
         }
      }
      return converted;
   }


   
   // Z implementation ----------------------------------------------
   
   // Y overrides ---------------------------------------------------
   
   // Package protected ---------------------------------------------
   
   // Protected -----------------------------------------------------

   // Private -------------------------------------------------------
   private int fieldOffset(CtClass clazz) throws NotFoundException
   {
      if (clazz == null)
         return 0;
      if (clazz.getName().equals("java.lang.Object"))
         return 0;
      int offset = fieldOffset(clazz.getSuperclass());

      CtField[] fields = clazz.getDeclaredFields();
      for (int i = 0; i < fields.length; i++)
      {
         if (instrumentor.isAdvisable(fields[i]))
         {
            offset++;
         }
      }
      return offset;
   }

   private boolean isAdvisableGetField(CtField field, ClassAdvisor advisor) throws NotFoundException
   {
      Iterator pointcuts = AspectManager.instance().getPointcuts().values().iterator();
      while (pointcuts.hasNext())
      {
         Pointcut pointcut = (Pointcut) pointcuts.next();
         if (pointcut.matchesGet(advisor, field))
            return true;
      }
      return false;
   }

   private boolean isAdvisableSetField(CtField field, ClassAdvisor advisor) throws NotFoundException
   {
      Iterator pointcuts = AspectManager.instance().getPointcuts().values().iterator();
      while (pointcuts.hasNext())
      {
         Pointcut pointcut = (Pointcut) pointcuts.next();
         if (pointcut.matchesSet(advisor, field))
            return true;
      }
      return false;
   }

   private String fieldRead(String fieldName)
   {
      return fieldName + "_r_" + ClassAdvisor.NOT_TRANSFORMABLE_SUFFIX;
   }

   private String fieldWrite(String fieldName)
   {
      return fieldName + "_w_" + ClassAdvisor.NOT_TRANSFORMABLE_SUFFIX;
   }

   private int getModifiers(CtField field)
   {
      int mod = Modifier.STATIC;
      if ((field.getModifiers() & Modifier.PUBLIC) != 0)
      {
         mod |= Modifier.PUBLIC;
      }
      else if ((field.getModifiers() & Modifier.PROTECTED) != 0)
      {
         mod |= Modifier.PROTECTED;
      }
      else if ((field.getModifiers() & Modifier.PRIVATE) != 0)
      {
         mod |= Modifier.PRIVATE;
      }
      else
      {
         mod |= Modifier.PUBLIC;
      }

      return mod;
   }

   /**
    * This function replaces internal field accesses with bytecode hooks into framework
    * todo this must do it for inherited protected/public fields as well
    *
    * @param clazz
    * @param field
    * @param doGet
    * @param doSet
    * @param index
    * @throws javassist.CannotCompileException
    */
   private void replaceFieldAccessInternally(CtClass clazz, CtField field, boolean doGet, boolean doSet, int index) throws CannotCompileException
   {
      FieldAccessExprEditor expr = new FieldAccessExprEditor(clazz, field, doGet, doSet, index);
      clazz.instrument(expr);
   }
   // Inner classes -------------------------------------------------
      
   class OptimizedTransformer
   {

      void buildFieldWrappers(CtClass clazz, ClassAdvisor advisor) throws NotFoundException, CannotCompileException
      {
         List fields = instrumentor.getAdvisableFields(clazz);

         int offset = fieldOffset(clazz.getSuperclass());

         Iterator it = fields.iterator();
         for (int index = offset; it.hasNext(); index++)
         {
            CtField field = (CtField) it.next();

            boolean doGet = isAdvisableGetField(field, advisor);
            boolean doSet = isAdvisableSetField(field, advisor);
            if (!doGet && !doSet)
            {
               continue;
            }

            instrumentor.setupBasics(clazz);
            int mod = getModifiers(field);
                        
            //Create placeholder static wrappers, since without these methods replaceFieldAccessInternally() 
            //will not compile. 
            //If we add the actual static wrappers before calling replaceFieldAccessInternally()
            //field access done in the inner invocation classes as well as in the static wrappers
            //is replaced with a call to the wrapper instead, which means infinite recursion
            buildOptimizedWrapperPlaceHolders(clazz, field, doGet, doSet, mod);              
            replaceFieldAccessInternally(clazz, field, doGet, doSet, index);
            buildOptimizedWrappers(clazz, field, doGet, doSet, index);
         }      
      }

      private void buildOptimizedWrapperPlaceHolders(CtClass clazz, CtField field, boolean doGet, boolean doSet, int mod)
         throws NotFoundException, CannotCompileException
      {
         if (doGet)
         {
            buildOptimizedReadWrapperPlaceHolder(clazz, field, mod);
         }
         if (doSet)
         {
            buildOptimizedWriteWrapperPlaceHolder(clazz, field, mod);
         }
      }

      private void buildOptimizedReadWrapperPlaceHolder(CtClass clazz, CtField field, int mod)
         throws NotFoundException, CannotCompileException
      {
         AOPClassPool classPool = (AOPClassPool) instrumentor.getClassPool();

         String name = field.getName();
         CtClass ftype = field.getType();
         CtClass[] readParam = new CtClass[] { classPool.get("java.lang.Object")};
         
         //Doesn't matter what we put in as long as it compiles, 
         //body will be replaced when we call buildOptimizedReadWrapper()
         String code = "{" + ftype.getName() + " var; return var;}";

         try
         {
            CtMethod rmethod = CtNewMethod.make(ftype, fieldRead(name), readParam, null, null, clazz);
            rmethod.setModifiers(mod);
            rmethod.setBody(code);
            clazz.addMethod(rmethod);
         }
         catch (CannotCompileException e)
         {
            System.out.println(code);
            System.out.println(e.getMessage());
            throw e;
         }

      }

      private void buildOptimizedWriteWrapperPlaceHolder(CtClass clazz, CtField field, int mod)
         throws NotFoundException, CannotCompileException
      {
         AOPClassPool classPool = (AOPClassPool) instrumentor.getClassPool();

         String name = field.getName();
         CtClass ftype = field.getType();

         // create field trapWrite memthod
         CtClass[] writeParam = new CtClass[2];
         writeParam[0] = classPool.get("java.lang.Object");
         writeParam[1] = ftype;

         //Doesn't matter what we put in as long as it compiles, 
         //body will be replaced when we call buildOptimizedWriteWrapper()
         String code = "{ System.out.print(\"\"); }";         

         try
         {
            CtMethod wmethod = CtNewMethod.make(CtClass.voidType, fieldWrite(name), writeParam, null, null, clazz);
            wmethod.setModifiers(mod);
            wmethod.setBody(code);
            clazz.addMethod(wmethod);
         }
         catch (CannotCompileException e)
         {
            System.out.println(code);
            System.out.println(e.getMessage());
            throw e;
         }
      }

      private String getOptimizedInvocationClassName(CtClass clazz, CtField field, boolean get) throws Exception
      {
         StringBuffer sb = new StringBuffer();
         sb.append(field.getModifiers());
         sb.append(field.getClass());
         sb.append(field.getDeclaringClass());
         sb.append(field.getFieldInfo().getAccessFlags());
         sb.append(field.getFieldInfo().getDescriptor());
         sb.append(field.getName());
         sb.append(field.getType());
         long hash = sb.hashCode();

         String className = clazz.getName() + "_" + field.getName() + "_" + hash;

         if (get)
         {
            className += "_OptimizedGetFieldInvocation";
         }
         else
         {
            className += "_OptimizedSetFieldInvocation";
         }

         className = className.replace('-', 'N');
         return className;
      }

      private void buildOptimizedWrappers(CtClass clazz, CtField field, boolean doGet, boolean doSet, int index)
         throws NotFoundException, CannotCompileException
      {
         if (doGet)
         {
            buildOptimizedReadWrapper(clazz, field, index);
         }
         if (doSet)
         {
            buildOptimizedWriteWrapper(clazz, field, index);
         }
      }

      private void buildOptimizedReadWrapper(CtClass clazz, CtField field, int index)
         throws NotFoundException, CannotCompileException
      {
         String wrappedName = field.getName();

         String optimizedInvocation;
         try
         {
            optimizedInvocation = createOptimizedInvocationClass(clazz, field, true);
         }
         catch (Exception e)
         {
            throw new CannotCompileException(e);
         }

         String name = field.getName();
         boolean isStatic = javassist.Modifier.isStatic(field.getModifiers());
         // create field trapWrite memthod

         String code;

         if (!isStatic)
         {
            code =
               "{ "
                  + "    org.jboss.aop.ClassInstanceAdvisor instAdv = (("   + clazz.getName()   + ")$1)._instanceAdvisor;"
                  + "    if (" + Instrumentor.HELPER_FIELD_NAME + ".doesHaveAspects || (instAdv != null && instAdv.hasInstanceAspects))"
                  + "    { "
                  + "       org.jboss.aop.advice.Interceptor[] interceptors = " + Instrumentor.HELPER_FIELD_NAME   + ".getFieldReadInterceptors()[" + index + "]; "
                  + "       if (instAdv != null) "
                  + "       { "
                  + "          interceptors = instAdv.getInterceptors("   + Instrumentor.HELPER_FIELD_NAME   + ".getFieldReadInterceptors()[" + index + "]); "
                  + "       } "
                  + "       "   + optimizedInvocation + " invocation = new "   + optimizedInvocation + "(" + Instrumentor.HELPER_FIELD_NAME + ".getAdvisedFields()[" + index + "]," + index + ", interceptors); "
                  + "       invocation.setTargetObject($1); "
                  + "       invocation.typedTargetObject = (" + clazz.getName() + ")$1; "
                  + "       invocation.setAdvisor(" + Instrumentor.HELPER_FIELD_NAME + "); "
                  + "       return ($r)invocation.invokeNext(); "
                  + "    } "
                  + "    else "
                  + "    {"
                  + "       return ((" + clazz.getName() + ")$1)." + wrappedName + ";"
                  + "    }"
                  + "}";
         }
         else
         {
            code =
               "{ "
                  + "    if (" + Instrumentor.HELPER_FIELD_NAME + ".doesHaveAspects) " 
                  + "    { " + "       org.jboss.aop.advice.Interceptor[] interceptors = " + Instrumentor.HELPER_FIELD_NAME + ".getFieldReadInterceptors()[" + index + "]; "
                  + "   "+ optimizedInvocation + " invocation = new " + optimizedInvocation + "(" + Instrumentor.HELPER_FIELD_NAME + ".getAdvisedFields()["   + index + "],"   + index + ", interceptors); "
                  + "       invocation.setTargetObject($1); "
                  + "       invocation.setAdvisor(" + Instrumentor.HELPER_FIELD_NAME + "); "
                  + "       return ($r)invocation.invokeNext(); "
                  + "    } "
                  + "    else "
                  + "    {"
                  + "       return " + clazz.getName() + "." + wrappedName + ";"
                  + "    }"
                  + "}";
         }

         try
         {
            //Find method we created in buildOptimizedReadWrapperPlaceHolder and set real body
            CtMethod rmethod = clazz.getDeclaredMethod(fieldRead(name)); 
            rmethod.setBody(code);
         }
         catch (CannotCompileException e)
         {
            System.out.println(code);
            System.out.println(e.getMessage());
            throw e;
         }

      }

      private void buildOptimizedWriteWrapper(CtClass clazz, CtField field, int index)
         throws NotFoundException, CannotCompileException
      {
         String wrappedName = field.getName();
         AOPClassPool classPool = (AOPClassPool) instrumentor.getClassPool();

         String optimizedInvocation;
         try
         {
            optimizedInvocation = createOptimizedInvocationClass(clazz, field, false);
         }
         catch (Exception e)
         {
            throw new CannotCompileException(e);
         }

         String name = field.getName();
         CtClass ftype = field.getType();
         boolean isStatic = javassist.Modifier.isStatic(field.getModifiers());
         // create field trapWrite memthod
         CtClass[] writeParam = new CtClass[2];
         writeParam[0] = classPool.get("java.lang.Object");
         writeParam[1] = ftype;

         String code;
         if (!isStatic)
         {
            code =
                         "{ "
                  + "    org.jboss.aop.ClassInstanceAdvisor instAdv = ((" + clazz.getName() + ")$1)._instanceAdvisor;"
                  + "    if (" + Instrumentor.HELPER_FIELD_NAME + ".doesHaveAspects || (instAdv != null && instAdv.hasInstanceAspects)) "
                  + "    { "
                  + "       org.jboss.aop.advice.Interceptor[] interceptors = " + Instrumentor.HELPER_FIELD_NAME + ".getFieldWriteInterceptors()[" + index + "]; "
                  + "       if (instAdv != null) "
                  + "       { "
                  + "          interceptors = instAdv.getInterceptors(" + Instrumentor.HELPER_FIELD_NAME + ".getFieldWriteInterceptors()[" + index + "]); "
                  + "       } "
                  + "       " + optimizedInvocation + " invocation = new " + optimizedInvocation + "(" + Instrumentor.HELPER_FIELD_NAME + ".getAdvisedFields()[" + index + "]," + index + ", ($w)$2" + ", interceptors); "
                  + "       invocation.setTargetObject($1); "
                  + "       invocation.typedTargetObject = (" + clazz.getName() + ")$1; "
                  + "       invocation.setAdvisor(" + Instrumentor.HELPER_FIELD_NAME + "); "
                  + "       invocation.invokeNext(); "
                  + "    } "
                  + "    else "
                  + "    {"
                  + "       ((" + clazz.getName() + ")$1)." + wrappedName + "=$2" + ";"
                  + "    }"
                  + "}";
         }
         else
         {
            code =
               "{ "
                  + "    if (" + Instrumentor.HELPER_FIELD_NAME + ".doesHaveAspects) "
                  + "    { "
                  + "       org.jboss.aop.advice.Interceptor[] interceptors = " + Instrumentor.HELPER_FIELD_NAME + ".getFieldWriteInterceptors()[" + index + "]; "
                  + "       " + optimizedInvocation + " invocation = new " + optimizedInvocation + "(" + Instrumentor.HELPER_FIELD_NAME + ".getAdvisedFields()[" + index + "]," + index + ", ($w)$2" + ", interceptors); "
                  + "       invocation.setTargetObject($1); "
                  + "       invocation.setAdvisor(" + Instrumentor.HELPER_FIELD_NAME + "); "
                  + "       invocation.invokeNext(); "
                  + "    } "
                  + "    else "
                  + "    {"
                  + "       " + clazz.getName() + "." + wrappedName + "=$2" + ";"
                  + "    }"
                  + "}";
         }

         try
         {
            //Find method we created in buildOptimizedReadWrapperPlaceHolder and set real body
            CtMethod rmethod = clazz.getDeclaredMethod(fieldWrite(name)); 
            rmethod.setBody(code);
         }
         catch (CannotCompileException e)
         {
            System.out.println(code);
            System.out.println(e.getMessage());
            throw e;
         }

      }

      private String createOptimizedInvocationClass(CtClass clazz, CtField field, boolean get) throws Exception
      {
         String wrappedName = field.getName();
         AOPClassPool pool = (AOPClassPool) instrumentor.getClassPool();
         CtClass fieldInvocation =
            get
               ? pool.get("org.jboss.aop.joinpoint.FieldReadInvocation")
               : pool.get("org.jboss.aop.joinpoint.FieldWriteInvocation");

         String className = getOptimizedInvocationClassName(clazz, field, get);
         boolean makeInnerClass = !Modifier.isPublic(field.getModifiers());

         CtClass invocation = TransformerCommon.makeInvocationClass(pool, makeInnerClass, clazz, className, fieldInvocation); 

         boolean isStatic = javassist.Modifier.isStatic(field.getModifiers());
         if (!isStatic)
         {
            CtField target = new CtField(field.getDeclaringClass(), "typedTargetObject", invocation);
            target.setModifiers(Modifier.PUBLIC);
            invocation.addField(target);
         }

         CtMethod in = fieldInvocation.getDeclaredMethod("invokeNext");
         CtMethod invokeNext =
            CtNewMethod.make(
               in.getReturnType(),
               "invokeNext",
               in.getParameterTypes(),
               in.getExceptionTypes(),
               null,
               invocation);
         invokeNext.setModifiers(in.getModifiers());

         String code = TransformerCommon.STANDARD_INTERCEPTOR_CODE;

         String ref = isStatic ? field.getDeclaringClass().getName() + "." : " typedTargetObject.";

         if (get)
         {
            code += "return ($w) " + ref + wrappedName + ";";
         }
         else
         {
            //TODO: Must be a better way to do the cast of the value that what is done by castInvocationValueToTypeString()?
            CtClass type = field.getType();
            code += ref + wrappedName + " = " + castInvocationValueToTypeString(type) + " return null;";
         }

         code += "}";
         try
         {
            invokeNext.setBody(code);
         }
         catch (CannotCompileException e)
         {
            System.out.println(code);
            throw e;
         }
         invocation.addMethod(invokeNext);
         addCopy(pool, invocation, isStatic, get);

         TransformerCommon.compileOrLoadClass(field.getDeclaringClass(), invocation);

         //Return fully qualified name of class (may be an inner class)
         return invocation.getName();
      }

      /* TODO: Should be a better way to do the cast of the invocation value performed, using something in javassist?
       */
      private String castInvocationValueToTypeString(CtClass type)
      {
         String cast = null;
         if (type.isPrimitive())
         {
            if (type.equals(CtClass.booleanType))
            {
               cast = "((Boolean)value).booleanValue();";
            }
            else if (type.equals(CtClass.byteType))
            {
               cast = "((Byte)value).byteValue()";
            }
            else if (type.equals(CtClass.charType))
            {
               cast = "((Character)value).charValue();";
            }
            else if (type.equals(CtClass.doubleType))
            {
               cast = "((Double)value).doubleValue();";
            }
            else if (type.equals(CtClass.floatType))
            {
               cast = "((Float)value).floatValue();";
            }
            else if (type.equals(CtClass.intType))
            {
               cast = "((Integer)value).intValue();";
            }
            else if (type.equals(CtClass.longType))
            {
               cast = "((Long)value).longValue();";
            }
            else if (type.equals(CtClass.shortType))
            {
               cast = "((Short)value).shortValue();";
            }
         }
         else if (type.isArray())
         {
            cast = "(" + type.getName() + "[])value;";
         }
         else
         {
            cast = "(" + type.getName() + ")value;";
         }

         return cast;
      }

      private void addCopy(ClassPool pool, CtClass invocation, boolean isStatic, boolean isGet) throws Exception
      {
         CtClass fieldInvocation =
            isGet
               ? pool.get("org.jboss.aop.joinpoint.FieldReadInvocation")
               : pool.get("org.jboss.aop.joinpoint.FieldWriteInvocation");
         CtMethod template = fieldInvocation.getDeclaredMethod("copy");

         CtMethod copy =
            CtNewMethod.make(
               template.getReturnType(),
               "copy",
               template.getParameterTypes(),
               template.getExceptionTypes(),
               null,
               invocation);
         copy.setModifiers(template.getModifiers());

         String newExpr =
            (isGet)
               ? invocation.getName()
                  + " wrapper = new "
                  + invocation.getName()
                  + "(this.field, this.index, this.interceptors); "
               : invocation.getName()
                  + " wrapper = new "
                  + invocation.getName()
                  + "(this.field, this.index, this.value, this.interceptors); ";

         String code =
            "{ "
               + "   "
               + newExpr
               + "   wrapper.metadata = this.metadata; "
               + "   wrapper.currentInterceptor = this.currentInterceptor; "
               + "   wrapper.instanceResolver = this.instanceResolver; ";
         if (!isStatic)
         {
            code += "   wrapper.typedTargetObject = this.typedTargetObject; ";
            code += "   wrapper.targetObject = this.targetObject; ";
         }

         code += "   return wrapper; }";

         copy.setBody(code);
         invocation.addMethod(copy);

      }
   }//End inner class OptimizedTransformer

   class NotOptimizedTransformer
   {
      public void buildFieldWrappers(CtClass clazz, ClassAdvisor advisor)
              throws NotFoundException, CannotCompileException
      {
         List fields = instrumentor.getAdvisableFields(clazz);
         ClassPool classPool = instrumentor.getClassPool();

         int offset = fieldOffset(clazz.getSuperclass());
         CtClass[] readParam = new CtClass[]{classPool.get("java.lang.Object")};
         Iterator it = fields.iterator();
         for (int index = offset; it.hasNext(); index++)
         {
            CtField field = (CtField) it.next();
            boolean doGet = isAdvisableGetField(field, advisor);
            boolean doSet = isAdvisableSetField(field, advisor);
            if (!doGet && !doSet)
            {
               continue;
            }


            instrumentor.setupBasics(clazz);

            replaceFieldAccessInternally(clazz, field, doGet, doSet, index);

            // don't need wrappers if the field is private as we inline
            // the conditional and interception directly within code.
            if (!javassist.Modifier.isPrivate(field.getModifiers()))
            {
               int mod = Modifier.STATIC;
               if ((field.getModifiers() & Modifier.PUBLIC) != 0)
               {
                  mod |= Modifier.PUBLIC;
               }
               else if ((field.getModifiers() & Modifier.PROTECTED) != 0)
               {
                  mod |= Modifier.PROTECTED;
               }
               else if ((field.getModifiers() & Modifier.PRIVATE) != 0)
               {
                  mod |= Modifier.PRIVATE;
               }
               else
               {
                  mod |= Modifier.PUBLIC;
               }

               String name = field.getName();
               CtClass ftype = field.getType();
               boolean isStatic = (javassist.Modifier.isStatic(field.getModifiers()));
               String access = "";
               String instanceCheck = "";
               if (!isStatic)
               {
                  access = "((" + clazz.getName() + ")$1).";
                  instanceCheck = " || ((org.jboss.aop.ClassInstanceAdvisor)((org.jboss.aop.InstanceAdvised)$1)._getInstanceAdvisor()).hasInstanceAspects";
               }

               // read wrapper
               if (doGet)
               {
                  CtMethod rmethod = CtNewMethod.make(ftype, fieldRead(name), readParam, null, null, clazz);
                  rmethod.setModifiers(mod);

                  String readcode =
                          "{ " +
                          "    if (" + Instrumentor.HELPER_FIELD_NAME + ".doesHaveAspects " + instanceCheck + " ) " +
                          "    { " +
                          "       return ($r)" + Instrumentor.HELPER_FIELD_NAME + ".invokeRead($1, (int)" + (index) + "); " +
                          "    } " +
                          "    return " + access + name + "; " +
                          "}";

                  rmethod.setBody(readcode);
                  clazz.addMethod(rmethod);
               }
               // create field trapWrite memthod
               if (doSet)
               {
                  CtClass[] writeParam = new CtClass[2];
                  writeParam[0] = classPool.get("java.lang.Object");
                  writeParam[1] = ftype;
                  String writecode =
                          "{ " +
                          "    if (" + Instrumentor.HELPER_FIELD_NAME + ".doesHaveAspects " + instanceCheck + " ) " +
                          "    { " +
                          "       " + Instrumentor.HELPER_FIELD_NAME + ".invokeWrite($1, (int)" + (index) + ", ($w)$2); " +
                          "    } " +
                          "    else " +
                          "    { " +
                          "       " + access + name + " = $2; " +
                          "    } " +
                          "}";
                  // read wrapper
                  CtMethod wmethod = CtNewMethod.make(CtClass.voidType, fieldWrite(name), writeParam, null, null, clazz);
                  wmethod.setModifiers(mod);
                  wmethod.setBody(writecode);
                  clazz.addMethod(wmethod);
               }
            }
         }
      }


   }// End Inner class NotOptimizedTransformer

   private class FieldAccessExprEditor extends ExprEditor
   {
      CtClass clazz;
      CtField field;
      boolean doGet;
      boolean doSet;
      int fieldIndex;

      public FieldAccessExprEditor(CtClass clazz, CtField field, boolean doGet, boolean doSet, int index)
      {
         this.clazz = clazz;
         this.field = field;
         this.doGet = doGet;
         this.doSet = doSet;
         this.fieldIndex = index;
      }

      public void edit(FieldAccess fieldAccess) throws CannotCompileException
      {
         if (!fieldAccess.getClassName().equals(clazz.getName())) return;
         if (!fieldAccess.getFieldName().equals(field.getName())) return;

         if (fieldAccess.isReader() && doGet)
         {
            replaceRead(fieldAccess);
         }
         if (fieldAccess.isWriter() && doSet)
         {
            replaceWrite(fieldAccess);
         }
      }

      private void replaceRead(FieldAccess fieldAccess) throws CannotCompileException
      {
         if (optimize)
         {
            replaceReadOptimized(fieldAccess);
         }
         else
         {
            replaceReadNotOptimized(fieldAccess);
         }
      }

      private void replaceReadOptimized(FieldAccess fieldAccess) throws CannotCompileException
      {
         if (fieldAccess.isStatic())
         {
            String code =
                    "    { " +
                    "       $_ = ($r)" + fieldRead(field.getName()) + "(null);" +
                    "    } " +
                    "";
            fieldAccess.replace(code);
         }
         else
         {
            String code =
                    "    { " +
                    "       $_ = ($r)" + fieldRead(field.getName()) + "($0);" +
                    "    } " +
                    "";
            fieldAccess.replace(code);
         }
      }

      private void replaceReadNotOptimized(FieldAccess fieldAccess) throws CannotCompileException
      {
         if (fieldAccess.isStatic())
         {
            String code =
                    "    if (" + Instrumentor.HELPER_FIELD_NAME + ".doesHaveAspects) " +
                    "    { " +
                    "       Object obj = null;" +
                    "       $_ = ($r)" + Instrumentor.HELPER_FIELD_NAME + ".invokeRead(obj, (int)" + (fieldIndex) + "); " +
                    "    } " +
                    "    else " +
                    "    { " +
                    "       $_ = " + clazz.getName() + "." + field.getName() + "; " +
                    "    } " +
                    "";
            fieldAccess.replace(code);
         }
         else
         {
            String code =
                    "    if (" + Instrumentor.HELPER_FIELD_NAME + ".doesHaveAspects || (_instanceAdvisor != null && _instanceAdvisor.hasInstanceAspects)) " +
                    "    { " +
                    "       $_ = ($r)" + Instrumentor.HELPER_FIELD_NAME + ".invokeRead($0, (int)" + (fieldIndex) + "); " +
                    "    } " +
                    "    else " +
                    "    { " +
                    "       $_ = this." + fieldAccess.getFieldName() + "; " +
                    "    } ";
            fieldAccess.replace(code);
         }
      }

      private void replaceWrite(FieldAccess fieldAccess) throws CannotCompileException
      {
         if (optimize)
         {
            replaceWriteOptimized(fieldAccess);
         }
         else
         {
            replaceWriteNotOptimized(fieldAccess);
         }
      }

      private void replaceWriteOptimized(FieldAccess fieldAccess) throws CannotCompileException
      {
         String fieldWrite = fieldWrite(field.getName());

         if (fieldAccess.isStatic())
         {
            String code =
                    "    { " +
                    "       " + fieldWrite + "(null, $1);" +
                    "    } " +
                    "";
            fieldAccess.replace(code);
            
         }
         else
         {
            String code =
                    "    { " +
                    "       " + fieldWrite + "($0, $1);" +
                    "    } " +
                    "";
            fieldAccess.replace(code);
         }
      }

      private void replaceWriteNotOptimized(FieldAccess fieldAccess) throws CannotCompileException
      {
         if (fieldAccess.isStatic())
         {
            String code =
                    "    if (" + Instrumentor.HELPER_FIELD_NAME + ".doesHaveAspects) " +
                    "    { " +
                    "       Object obj = null;" +
                    "      " + Instrumentor.HELPER_FIELD_NAME + ".invokeWrite(obj, (int)" + (fieldIndex) + ", ($w)$1); " +
                    "    } " +
                    "    else " +
                    "    { " +
                    "       " + clazz.getName() + "." + fieldAccess.getFieldName() + " = $1; " +
                    "    } ";
            fieldAccess.replace(code);
         }
         else
         {
            String code =
                    "    if (" + Instrumentor.HELPER_FIELD_NAME + ".doesHaveAspects || (_instanceAdvisor != null && _instanceAdvisor.hasInstanceAspects)) " +
                    "    { " +
                    "       " + Instrumentor.HELPER_FIELD_NAME + ".invokeWrite($0, (int)" + (fieldIndex) + ", ($w)$1); " +
                    "    } " +
                    "    else " +
                    "    { " +
                    "       this." + fieldAccess.getFieldName() + " = $1; " +
                    "    } ";
            fieldAccess.replace(code);
         }
      }


   }//End Inner class FieldAccessExprEditor


}
