/***************************************
 *                                     *
 *  JBoss: The OpenSource J2EE WebOS   *
 *                                     *
 *  Distributable under LGPL license.  *
 *  See terms of license at gnu.org.   *
 *                                     *
 ***************************************/

package org.jboss.aop.instrument;

import javassist.CannotCompileException;
import javassist.ClassPool;
import javassist.CodeConverter;
import javassist.CtClass;
import javassist.CtConstructor;
import javassist.CtField;
import javassist.CtMethod;
import javassist.CtNewMethod;
import javassist.Modifier;
import javassist.NotFoundException;
import javassist.SerialVersionUID;
import javassist.bytecode.AnnotationsAttribute;
import javassist.bytecode.FieldInfo;
import javassist.bytecode.MethodInfo;
import org.jboss.aop.AOPClassPool;
import org.jboss.aop.Advised;
import org.jboss.aop.Advisor;
import org.jboss.aop.AspectManager;
import org.jboss.aop.ClassAdvisor;
import org.jboss.aop.annotation.compiler.AnnotationInfoCreator;
import org.jboss.aop.introduction.AnnotationIntroduction;
import org.jboss.aop.introduction.InterfaceIntroduction;
import org.jboss.aop.util.CtConstructorComparator;
import org.jboss.aop.util.CtFieldComparator;
import org.jboss.aop.util.JavassistMethodHashing;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

/**
 * Transforms byte code, making a class advisable. Implements
 * command line class instrumentor as well. Reads classes from class path and creates
 * advised versions in specified directory. Usage:
 * <pre>
 * Instrumentor [dest. directory] [class[ class...]]
 * </pre>
 *
 * @author <a href="mailto:bill@jboss.org">Bill Burke</a>
 * @author <a href="mailto:gte863h@prism.gatech.edu">Austin Chau</a>
 * @author <a href="mailto:crazybob@crazybob.org">Bob Lee</a>
 * @version $Revision: 1.21.2.1 $
 */
public class Instrumentor
{
   /**
    * Package of AOP classes.
    */
   public static final String AOP_PACKAGE =
           Advised.class.getPackage().getName();

   /**
    * Name of helper class.
    */
   public static final String HELPER_CLASS_NAME =
           AOP_PACKAGE + ".ClassAdvisor";

   /**
    * Name of helper class.
    */
   public static final String ASPECT_MANAGER_CLASS_NAME =
           AOP_PACKAGE + ".AspectManager";

   /**
    * Helper class's field name.
    */
   public static final String HELPER_FIELD_NAME = "aop$classAdvisor" + ClassAdvisor.NOT_TRANSFORMABLE_SUFFIX;

   /**
    * Class reference's field name.
    */
   public static final String CLASS_FIELD_NAME = "aop$class" + ClassAdvisor.NOT_TRANSFORMABLE_SUFFIX;

   protected AOPClassPool classPool;
   protected boolean basicsSet = false;


   protected CodeConverter converter;


   // Transformers, more than meets the eye!
   MethodExecutionTransformer methodExecutionTransformer = new MethodExecutionTransformer(this);
   ConstructorExecutionTransformer constructorExecutionTransformer = new ConstructorExecutionTransformer(this);
   FieldAccessTransformer fieldAccessTransformer = new FieldAccessTransformer(this);
   CallerTransformer callerTransformer = new CallerTransformer(this);

   /**
    * Constructs new instrumentor.
    */
   public Instrumentor(AOPClassPool pool)
   {
      this.classPool = pool;
      this.converter = new CodeConverter();
   }

   public ClassPool getClassPool()
   {
      return classPool;
   }

   CodeConverter getCodeConverter()
   {
      return converter;
   }

   /**
    * Is the method advisable?
    */
   public static boolean isAdvisable(CtMethod method)
   {
      return
              (method.getModifiers() & Modifier.ABSTRACT) == 0 &&
              (method.getModifiers() & Modifier.NATIVE) == 0 &&
              !method.getName().endsWith(ClassAdvisor.NOT_TRANSFORMABLE_SUFFIX) &&
              !method.getName().equals("_getAdvisor") &&
              !method.getName().equals("_getInstanceAdvisor") &&
              !method.getName().equals("_setInstanceAdvisor") &&
              !method.getName().equals("class$") &&
              !method.getName().equals("");
   }

   /**
    * Is the method advisable?
    */
   public boolean isAdvisable(CtField field)
   {
      return (field.getModifiers() & Modifier.FINAL) == 0 &&
              (!field.getName().endsWith(ClassAdvisor.NOT_TRANSFORMABLE_SUFFIX)) &&
              !field.getName().equals("_instanceAdvisor") &&
              !field.getName().endsWith("$aop$mixin") &&
              (field.getName().indexOf('$') == -1)
              ;

   }

   public boolean isAdvised(CtClass clazz) throws NotFoundException
   {
      CtClass[] interfaces = clazz.getInterfaces();
      CtClass advised = forName(AOP_PACKAGE + ".Advised");
      for (int i = 0; i < interfaces.length; i++)
      {
         if (interfaces[i].equals(advised)) return true;
         if (interfaces[i].getName().equals(AOP_PACKAGE + ".Advised")) return true;
      }
      return false;
   }

   public boolean isTransformable(CtClass clazz) throws NotFoundException
   {
      CtClass[] interfaces = clazz.getInterfaces();
      CtClass advised = forName(AOP_PACKAGE + ".instrument.Untransformable");
      for (int i = 0; i < interfaces.length; i++)
      {
         if (interfaces[i].equals(advised)) return false;
         if (interfaces[i].getName().equals(AOP_PACKAGE + ".instrument.Untransformable")) return false;
      }
      return true;
   }

   private boolean isBaseClass(CtClass clazz)
           throws NotFoundException
   {
      if (clazz.getSuperclass() != null)
      {
         return !isAdvised(clazz.getSuperclass());
      }
      return true;
   }


   private String mixinFieldName(CtClass mixinClass)
   {
      StringBuffer buf = new StringBuffer("_");
      buf.append(mixinClass.getName().replace('.', '$'));
      buf.append("$aop$mixin");
      return buf.toString();
   }

   private void addMixinMethod(Advisor advisor, CtMethod method, CtClass clazz, CtMethod delegate, long hash) throws Exception
   {
      CtClass[] exceptions = method.getExceptionTypes();

      // create base, delegating method.
      CtMethod newMethod = CtNewMethod.wrapped(method.getReturnType(),
              method.getName(),
              method.getParameterTypes(),
              exceptions,
              delegate,
              CtMethod.ConstParameter.integer(hash),
              clazz);
      newMethod.setModifiers(Modifier.PUBLIC);
      clazz.addMethod(newMethod);
   }

   private void addMixin(CtClass clazz, InterfaceIntroduction pointcut, InterfaceIntroduction.Mixin mixin, HashMap baseMethods) throws Exception
   {
      // REVISIT:
      // Later on we should follow the same pattern as
      // C++ public virtual Mixins
      // But, for now, just throw an exception if the
      // mixin is adding any interfaces already
      // defined in base class or another mixin.
      CtClass mixinClass = classPool.get(mixin.getClassName());
      String initializer = (mixin.getConstruction() == null) ? ("new " + mixinClass.getName() + "()") : mixin.getConstruction();
      CtClass type = forName(mixinClass.getName());
      CtField field = new CtField(type, mixinFieldName(mixinClass), clazz);
      int modifiers = Modifier.PRIVATE;
      if (mixin.isTransient()) modifiers = modifiers | Modifier.TRANSIENT;
      field.setModifiers(modifiers);
      clazz.addField(field, CtField.Initializer.byExpr(initializer));
      HashSet addedMethods = new HashSet();

      String[] interfaces = mixin.getInterfaces();
      for (int i = 0; i < interfaces.length; i++)
      {
         CtClass intf = classPool.get(interfaces[i]);
         if (clazz.subtypeOf(intf)) continue;
         clazz.addInterface(intf);
         HashMap intfMap = JavassistMethodHashing.getMethodMap(intf);
         Iterator entries = intfMap.entrySet().iterator();
         while (entries.hasNext())
         {
            Map.Entry entry = (Map.Entry) entries.next();
            Long hash = (Long) entry.getKey();
            CtMethod method = (CtMethod) entry.getValue();
            if (baseMethods.containsKey(hash) && !addedMethods.contains(hash))
               throw new Exception("Mixin " + mixinClass.getName() +
                       " of pointcut " + pointcut.getName() +
                       " is trying to apply an already existing method" + method.getName() + " for class " + clazz.getName());

            // If another interface of this mixin has a duplicate method, then its ok, but don't re-add
            if (addedMethods.contains(hash)) continue;
            createMixinInvokeMethod(clazz, mixinClass, initializer, method, hash.longValue());
            baseMethods.put(hash, method);
            addedMethods.add(hash);
         }
      }
   }

   private void addIntroductionPointcutInterface(CtClass clazz, Advisor advisor, String intf, HashMap baseMethods) throws Exception
   {
      CtClass iface = classPool.get(intf);
      if (clazz.subtypeOf(iface)) return;
      if (clazz.subclassOf(iface)) return;

      clazz.addInterface(iface);

      CtMethod mixinInvokeMethod = createInvokeMethod(clazz);
      HashMap intfMap = JavassistMethodHashing.getMethodMap(iface);
      Iterator entries = intfMap.entrySet().iterator();
      while (entries.hasNext())
      {
         Map.Entry entry = (Map.Entry) entries.next();
         Long hash = (Long) entry.getKey();
         if (baseMethods.containsKey(hash)) continue;
         CtMethod method = (CtMethod) entry.getValue();
         addMixinMethod(advisor, method, clazz, mixinInvokeMethod, hash.longValue());
         baseMethods.put(hash, method);
      }
   }

   private void instrumentIntroductions(CtClass clazz, Advisor advisor)
           throws Exception
   {
      ArrayList pointcuts = advisor.getInterfaceIntroductions();
      if (pointcuts.size() == 0) return;
      HashMap baseMethods = JavassistMethodHashing.getMethodMap(clazz);
      Iterator it = pointcuts.iterator();
      if (it.hasNext()) setupBasics(clazz);
      while (it.hasNext())
      {

         InterfaceIntroduction pointcut = (InterfaceIntroduction) it.next();
         ArrayList mixins = pointcut.getMixins();
         for (int i = 0; i < mixins.size(); i++)
         {
            InterfaceIntroduction.Mixin mixin = (InterfaceIntroduction.Mixin) mixins.get(i);
            addMixin(clazz, pointcut, mixin, baseMethods);
         }
      }

      // pointcut interfaces.  If a method is already implemented for it then use that method
      // otherwise delegate to an interceptor
      it = pointcuts.iterator();
      while (it.hasNext())
      {
         InterfaceIntroduction pointcut = (InterfaceIntroduction) it.next();
         String[] interfaces = pointcut.getInterfaces();
         if (interfaces == null) continue;
         for (int i = 0; i < interfaces.length; i++)
         {
            addIntroductionPointcutInterface(clazz, advisor, interfaces[i], baseMethods);
         }
      }
   }

   private boolean instrumentAnnotationIntroductions(CtClass clazz, ClassAdvisor advisor)
           throws Exception
   {
      boolean changed = false;
      Iterator it = AspectManager.instance().getAnnotationIntoductions();
      while (it.hasNext())
      {
         AnnotationIntroduction introduction = (AnnotationIntroduction) it.next();
         if (AspectManager.verbose) System.out.println("**** " + introduction.getOriginalAnnotationExpr() + " invisible: " + introduction.isInvisible() + " expr: " + introduction.getOriginalExpression());
         if (introduction.matches(advisor, clazz))
         {
            if (AspectManager.verbose) System.out.println(introduction.getAnnotation() + " binds to " + clazz.getName());
            javassist.bytecode.annotation.Annotation info = AnnotationInfoCreator.createAnnotationInfo(classPool, clazz.getClassFile2().getConstPool(), introduction.getAnnotation());
            if (introduction.isInvisible())
            {
               AnnotationsAttribute invisible = (AnnotationsAttribute) clazz.getClassFile2().getAttribute(AnnotationsAttribute.invisibleTag);
               if (invisible == null)
               {
                  invisible = new AnnotationsAttribute(clazz.getClassFile2().getConstPool(), AnnotationsAttribute.invisibleTag);
                  clazz.getClassFile2().addAttribute(invisible);
               }
               changed = true;
               invisible.addAnnotation(info);
            }
            else
            {
               AnnotationsAttribute visible = (AnnotationsAttribute) clazz.getClassFile2().getAttribute(AnnotationsAttribute.visibleTag);
               if (visible == null)
               {
                  visible = new AnnotationsAttribute(clazz.getClassFile2().getConstPool(), AnnotationsAttribute.visibleTag);
                  clazz.getClassFile2().addAttribute(visible);
               }
               changed = true;
               visible.addAnnotation(info);
            }
         }

         CtMethod[] methods = clazz.getDeclaredMethods();
         for (int i = 0; i < methods.length; i++)
         {
            if (introduction.matches(advisor, methods[i]))
            {
               javassist.bytecode.annotation.Annotation info = AnnotationInfoCreator.createAnnotationInfo(classPool, methods[i].getMethodInfo2().getConstPool(), introduction.getAnnotation());
               MethodInfo mi = methods[i].getMethodInfo2();
               if (introduction.isInvisible())
               {
                  AnnotationsAttribute invisible = (AnnotationsAttribute) mi.getAttribute(AnnotationsAttribute.invisibleTag);
                  if (invisible == null)
                  {
                     invisible = new AnnotationsAttribute(mi.getConstPool(), AnnotationsAttribute.invisibleTag);
                     mi.addAttribute(invisible);
                  }
                  changed = true;
                  invisible.addAnnotation(info);
               }
               else
               {
                  AnnotationsAttribute visible = (AnnotationsAttribute) mi.getAttribute(AnnotationsAttribute.visibleTag);
                  if (visible == null)
                  {
                     visible = new AnnotationsAttribute(mi.getConstPool(), AnnotationsAttribute.visibleTag);
                     mi.addAttribute(visible);
                  }
                  changed = true;
                  visible.addAnnotation(info);
               }
            }

         }

         CtConstructor[] cons = clazz.getDeclaredConstructors();
         for (int i = 0; i < cons.length; i++)
         {
            if (introduction.matches(advisor, cons[i]))
            {
               javassist.bytecode.annotation.Annotation info = AnnotationInfoCreator.createAnnotationInfo(classPool, cons[i].getMethodInfo2().getConstPool(), introduction.getAnnotation());
               MethodInfo mi = cons[i].getMethodInfo2();
               if (introduction.isInvisible())
               {
                  AnnotationsAttribute invisible = (AnnotationsAttribute) mi.getAttribute(AnnotationsAttribute.invisibleTag);
                  if (invisible == null)
                  {
                     invisible = new AnnotationsAttribute(mi.getConstPool(), AnnotationsAttribute.invisibleTag);
                     mi.addAttribute(invisible);
                  }
                  changed = true;
                  invisible.addAnnotation(info);
               }
               else
               {
                  AnnotationsAttribute visible = (AnnotationsAttribute) mi.getAttribute(AnnotationsAttribute.visibleTag);
                  if (visible == null)
                  {
                     visible = new AnnotationsAttribute(mi.getConstPool(), AnnotationsAttribute.visibleTag);
                     mi.addAttribute(visible);
                  }
                  changed = true;
                  visible.addAnnotation(info);
               }
            }
         }

         CtField[] fields = clazz.getDeclaredFields();
         for (int i = 0; i < fields.length; i++)
         {
            if (introduction.matches(advisor, fields[i]))
            {
               javassist.bytecode.annotation.Annotation info = AnnotationInfoCreator.createAnnotationInfo(classPool, fields[i].getFieldInfo2().getConstPool(), introduction.getAnnotation());
               FieldInfo mi = fields[i].getFieldInfo2();
               if (introduction.isInvisible())
               {
                  AnnotationsAttribute invisible = (AnnotationsAttribute) mi.getAttribute(AnnotationsAttribute.invisibleTag);
                  if (invisible == null)
                  {
                     invisible = new AnnotationsAttribute(mi.getConstPool(), AnnotationsAttribute.invisibleTag);
                     mi.addAttribute(invisible);
                  }
                  changed = true;
                  invisible.addAnnotation(info);
               }
               else
               {
                  AnnotationsAttribute visible = (AnnotationsAttribute) mi.getAttribute(AnnotationsAttribute.visibleTag);
                  if (visible == null)
                  {
                     visible = new AnnotationsAttribute(mi.getConstPool(), AnnotationsAttribute.visibleTag);
                     mi.addAttribute(visible);
                  }
                  changed = true;
                  visible.addAnnotation(info);
               }
            }
         }
      }
      return changed;
   }

   public boolean applyCallerPointcuts(CtClass clazz, ClassAdvisor advisor) throws CannotCompileException
   {
      return callerTransformer.applyCallerPointcuts(clazz, advisor);
   }

   /**
    * Find all classes that this class references.  If any of those classes are advised and have field and/or constructor
    * interception, do instrumentation on this class so that those fields and constructors are instrumented
    */
   protected boolean convertReferences(CtClass clazz) throws Exception
   {
      boolean converted = false;
      String ref = null;
      try
      {
         AOPClassPool pool = AOPClassPool.createAOPClassPool(clazz.getClassPool(), AspectManager.instance());

         Iterator it = clazz.getRefClasses().iterator();
         while (it.hasNext())
         {
            ref = (String) it.next();
            if (ref.startsWith("java.")) continue;
            if (ref.startsWith("javax.")) continue;
            if (ref.startsWith("[")) continue;
            if (ref.equals(clazz.getName())) continue;
            // Only need a temporary advisor for resolving metadata
            CtClass ctRef = pool.get(ref);
            if (!isTransformable(ctRef)) continue;

            ClassAdvisor advisor = AspectManager.instance().getTempClassAdvisor(ctRef);
            List fields = getAdvisableFields(ctRef);
            if (fieldAccessTransformer.replaceFieldAccess(fields, ctRef, advisor)) converted = true;
            if (constructorExecutionTransformer.replaceConstructorAccess(advisor, ctRef)) converted = true;
            ref = null;
         }
      }
      catch (Exception ex)
      {
         if (ref != null)
         {
            throw new TransformationException("Failed to aspectize class " + clazz.getName() + ".  Could not find class it references " + ref);
         }
         throw ex;
      }
      return converted;
   }

   /**
    * Makes class advisable.
    */
   public boolean transform(CtClass clazz,
                            ClassAdvisor advisor)
   {
      try
      {
         if (clazz.isInterface() ||
                 clazz.isFrozen() ||
                 clazz.isArray() ||
                 clazz.getName().startsWith("org.jboss.aop") ||
                 isAdvised(clazz) ||
                 !isTransformable(clazz))
            return false;
         if (AspectManager.verbose) System.out.println("[trying to transform] " + clazz.getName());

         boolean converted = instrumentAnnotationIntroductions(clazz, advisor);
         boolean constructorAccessConverted = false;
         converted = applyCallerPointcuts(clazz, advisor) || converted;
         methodExecutionTransformer.instrument(clazz, advisor);
         constructorAccessConverted = constructorExecutionTransformer.replaceConstructorAccess(advisor, clazz);
         converted = converted || constructorAccessConverted;

         instrumentIntroductions(clazz, advisor);

         converted = convertReferences(clazz) || converted;

         // need to instrument no matter what so that
         // previously declared field and constructor interceptions
         // get instrumented within this class.
         if (converted || basicsSet)
         {
            clazz.instrument(converter);
         }

         // create static wrapper methods after
         // clazz.instrument because the wrappers may call cons or fields
         fieldAccessTransformer.buildFieldWrappers(clazz, advisor);
         if (constructorAccessConverted)
         {
            constructorExecutionTransformer.buildConstructorWrappers(clazz, advisor);
         }
         if (AspectManager.verbose) System.out.println("[debug] was " + clazz.getName() + " converted: " + (basicsSet || converted));

         if (basicsSet || converted)
         {
            return true;
         }
         else
         {
            //classPool.flushClass(clazz.getName());
            return false;
         }

      }
      catch (Exception e)
      {
         if (AspectManager.suppressTransformationErrors)
         {
            System.err.println("[warn] AOP Instrumentor failed to transform " + clazz.getName());
            if (AspectManager.verbose) e.printStackTrace();
            return false;
         }
         else
         {
            if (e instanceof TransformationException)
            {
               throw ((TransformationException) e);
            }
            else
            {
               throw new RuntimeException("failed to transform: " + clazz.getName(), e);
            }
         }

      }
   }

   private void addBaseElements(CtClass clazz)
           throws NotFoundException, CannotCompileException
   {
      addProtectedField(clazz,
              "_instanceAdvisor",
              "org.jboss.aop.ClassInstanceAdvisor", null
              //CtField.Initializer.byExpr("new org.jboss.aop.ClassInstanceAdvisor(this)")
      );

      CtMethod getter = CtNewMethod.make("public synchronized org.jboss.aop.InstanceAdvisor _getInstanceAdvisor()" +
              "{ " +
              "    if (_instanceAdvisor == null) { _instanceAdvisor = new org.jboss.aop.ClassInstanceAdvisor(this); }" +
              "    return _instanceAdvisor;" +
              "}",
              clazz);
      clazz.addMethod(getter);

      CtMethod setter = CtNewMethod.make("public synchronized void _setInstanceAdvisor(org.jboss.aop.InstanceAdvisor newAdvisor)" +
              "{ " +
              "    _instanceAdvisor = (org.jboss.aop.ClassInstanceAdvisor)newAdvisor;" +
              "}",
              clazz);
      clazz.addMethod(setter);

   }

   public List getConstructors(CtClass clazz)
   {
      List list = new ArrayList();

      CtConstructor[] constructors = clazz.getDeclaredConstructors();

      for (int i = 0; i < constructors.length; i++)
      {
         list.add(constructors[i]);
      }
      Collections.sort(list, CtConstructorComparator.INSTANCE);

      return list;
   }

   /**
    * Gets sorted collection of advisable methods.
    */
   public List getAdvisableFields(CtClass clazz) throws NotFoundException
   {
      List list = new ArrayList();
      CtField[] fields = clazz.getDeclaredFields();
      for (int i = 0; i < fields.length; i++)
      {
         if (isAdvisable(fields[i]))
         {
            list.add(fields[i]);
         }
      }
      Collections.sort(list, CtFieldComparator.INSTANCE);

      return list;
   }

   /**
    * Creates generic invoke method to be wrapped by real signatures.
    */
   private CtMethod createMixinInvokeMethod(CtClass clazz, CtClass mixinClass, String initializer, CtMethod method, long hash)
           throws CannotCompileException, NotFoundException, Exception
   {
      String code = null;
      String aopReturnStr = (method.getReturnType().equals(CtClass.voidType)) ? "" : "return ($r)";
      String returnStr = (method.getReturnType().equals(CtClass.voidType)) ? "" : "return ";
      String args = "null";
      if (method.getParameterTypes().length > 0) args = "$args";
      //If a call to a method introduced is intercepted, this original code uses the mixin class as the target
      //object. This causes a ClassCastException in PerInstanceAdvice since the mixin class does not implement
      //Advised
//			 code =
//						"{ " +
//						"   if (" + mixinFieldName(mixinClass) + " == null) { " +
//						"      this." + mixinFieldName(mixinClass) + " = " + initializer + "; " +
//						"   } " +
//						"   if (" + HELPER_FIELD_NAME + ".doesHaveAspects || (_instanceAdvisor != null && _instanceAdvisor.hasInstanceAspects)) " +
//						"   { " +
//						"      Object[] ags = " + args + "; " +
//						"      " + aopReturnStr  + HELPER_FIELD_NAME + ".invokeMethod(_instanceAdvisor, this." + mixinFieldName(mixinClass) + ", " + hash + "L, ags); " +
//						"   } " +
//						"   else " +
//						"   {" +
//						"      " + returnStr + " " + mixinFieldName(mixinClass) + "." + method.getName() + "($$); " +
//						"   }" +
//						"}";



      //1) add a not advised method that performs the call to the mixin class
      code =
              "{ " +
              "" + returnStr + " " + mixinFieldName(mixinClass) + "." + method.getName() + "($$); " +
              "}";
      String wrappedName = ClassAdvisor.notAdvisedMethodName(clazz.getName(), method.getName());
      CtMethod nmethod = CtNewMethod.make(method.getReturnType(), wrappedName, method.getParameterTypes(),
              method.getExceptionTypes(), null, clazz);
      nmethod.setModifiers(method.getModifiers());
      nmethod.setBody(code);
      clazz.addMethod(nmethod);



      //2 make the wrapped invocation call through to the helper method, using this as the target object
      code =
              "{ " +
              "   if (" + mixinFieldName(mixinClass) + " == null) { " +
              "      this." + mixinFieldName(mixinClass) + " = " + initializer + "; " +
              "   } " +
              "   if (" + HELPER_FIELD_NAME + ".doesHaveAspects || (_instanceAdvisor != null && _instanceAdvisor.hasInstanceAspects)) " +
              "   { " +
              "      Object[] ags = " + args + "; " +
              "      " + aopReturnStr + HELPER_FIELD_NAME + ".invokeMethod(_instanceAdvisor, this, " + hash + "L, ags); " +
              "   } " +
              "   else " +
              "   {" +
              "      " + returnStr + " " + mixinFieldName(mixinClass) + "." + method.getName() + "($$); " +
              "   }" +
              "}";

      CtMethod wmethod = CtNewMethod.make(method.getReturnType(), method.getName(), method.getParameterTypes(),
              method.getExceptionTypes(), null, clazz);
      wmethod.setModifiers(method.getModifiers());
      try
      {
         wmethod.setBody(code);
      }
      catch (CannotCompileException e)
      {
         throw new RuntimeException("<mixin> construction may have syntax error: '" + initializer + "'");
      }
      clazz.addMethod(wmethod);

      return wmethod;

   }

   /**
    * Creates generic invoke method to be wrapped by real signatures.
    */
   private CtMethod createInvokeMethod(CtClass clazz)
           throws CannotCompileException
   {
      return CtNewMethod.make("public java.lang.Object invoke(java.lang.Object[] args, long i)" +
              "       throws java.lang.Throwable {" +
              "   return ((org.jboss.aop.ClassAdvisor)this._getAdvisor()).invokeMethod(this, i, args);" +
              "}",
              clazz);
   }

   /**
    * Gets a class by its name.
    */
   public CtClass forName(String name) throws NotFoundException
   {
      return this.classPool.get(name);
   }

   /**
    * Adds a reference to this class's Closs object.
    */
   private void addClassReference(CtClass clazz) throws CannotCompileException,
           NotFoundException
   {
      addStaticField(clazz,
              CLASS_FIELD_NAME,
              "java.lang.Class",
              CtField.Initializer.byExpr(Class.class.getName() + "#forName(\"" +
              clazz.getName() + "\")"));
   }


   private void addHelperField(CtClass clazz) throws CannotCompileException, NotFoundException
   {
      addStaticField(clazz,
              HELPER_FIELD_NAME,
              HELPER_CLASS_NAME,
              CtField.Initializer.byExpr(ASPECT_MANAGER_CLASS_NAME + ".instance().getAdvisor(" +
              CLASS_FIELD_NAME + ")"));
   }

   /**
    * Adds a static helper class instance to passed class.
    */
   private void addHelperClass(CtClass clazz)
           throws CannotCompileException, NotFoundException
   {
      addHelperField(clazz);
      // To make sure that correct Advisor gets called
      CtMethod getter = CtNewMethod.make("public org.jboss.aop.Advisor _getAdvisor()" +
              "{ " +
              "    return " + HELPER_FIELD_NAME + ";" +
              "}",
              clazz);
      clazz.addMethod(getter);
   }

   /**
    * Adds a static field to a class.
    */
   private CtField addStaticField(CtClass clazz, String name, String typeName,
                                  CtField.Initializer initializer)
           throws CannotCompileException, NotFoundException
   {
      CtClass type = forName(typeName);
      CtField field = new CtField(type, name, clazz);
      field.setModifiers(Modifier.PRIVATE | Modifier.STATIC | Modifier.FINAL);
      clazz.addField(field, initializer);

      return field;
   }

   /**
    * Adds a protected field to a class.
    */
   private CtField addProtectedField(CtClass clazz, String name, String typeName,
                                     CtField.Initializer initializer)
           throws CannotCompileException, NotFoundException
   {
      CtClass type = forName(typeName);
      CtField field = new CtField(type, name, clazz);
      field.setModifiers(Modifier.PROTECTED | Modifier.TRANSIENT);
      if (initializer != null)
      {
         clazz.addField(field, initializer);
      }
      else
      {
         clazz.addField(field);
      }
      return field;
   }

   public void setupBasics(CtClass clazz) throws CannotCompileException, NotFoundException
   {
      if (basicsSet) return;
      basicsSet = true;
      // add serialVersionUID.
      SerialVersionUID.setSerialVersionUID(clazz);

      // add marker interface.
      clazz.addInterface(forName(AOP_PACKAGE + ".Advised"));

      // add reference to this class's Class object.
      addClassReference(clazz);

      // add aop helper class.
      addHelperClass(clazz);

      if (isBaseClass(clazz))
      {
         addBaseElements(clazz);
      }
   }
}
