/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.aop.instrument;

import javassist.CannotCompileException;
import javassist.ClassPool;
import javassist.CtClass;
import javassist.CtField;
import javassist.CtMethod;
import javassist.CtNewMethod;
import javassist.CtPrimitiveType;
import javassist.Modifier;
import javassist.NotFoundException;
import javassist.bytecode.AnnotationsAttribute;
import javassist.bytecode.MethodInfo;
import org.jboss.aop.AOPClassPool;
import org.jboss.aop.AspectManager;
import org.jboss.aop.ClassAdvisor;
import org.jboss.aop.pointcut.Pointcut;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.HashMap;

/**
 * Comment
 *
 * @author <a href="mailto:bill@jboss.org">Bill Burke</a>
 * @version $Revision: 1.15.2.5 $
 */
public class MethodExecutionTransformer
{
   private Instrumentor instrumentor;

   public MethodExecutionTransformer(Instrumentor instrumentor)
   {
      this.instrumentor = instrumentor;
   }

   private String addMethodInfoField(CtClass clazz, String methodName, long methodHash) throws Exception
   {
      CtClass type = instrumentor.forName("org.jboss.aop.MethodInfo");
      String name = getMethodInfoFieldName(methodName, methodHash);
      CtField field = new CtField(type, name, clazz);
      field.setModifiers(Modifier.PRIVATE | Modifier.STATIC);
      clazz.addField(field);
      return name;
   }

   public static String getMethodInfoFieldName(String methodName, long methodHash)
   {
      String hash;
      if (methodHash < 0)
         hash = "_N_" + (-1 * methodHash);
      else
         hash = "" + methodHash;
      String name = "aop$MethodInfo_" + methodName + hash;
      return name;
   }


   /**
    * Gets sorted collection of advisable methods.
    */
   private List getAdvisableMethods(CtClass clazz, ClassAdvisor advisor) throws NotFoundException
   {
      List list = new ArrayList();
      CtMethod[] methods = clazz.getDeclaredMethods();
      for (int i = 0; i < methods.length; i++)
      {
         if (Instrumentor.isAdvisable(methods[i]))
         {
            if (AspectManager.verbose) System.out.println("[debug] is advisable method: " + methods[i].getName());
            Iterator pointcuts = AspectManager.instance().getPointcuts().values().iterator();
            while (pointcuts.hasNext())
            {
               Pointcut pointcut = (Pointcut) pointcuts.next();

               boolean matches = pointcut.matchesExecution(advisor, methods[i]);
               if (AspectManager.verbose) System.out.println("[debug] does " + methods[i].getName() + " match " + pointcut.getExpr() + ": " + matches);
               if (matches)
               {
                  list.add(methods[i]);
                  break;
               }
            }
         }
      }
      return list;
   }

   public static String getOptimizedInvocationClassName(CtClass clazz, CtMethod method) throws Exception
   {
      long hash = org.jboss.aop.util.JavassistMethodHashing.methodHash(method);
      String className = clazz.getName() + "_" + method.getName() + "_" + hash + "_OptimizedMethodInvocation";
      className = className.replace('-', 'N');
      return className;
   }

   protected String createOptimizedInvocationClass(CtClass clazz, CtMethod method) throws Exception
   {
      String wrappedName = ClassAdvisor.notAdvisedMethodName(clazz.getName(),
              method.getName());
      AOPClassPool pool = (AOPClassPool) instrumentor.getClassPool();
      CtClass methodInvocation = pool.get("org.jboss.aop.joinpoint.MethodInvocation");

      String className = getOptimizedInvocationClassName(clazz, method);
      boolean makeInnerClass = !Modifier.isPublic(method.getModifiers());

      CtClass invocation = TransformerCommon.makeInvocationClass(pool, makeInnerClass, clazz, className, methodInvocation);
      CtClass[] params = method.getParameterTypes();
      TransformerCommon.addArgumentFieldsToInvocation(invocation, params);

      boolean isStatic = javassist.Modifier.isStatic(method.getModifiers());
      if (!isStatic)
      {
         CtField target = new CtField(method.getDeclaringClass(), "typedTargetObject", invocation);
         target.setModifiers(Modifier.PUBLIC);
         invocation.addField(target);
      }

      CtMethod in = methodInvocation.getDeclaredMethod("invokeNext");
      CtMethod invokeNext = CtNewMethod.make(in.getReturnType(), "invokeNext", in.getParameterTypes(), in.getExceptionTypes(), null, invocation);
      invokeNext.setModifiers(in.getModifiers());

      String code = TransformerCommon.STANDARD_INTERCEPTOR_CODE;

      String returnStr = (method.getReturnType().equals(CtClass.voidType)) ? "" : "return ($w)";
      if (isStatic)
      {
         code +=
                 "   " + returnStr + " " + method.getDeclaringClass().getName() + ".";
      }
      else
      {
         code +=
                 "   " + returnStr + " typedTargetObject.";
      }
      code += wrappedName + "(";
      for (int i = 0; i < params.length; i++)
      {
         if (i > 0) code += ", ";
         code += "arg" + i;
      }
      code += ");  ";
      if (method.getReturnType().equals(CtClass.voidType))
      {
         code += " return null; ";
      }
      code += "}";
      try
      {
         invokeNext.setBody(code);
      }
      catch (CannotCompileException e)
      {
         System.out.println(code);
         throw e;
      }
      invocation.addMethod(invokeNext);
      TransformerCommon.addGetArguments(pool, invocation, method.getParameterTypes());
      addSetArguments(pool, invocation, method.getParameterTypes());
      addCopy(pool, invocation, method.getParameterTypes(), isStatic);

      TransformerCommon.compileOrLoadClass(method.getDeclaringClass(), invocation);

      //Return fully qualified name of class (may be an inner class)
      return invocation.getName();
   }

   public void instrument(CtClass clazz, ClassAdvisor advisor)
           throws Exception
   {
      Iterator iterator = getAdvisableMethods(clazz, advisor).iterator();
      while (iterator.hasNext())
      {
         instrumentor.setupBasics(clazz);
         CtMethod method = (CtMethod) iterator.next();
         long hash = org.jboss.aop.util.JavassistMethodHashing.methodHash(method);
         String name = method.getName();
         String methodInfoField = addMethodInfoField(clazz, name, hash);

         if (AspectManager.optimize)
         {
            optimized(method, clazz, name, methodInfoField);
         }
         else
         {
            unoptimized(method, clazz, name, hash, methodInfoField);
         }


      }
   }


   public static String setArguments(CtClass[] params) throws Exception
   {
      String code = "";
      for (int i = 0; i < params.length; i++)
      {
         code += "   invocation.arg" + i + " = $" + (i + 1) + "; ";
      }
      return code;
   }

   public static void addSetArguments(ClassPool pool, CtClass invocation, CtClass[] params) throws Exception
   {
      if (params == null || params.length == 0) return;
      CtClass methodInvocation = pool.get("org.jboss.aop.joinpoint.MethodInvocation");
      CtMethod template = methodInvocation.getDeclaredMethod("setArguments");

      String code =
              "public void setArguments(java.lang.Object[] args)" +
              "{ ";
      code += "   arguments = args; ";
      for (int i = 0; i < params.length; i++)
      {
         if (params[i].isPrimitive())
         {
            CtPrimitiveType primitive = (CtPrimitiveType) params[i];
            code += "   arg" + i + " = ((" + primitive.getWrapperName() + ")args[" + i + "])." + primitive.getGetMethodName() + "(); ";
         }
         else
         {
            code += "   Object warg" + i + " = args[" + i + "]; ";
            code += "   arg" + i + " = (" + params[i].getName() + ")warg" + i + "; ";
         }
      }
      code += "}";
      CtMethod setArguments = CtNewMethod.make(code, invocation);
      //CtMethod setArguments = CtNewMethod.make(template.getReturnType(), "setArguments", template.getParameterTypes(), template.getExceptionTypes(), code, invocation);
      setArguments.setModifiers(template.getModifiers());
      invocation.addMethod(setArguments);
   }

   private void addCopy(ClassPool pool, CtClass invocation, CtClass[] params, boolean isStatic) throws Exception
   {
      CtClass methodInvocation = pool.get("org.jboss.aop.joinpoint.MethodInvocation");
      CtMethod template = methodInvocation.getDeclaredMethod("copy");

      CtMethod copy = CtNewMethod.make(template.getReturnType(), "copy", template.getParameterTypes(), template.getExceptionTypes(), null, invocation);
      copy.setModifiers(template.getModifiers());

      String code =
              "{ " +
              "   " + invocation.getName() + " wrapper = new " + invocation.getName() + "(this.info, this.interceptors); " +
              "   wrapper.arguments = this.arguments; " +
              "   wrapper.metadata = this.metadata; " +
              "   wrapper.currentInterceptor = this.currentInterceptor; " +
              "   wrapper.instanceResolver = this.instanceResolver; ";
      if (!isStatic)
      {
         code += "   wrapper.typedTargetObject = this.typedTargetObject; ";
         code += "   wrapper.targetObject = this.targetObject; ";
      }

      for (int i = 0; i < params.length; i++)
      {
         code += "   wrapper.arg" + i + " = this.arg" + i + "; ";
      }
      code += "   return wrapper; }";

      copy.setBody(code);
      invocation.addMethod(copy);

   }

   private void copyAnnotations(CtMethod src, CtMethod dest)
   {
      MethodInfo mi = src.getMethodInfo2();
      MethodInfo wmi = dest.getMethodInfo2();
      AnnotationsAttribute invisible = (AnnotationsAttribute) mi.getAttribute(AnnotationsAttribute.invisibleTag);
      if (invisible != null)
      {
         wmi.addAttribute(invisible.copy(wmi.getConstPool(), new HashMap()));
      }
      AnnotationsAttribute visible = (AnnotationsAttribute) mi.getAttribute(AnnotationsAttribute.visibleTag);
      if (visible != null)
      {
         wmi.addAttribute(visible.copy(wmi.getConstPool(), new HashMap()));
      }
   }

   private void optimized(CtMethod method, CtClass clazz, String name, String methodInfoField) throws Exception
   {
      boolean isStatic = Modifier.isStatic(method.getModifiers());
      String wrappedName = ClassAdvisor.notAdvisedMethodName(clazz.getName(),
              method.getName());
      CtMethod wmethod = CtNewMethod.copy(method, clazz, null);

      String originalName = method.getName();
      wmethod.setName(wrappedName);
      clazz.addMethod(wmethod);
      copyAnnotations(method, wmethod);
      String optimizedInvocation = createOptimizedInvocationClass(clazz, method);
      method.setName(wrappedName);
      wmethod.setName(originalName);

      //make a copy of method and modify the existing method because we need to retain annotations
      //of existing method
      //wmethod = method;
      String code = null;
      String aopReturnStr = (method.getReturnType().equals(CtClass.voidType)) ? "" : "return ($r)";
      String returnStr = (method.getReturnType().equals(CtClass.voidType)) ? "" : "return ";
      if (!isStatic)
      {
         code =
                 "{ " +
                 "    if (" + Instrumentor.HELPER_FIELD_NAME + ".doesHaveAspects || (_instanceAdvisor != null && _instanceAdvisor.hasInstanceAspects)) " +
                 "    { " +
                 "       org.jboss.aop.advice.Interceptor[] interceptors = " + methodInfoField + ".interceptors;" +
                 "       if (_instanceAdvisor != null) " +
                 "       { " +
                 "          interceptors = _instanceAdvisor.getInterceptors(" + methodInfoField + ".interceptors); " +
                 "       } " +
                 "       " + optimizedInvocation + " invocation = new " + optimizedInvocation + "(" + methodInfoField + ", interceptors); " +
                 setArguments(method.getParameterTypes()) +
                 "       invocation.setTargetObject(this); " +
                 "       invocation.typedTargetObject = this; " +
                 "       invocation.setAdvisor(" + Instrumentor.HELPER_FIELD_NAME + "); " +
                 "       " + aopReturnStr + "invocation.invokeNext(); " +
                 "    } " +
                 "    else " +
                 "    {" +
                 "       " + returnStr + " " + wrappedName + "($$); " +
                 "    }" +
                 "}";
      }
      else
      {
         code =
                 "{ " +
                 "    if (" + Instrumentor.HELPER_FIELD_NAME + ".doesHaveAspects) " +
                 "    { " +
                 "       org.jboss.aop.advice.Interceptor[] interceptors = " + methodInfoField + ".interceptors;" +
                 "       " + optimizedInvocation + " invocation = new " + optimizedInvocation + "(" + methodInfoField + ", interceptors); " +
                 setArguments(method.getParameterTypes()) +
                 "       invocation.setAdvisor(" + Instrumentor.HELPER_FIELD_NAME + "); " +
                 "       " + aopReturnStr + "invocation.invokeNext(); " +
                 "    } " +
                 "    else " +
                 "    {" +
                 "       " + returnStr + " " + wrappedName + "($$); " +
                 "    }" +
                 "}";
      }

      try
      {
         wmethod.setBody(code);
      }
      catch (CannotCompileException e)
      {
         e.printStackTrace();
         throw new RuntimeException("code was: " + code + " for method " + name);  //To change body of catch statement use Options | File Templates.
      }
   }

   private void unoptimized(CtMethod method, CtClass clazz, String name, long hash, String methodInfoField) throws NotFoundException, CannotCompileException
   {
      boolean isStatic = Modifier.isStatic(method.getModifiers());
      String wrappedName = ClassAdvisor.notAdvisedMethodName(clazz.getName(),
              method.getName());
      CtMethod wmethod = CtNewMethod.copy(method, clazz, null);

      String originalName = method.getName();
      wmethod.setName(wrappedName);
      clazz.addMethod(wmethod);
      copyAnnotations(method, wmethod);
      method.setName(wrappedName);
      wmethod.setName(originalName);


      //make a copy of method and modify the existing method because we need to retain annotations
      //of existing method
      //wmethod = method;
      String code = null;
      String aopReturnStr = (method.getReturnType().equals(CtClass.voidType)) ? "" : "return ($r)";
      String returnStr = (method.getReturnType().equals(CtClass.voidType)) ? "" : "return ";
      String args = "null";
      if (method.getParameterTypes().length > 0) args = "$args";
      if (!isStatic)
      {
         code =
                 "{ " +
                 "    if (" + Instrumentor.HELPER_FIELD_NAME + ".doesHaveAspects || (_instanceAdvisor != null && _instanceAdvisor.hasInstanceAspects)) " +
                 "    { " +
                 "       Object[] ags = " + args + "; " +
                 "       " + aopReturnStr + Instrumentor.HELPER_FIELD_NAME + ".invokeMethod(_instanceAdvisor, this, " + hash + "L, ags, " + methodInfoField + "); " +
                 "    } " +
                 "    else " +
                 "    {" +
                 "       " + returnStr + " " + wrappedName + "($$); " +
                 "    }" +
                 "}";
      }
      else
      {
         code =
                 "{ " +
                 "    if (" + Instrumentor.HELPER_FIELD_NAME + ".doesHaveAspects) " +
                 "    { " +
                 "       org.jboss.aop.ClassInstanceAdvisor ia = null; " +
                 "       Object[] ags = " + args + "; " +
                 "       Object target = null; " +
                 "       " + aopReturnStr + Instrumentor.HELPER_FIELD_NAME + ".invokeMethod(ia, target, " + hash + "L, ags, " + methodInfoField + "); " +
                 "    } " +
                 "    else " +
                 "    {" +
                 "       " + returnStr + " " + wrappedName + "($$); " +
                 "    }" +
                 "}";
      }


      try
      {
         wmethod.setBody(code);
      }
      catch (CannotCompileException e)
      {
         throw new RuntimeException("code was: " + code + " for method " + name, e);  //To change body of catch statement use Options | File Templates.
      }
   }
}
