/* *************************************
 *                                     *
 *  JBoss: The OpenSource J2EE WebOS   *
 *                                     *
 *  Distributable under LGPL license.  *
 *  See terms of license at gnu.org.   *
 *                                     *
 ************************************* */
package org.jboss.aop.instrument;

import javassist.CannotCompileException;
import javassist.ClassPool;
import javassist.CtClass;
import javassist.CtConstructor;
import javassist.CtField;
import javassist.CtMethod;
import javassist.CtNewConstructor;
import javassist.CtNewMethod;
import javassist.Modifier;
import javassist.NotFoundException;
import org.jboss.aop.AOPClassPool;
import org.jboss.aop.standalone.Compiler;

import java.io.File;
import java.io.FileOutputStream;
import java.net.URI;
import java.net.URL;

/** 
 *  A few handy methods and common things used by the other Transformers
 * @author <a href="mailto:kabirkhan@bigfoot.com">Kabir Khan</a>
 * @version $Revision: 1.2 $
 *
 */
public class TransformerCommon {

   final static String STANDARD_INTERCEPTOR_CODE 
      = "{ "
      + "   if (currentInterceptor < interceptors.length) "
      + "   { "
      + "      try   "
      + "      { "
      + "         return interceptors[currentInterceptor++].invoke(this); "
      + "      }  " + "      catch (Throwable t)  " + "      {  "
      + "         currentInterceptor--;  " + "         throw t; "
      + "      } " + "   } ";

   
   /**
    * 
    * @param pool The AOPClassPool to create the optimized invocation class in
    * @param makeInnerClass If true creates the new class as an inner class of className
    * @param outerClass The class to create the invocation class as an inner class of if makeInnerClass==true
    * @param className The full class name (including package info) of the invocation class to be created
    * @param superInvocation The super class of this invocation
    * @return The created invocation class
    */
   public static CtClass makeInvocationClass(AOPClassPool pool, 
         boolean makeInnerClass, 
         CtClass outerClass, 
         String className, 
         CtClass superInvocation)throws CannotCompileException, NotFoundException
   {
      CtClass untransformable = pool.get("org.jboss.aop.instrument.Untransformable");
     
      CtClass invocation;
      if (makeInnerClass)
      {
         //Strip away the package from classname
         String innerClassName = className.substring(className.lastIndexOf('.') + 1);
      
         //Only static nested classes are supported
         boolean classStatic = true;
         invocation = outerClass.makeNestedClass(innerClassName, classStatic); 
         invocation.setSuperclass(superInvocation);
      }
      else
      {
         invocation = pool.makeClass(className, superInvocation);
      }

      //Add the invocation constructor
      invocation.addInterface(untransformable);
      CtConstructor conTemplate = superInvocation.getDeclaredConstructors()[0];
      CtConstructor icon = CtNewConstructor.make(conTemplate.getParameterTypes(), 
               conTemplate.getExceptionTypes(),
               invocation);
      invocation.addConstructor(icon);

      return invocation;
   }

   /** Compiles the class to file or adds it to the class pool  
    * 
    * @param classForPackage The class to be used to determine the directory to place the class in
    * @param invocation The class to be comiled/added to class pool
    * @throws Exception
    */
   public static void compileOrLoadClass(CtClass classForPackage, CtClass invocation)throws Exception
   {
      // If compile time
      if (Compiler.loader != null)
      {
         File file;
         URL url = Compiler.loader.getResource(
               classForPackage.getName().replace('.', '/') + ".class");
         String path = url.toString();
         path = path.substring(0, path.lastIndexOf('/') + 1);
         path = path + invocation.getSimpleName() + ".class";
         URI newUrl = new URI(path);
         file = new File(newUrl);
         FileOutputStream fp = new FileOutputStream(file);
         fp.write(invocation.toBytecode());
         fp.close();
      }
      else
      // if load time
      {
         invocation.toClass();
      }
   }

   /** Adds fields arg0, arg1 etc. to the invocation class for storing the parameters for a method
    * 
    * @param invocation The invocation we want to add 
    * @param params Array of the types of the parameters
    * @throws CannotCompileException 
    */
   public static void addArgumentFieldsToInvocation(CtClass invocation, CtClass[] params)throws CannotCompileException
   {
      for (int i = 0 ; i < params.length ; i++)
      {
         CtField field = new CtField(params[i], "arg" + i, invocation);
         field.setModifiers(Modifier.PUBLIC);
         invocation.addField(field);
      }
   }
   
   public static void addGetArguments(ClassPool pool, CtClass invocation, CtClass[] params) throws Exception
   {
      if (params == null || params.length == 0) return;
      try {
         CtClass superInvocation = invocation.getSuperclass();
         CtMethod template = superInvocation.getDeclaredMethod("getArguments");
  
         String code =
                 "public Object[] getArguments()" +
                 "{ " +
                 "   if (arguments != (Object[])null) { return (Object[])arguments; } " +
                 "   arguments = new Object[" + params.length + "]; " +
                 "";
         for (int i = 0; i < params.length; i++)
         {
            code += "   arguments[" + i + "] = ($w)arg" + i + "; ";
         }
  
         code += "   return arguments; }";
         CtMethod getArguments = CtNewMethod.make(code, invocation);
         getArguments.setModifiers(template.getModifiers());
         invocation.addMethod(getArguments);
      } catch (NotFoundException e) {
        //Field invocations don't have a getArguments() method, that's fine
      } 
   }
      
}
