/***************************************
 *                                     *
 *  JBoss: The OpenSource J2EE WebOS   *
 *                                     *
 *  Distributable under LGPL license.  *
 *  See terms of license at gnu.org.   *
 *                                     *
 ***************************************/

package org.jboss.aop.introduction;

import javassist.CtClass;
import org.jboss.aop.Advisor;
import org.jboss.aop.AspectManager;
import org.jboss.aop.ClassAdvisor;
import org.jboss.aop.pointcut.TypeMatcher;
import org.jboss.aop.pointcut.Util;
import org.jboss.aop.pointcut.ast.ASTStart;
import org.jboss.aop.pointcut.ast.ClassExpression;

import java.util.ArrayList;

/**
 *
 *
 * @author <a href="mailto:bill@jboss.org">Bill Burke</a>
 * @version $Revision: 1.5.2.4 $
 *
 */
public class InterfaceIntroduction
{
   public static class Mixin
   {
      protected final String classname;
      protected final String[] interfaces;
      protected final String construction;
      protected final boolean trans;

      public Mixin(String classname, String[] interfaces, String construction, boolean trans)
      {
         this.classname = classname;
         this.interfaces = interfaces;
         this.construction = construction;
         this.trans = trans;
      }

      public String getClassName()
      {
         return classname;
      }

      public String[] getInterfaces()
      {
         return interfaces;
      }

      public String getConstruction()
      {
         return construction;
      }

      public boolean isTransient()
      {
         return trans;
      }
      
      public String toString()
      {
         return "Mixin: Classname=" + classname + ";interfaces=" + interfaces + ";construction=" + construction + ";transient=" + trans;
      }
   }

   protected String name;
   protected ArrayList advisors = new ArrayList();
   protected String[] interfaces;
   protected ArrayList mixins = new ArrayList();
   protected ClassExpression classExpr;
   protected ASTStart ast;

   public InterfaceIntroduction(String name, String exp, String[] interfaces)
   {
      this.name = name;
      this.interfaces = interfaces;
      this.classExpr = new ClassExpression(exp);
   }

   public InterfaceIntroduction(String name, ASTStart ast, String[] interfaces)
   {
      this.name = name;
      this.ast = ast;
      this.interfaces = interfaces;
   }

   public String getName()
   {
      return name;
   }

   public String[] getInterfaces()
   {
      return interfaces;
   }

   public ArrayList getMixins()
   {
      return mixins;
   }


   public void addAdvisor(Advisor advisor)
   {
      // Don't hold a direct reference to an advisor because of undeploy and redeploy
      advisors.add(advisor.getName());
      advisor.addInterfaceIntroduction(this);
   }

   public void clearAdvisors()
   {
      for (int i = 0; i < advisors.size(); i++)
      {
         String advisorName = (String) advisors.get(i);
         Advisor advisor = AspectManager.instance().getAdvisor(advisorName);
         if (advisor != null)
            advisor.removeInterfaceIntroduction(this);
      }
   }

   public boolean equals(Object obj)
   {
      if (obj == this) return true;
      if (!(obj instanceof InterfaceIntroduction)) return false;
      return ((InterfaceIntroduction) obj).getName().equals(name);
   }

   public int hashCode()
   {
      return name.hashCode();
   }

   public String getClassExpr()
   {
      if (classExpr == null) return null;
      return classExpr.getOriginal();
   }

   public ASTStart getAst()
   {
      return ast;
   }

   public boolean matches(Advisor advisor, CtClass clazz)  throws Exception
   {
      boolean match = false;
      if (classExpr != null) return Util.matchesClassExpr(classExpr, clazz, advisor);
      else
      {
         TypeMatcher matcher = new TypeMatcher((ClassAdvisor)advisor, clazz);
         return ((Boolean)ast.jjtAccept(matcher, null)).booleanValue();
      }
   }
   public boolean matches(Advisor advisor, Class clazz)
   {
      if (classExpr != null) return Util.matchesClassExpr(classExpr, clazz, advisor);
      else
      {
         TypeMatcher matcher = new TypeMatcher((ClassAdvisor)advisor, clazz);
         return ((Boolean)ast.jjtAccept(matcher, null)).booleanValue();
      }
   }
   
   public String toString()
   {
      return "InterfaceIntroduction:{ name=" + name + "; class=" + classExpr.getClass().getName() + "; ast=" + ast + ";interfaces=" + interfaces + ";mixins=" + mixins + "}";
   }
   
}
