/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */

package org.jboss.aop.joinpoint;

import org.jboss.aop.CallerConstructorInfo;
import org.jboss.aop.advice.Interceptor;
import org.jboss.aop.util.MethodHashing;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;


/**
 * This is a helper wrapper class for an Invocation object.
 * It is used to add or get values or metadata that pertains to
 * an AOP method invocation.
 *
 * @author <a href="mailto:bill@jboss.org">Bill Burke</a>
 * @version $Revision: 1.9 $
 */
public class ConstructorCalledByMethodInvocation extends InvocationBase
{
   protected Object[] arguments = null;
   protected long callingMethodHash;
   protected CallerConstructorInfo info;
   protected Class callingClass;
   protected Method callingMethod;

   public ConstructorCalledByMethodInvocation(Interceptor[] interceptors, CallerConstructorInfo info, long callingHash, Class callingClass, Object[] args)
   {
      super(interceptors);
      this.info = info;
      this.callingMethodHash = callingHash;
      this.callingClass = callingClass;
      this.arguments = args;
   }

   protected ConstructorCalledByMethodInvocation(Interceptor[] interceptors)
   {
      super(interceptors);
   }

   /**
    *
    * @return the arguments of the called constructor
    */
   public Object[] getArguments()
   {
      return arguments;
   }

   /**
    * change the arguments to the called constructor
    * @param arguments
    */
   public void setArguments(Object[] arguments)
   {
      this.arguments = arguments;
   }

   /**
    *
    * @return The class that is making the call on the constructor
    */
   public Class getCallingClass()
   {
      return callingClass;
   }

   /**
    *
    * @return The method that is making the call on the constructor
    */
   public Method getCallingMethod()
   {
      if (callingMethod != null) return callingMethod;
      try
      {
         callingMethod = MethodHashing.findMethodByHash(callingClass, callingMethodHash);
      }
      catch (Exception e)
      {
         throw new RuntimeException(e);  //To change body of catch statement use Options | File Templates.
      }
      return callingMethod;
   }

   /**
    *
    * @return the constructor call being executed by the calling method
    */
   public Constructor getCalledConstructor() { return info.constructor; }

   /**
    * Is the called constructor aspectized?  If so then there is a wrapping
    * method that must be called.
    * @return
    */
   public boolean isWrapped() { return info.wrappingMethod != null; }

   /**
    * Is the called constructor aspectized?  If so then this method
    * returns the method that wraps the constructor.
    * @return
    */
   public Method getWrappingMethod() { return info.wrappingMethod; }

   /**
    * Invoke on the next interceptor in the chain.  If this is already
    * the end of the chain, reflection will call the constructor, field, or
    * method you are invoking on.
    */
   public Object invokeNext() throws Throwable
   {
      if (currentInterceptor < interceptors.length)
      {
         try
         {
            return interceptors[currentInterceptor++].invoke(this);
         }
         finally
         {
            // so that interceptors like clustering can reinvoke down the chain
            currentInterceptor--;
         }
      }

      if (info.wrappingMethod != null)
      {
         try
         {
            return info.wrappingMethod.invoke(null, arguments);
         }
         catch (InvocationTargetException e)
         {
            throw e.getTargetException();
         }
      }
      else
      {
         try
         {
            return info.constructor.newInstance(arguments);
         }
         catch (InstantiationException e)
         {
            throw new RuntimeException(e);  //To change body of catch statement use Options | File Templates.
         }
         catch (IllegalAccessException e)
         {
            throw new RuntimeException(e);  //To change body of catch statement use Options | File Templates.
         }
         catch (InvocationTargetException e)
         {
            throw e.getCause();
         }
      }

   }

   /**
    * This method resolves metadata based on the context of the invocation.
    * It iterates through its list of MetaDataResolvers to find out the
    * value of the metadata desired.
    *
    * This list usually is ThreadMetaData, InstanceAdvisor.getMetaData
    * ClassAdvisor.getMethodMetaData (or field, or constructor)
    * ClassAdvisor.getDefaultMetaData
    */
   public Object getMetaData(Object group, Object attr)
   {
      Object val = super.getMetaData(group, attr);
      if (val != null) return val;

      // todo what to do for caller side metadata?
      return null;
   }

   /**
    * Get a wrapper invocation object that can insert a new chain of interceptors
    * at runtime to the invocation flow.  CFlow makes use of this.
    * When the wrapper object finishes its invocation chain it delegates back to
    * the wrapped invocation.
    * @param newchain
    * @return
    */
   public Invocation getWrapper(Interceptor[] newchain)
   {
      ConstructorCalledByMethodInvocationWrapper wrapper = new ConstructorCalledByMethodInvocationWrapper(this, newchain);
      return wrapper;
   }

   /**
    * Copies complete state of Invocation object.
    * @return
    */
   public Invocation copy()
   {
      ConstructorCalledByMethodInvocation wrapper = new ConstructorCalledByMethodInvocation(interceptors, info, callingMethodHash, callingClass, arguments);
      wrapper.setAdvisor(this.getAdvisor());
      wrapper.setTargetObject(this.getTargetObject());
      wrapper.metadata = this.metadata;
      wrapper.instanceResolver = this.instanceResolver;
      wrapper.currentInterceptor = this.currentInterceptor;
      return wrapper;
   }
}
