/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */

package org.jboss.aop.joinpoint;

import org.jboss.aop.advice.Interceptor;

import java.lang.reflect.Field;

/**
 * This is a helper wrapper class for an Invocation object.
 * It is used to add or get values or metadata that pertains to
 * an AOP field access or field update.
 *
 * @author <a href="mailto:bill@jboss.org">Bill Burke</a>
 * @version $Revision: 1.10 $
 */
public abstract class FieldInvocation extends InvocationBase
{
   protected transient Field field = null;
   protected int index;

   public FieldInvocation(Field field, int fieldIndex, Interceptor[] interceptors)
   {
      super(interceptors);
      this.field = field;
      this.index = fieldIndex;
   }

   protected FieldInvocation(Interceptor[] interceptors)
   {
      super(interceptors);
   }

   /**
    * This method resolves an annotation based on the context of the invocation.
    *
    */
   public Object resolveAnnotation(Class annotation)
   {
      Object val = super.resolveAnnotation(annotation);
      if (val != null) return val;

      if (getAdvisor() != null)
      {
         val = getAdvisor().resolveAnnotation(field, annotation);
         if (val != null) return val;
      }

      return null;
   }
   /**
    * This method resolves metadata based on the context of the invocation.
    * It iterates through its list of MetaDataResolvers to find out the
    * value of the metadata desired.
    *
    * This list usually is ThreadMetaData, InstanceAdvisor.getMetaData
    * ClassAdvisor.getMethodMetaData (or field, or constructor)
    * ClassAdvisor.getDefaultMetaData
    */
   public Object getMetaData(Object group, Object attr)
   {
      Object val = super.getMetaData(group, attr);
      if (val != null) return val;

      if (getAdvisor() != null)
      {
         val = getAdvisor().getFieldMetaData().resolve(this, group, attr);
         if (val != null) return val;
      }

      if (getAdvisor() != null)
      {
         val = getAdvisor().getDefaultMetaData().resolve(this, group, attr);
         if (val != null) return val;
      }

      return null;
   }

	/**
	 * This is the field the invocation is accessing
	 *
	 * @return
	 */
	public Field getField()
	{
		return field;
	}

   /**
    * This is an index into the Field[] array accessed through the ClassAdvisor
    *
    * @return
    */
   public int getIndex()
   {
      return index;
   }

}
