/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */

package org.jboss.aop.joinpoint;

import org.jboss.aop.advice.Interceptor;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;


/**
 * This is a helper wrapper class for an Invocation object.
 * It is used to add or get values or metadata that pertains to
 * an AOP method invocation.
 *
 * @author <a href="mailto:bill@jboss.org">Bill Burke</a>
 * @version $Revision: 1.5 $
 */
public class MethodCalledByConstructorInvocation extends InvocationBase
{
   protected Object[] arguments = null;
   protected Constructor calling;
   protected Method calledMethod;

   public MethodCalledByConstructorInvocation(Method called, Object[] args, Constructor calling, Interceptor[] interceptors)
   {
      super(interceptors);
      this.calledMethod = called;
      this.arguments = args;
      this.calling = calling;
   }

   protected MethodCalledByConstructorInvocation(Interceptor[] interceptors)
   {
      super(interceptors);
   }

   /**
    * Invoke on the next interceptor in the chain.  If this is already
    * the end of the chain, reflection will call the constructor, field, or
    * method you are invoking on.
    */
   public Object invokeNext() throws Throwable
   {
      if (currentInterceptor < interceptors.length)
      {
         try
         {
            return interceptors[currentInterceptor++].invoke(this);
         }
         finally
         {
            // so that interceptors like clustering can reinvoke down the chain
            currentInterceptor--;
         }
      }

      try
      {
         return calledMethod.invoke(getTargetObject(), arguments);
      }
      catch (InvocationTargetException e)
      {
         throw e.getTargetException();
      }

   }
   /**
    * This method resolves metadata based on the context of the invocation.
    * It iterates through its list of MetaDataResolvers to find out the
    * value of the metadata desired.
    *
    * This list usually is ThreadMetaData, InstanceAdvisor.getMetaData
    * ClassAdvisor.getMethodMetaData (or field, or constructor)
    * ClassAdvisor.getDefaultMetaData
    */
   public Object getMetaData(Object group, Object attr)
   {
      Object val = super.getMetaData(group, attr);
      if (val != null) return val;

      // todo what to do for caller side metadata?
      return null;
   }

   /**
    * Get a wrapper invocation object that can insert a new chain of interceptors
    * at runtime to the invocation flow.  CFlow makes use of this.
    * When the wrapper object finishes its invocation chain it delegates back to
    * the wrapped invocation.
    * @param newchain
    * @return
    */
   public Invocation getWrapper(Interceptor[] newchain)
   {
      MethodCalledByConstructorInvocationWrapper wrapper = new MethodCalledByConstructorInvocationWrapper(this, newchain);
      return wrapper;
   }

   /**
    * Copies complete state of Invocation object.
    * @return
    */
   public Invocation copy()
   {
      MethodCalledByConstructorInvocation wrapper = new MethodCalledByConstructorInvocation(calledMethod, arguments, calling, interceptors);
      wrapper.setAdvisor(this.getAdvisor());
      wrapper.setTargetObject(this.getTargetObject());
      wrapper.currentInterceptor = this.currentInterceptor;
      wrapper.metadata = this.metadata;
      wrapper.instanceResolver = this.instanceResolver;
      return wrapper;
   }

   /**
    * What are the arguments of the method call
    * The are expressed as an Object[] array.
    *
    * @return the arguments of the method call
    */
   public Object[] getArguments()
   {
      return arguments;
   }

   /**
    * Change the arguments of the method call
    * @param arguments
    */
   public void setArguments(Object[] arguments)
   {
      this.arguments = arguments;
   }

   /**
    * The constructor that is calling the method
    *
    * @return The constructor that is calling the method
    */
   public Constructor getCalling()
   {
      return calling;
   }

   /**
    *
    * @return the method that is being called
    */
   public Method getCalledMethod()
   {
      return calledMethod;
   }
}
