/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */

package org.jboss.aop.joinpoint;

import org.jboss.aop.advice.Interceptor;
import org.jboss.aop.util.MethodHashing;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;


/**
 * This is a helper wrapper class for an Invocation object.
 * It is used to add or get values or metadata that pertains to
 * an AOP method invocation.
 *
 * @author <a href="mailto:bill@jboss.org">Bill Burke</a>
 * @version $Revision: 1.7 $
 */
public class MethodCalledByMethodInvocation extends InvocationBase
{
   protected Object[] arguments = null;
   protected long callingMethodHash;
   protected Class callingClass;
   protected Method calledMethod;
   protected Method callingMethod;

   public MethodCalledByMethodInvocation(Interceptor[] interceptors, Method calledMethod, Object[] arguments, Class callingClass, long callingMethodHash)
   {
      super(interceptors);
      this.calledMethod = calledMethod;
      this.arguments = arguments;
      this.callingClass = callingClass;
      this.callingMethodHash = callingMethodHash;
   }


   protected MethodCalledByMethodInvocation(Interceptor[] interceptors)
   {
      super(interceptors);
   }


   /**
    * Invoke on the next interceptor in the chain.  If this is already
    * the end of the chain, reflection will call the constructor, field, or
    * method you are invoking on.
    */
   public Object invokeNext() throws Throwable
   {
      if (currentInterceptor < interceptors.length)
      {
         try
         {
            return interceptors[currentInterceptor++].invoke(this);
         }
         finally
         {
            // so that interceptors like clustering can reinvoke down the chain
            currentInterceptor--;
         }
      }

      try
      {
         return calledMethod.invoke(getTargetObject(), arguments);
      }
      catch (InvocationTargetException e)
      {
         throw e.getTargetException();
      }

   }

   /**
    * This method resolves metadata based on the context of the invocation.
    * It iterates through its list of MetaDataResolvers to find out the
    * value of the metadata desired.
    *
    * This list usually is ThreadMetaData, InstanceAdvisor.getMetaData
    * ClassAdvisor.getMethodMetaData (or field, or constructor)
    * ClassAdvisor.getDefaultMetaData
    */
   public Object getMetaData(Object group, Object attr)
   {
      Object val = super.getMetaData(group, attr);
      if (val != null) return val;

      // todo what to do for caller side metadata?
      return null;
   }

   /**
    * Get a wrapper invocation object that can insert a new chain of interceptors
    * at runtime to the invocation flow.  CFlow makes use of this.
    * When the wrapper object finishes its invocation chain it delegates back to
    * the wrapped invocation.
    * @param newchain
    * @return
    */
   public Invocation getWrapper(Interceptor[] newchain)
   {
      MethodCalledByMethodInvocationWrapper wrapper = new MethodCalledByMethodInvocationWrapper(this, newchain);
      return wrapper;
   }

   /**
    * Copies complete state of Invocation object.
    * @return
    */
   public Invocation copy()
   {
      MethodCalledByMethodInvocation wrapper = new MethodCalledByMethodInvocation(interceptors, calledMethod, arguments, callingClass, callingMethodHash);
      wrapper.setAdvisor(this.getAdvisor());
      wrapper.setTargetObject(this.getTargetObject());
      wrapper.currentInterceptor = this.currentInterceptor;
      wrapper.instanceResolver = this.instanceResolver;
      wrapper.metadata = this.metadata;
      return wrapper;
   }

   /**
    *
    * @return the arguments of the called method
    */
   public Object[] getArguments()
   {
      return arguments;
   }

   /**
    * change the arguments to the called method
    * @param arguments
    */
   public void setArguments(Object[] arguments)
   {
      this.arguments = arguments;
   }

   /**
    *
    * @return The class that is making the call on the method
    */
   public Class getCallingClass()
   {
      return callingClass;
   }

   /**
    *
    * @return The method that is making the call on the method
    */
   public Method getCallingMethod()
   {
      if (callingMethod != null) return callingMethod;
      try
      {
         callingMethod = MethodHashing.findMethodByHash(callingClass, callingMethodHash);
      }
      catch (Exception e)
      {
         throw new RuntimeException(e);  //To change body of catch statement use Options | File Templates.
      }
      return callingMethod;
   }

   public Method getCalledMethod()
   {
      return calledMethod;
   }

}
