/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */

package org.jboss.aop.joinpoint;

import org.jboss.aop.Advisor;
import org.jboss.aop.MethodInfo;
import org.jboss.aop.advice.Interceptor;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

/**
 *   
 *
 * @author <a href="mailto:bill@jboss.org">Bill Burke</a>
 * @version $Revision: 1.20 $
 */
public class MethodInvocation extends InvocationBase
{
   /**
    * arguments of the method call expressed as an array of objects
    */
   protected java.lang.Object[] arguments;
   protected long methodHash; // needs to be member variable as MethodInfo is not serializable across the wire.

   protected transient MethodInfo info;

   public MethodInvocation(MethodInfo info, org.jboss.aop.advice.Interceptor[] interceptors)
   {
      super(interceptors);
      this.info = info;
      methodHash = info.hash;

   }

   protected MethodInvocation(org.jboss.aop.advice.Interceptor[] interceptors)
   {
      super(interceptors);
   }

   /**
    * Invoke on the next interceptor in the chain.  If this is already
    * the end of the chain, reflection will call the constructor, field, or
    * method you are invoking on.
    */
   public Object invokeNext() throws Throwable
   {
      if (currentInterceptor < interceptors.length)
      {
         try
         {
            return interceptors[currentInterceptor++].invoke(this);
         }
         finally
         {
            // so that interceptors like clustering can reinvoke down the chain
            currentInterceptor--;
         }
      }

      try
      {
         return getActualMethod().invoke(getTargetObject(),arguments);
      }
      catch (InvocationTargetException e)
      {
         throw e.getTargetException();
      }

   }

   /**
    * This method resolves an annotation based on the context of the invocation.
    *
    */
   public Object resolveAnnotation(Class annotation)
   {
      Object val = super.resolveAnnotation(annotation);
      if (val != null) return val;

      if (getAdvisor() != null)
      {
         val = getAdvisor().resolveAnnotation(getMethod(), annotation);
         if (val != null) return val;
      }

      return null;
   }

   /**
    * This method resolves metadata based on the context of the invocation.
    * It iterates through its list of MetaDataResolvers to find out the
    * value of the metadata desired.
    *
    * This list usually is ThreadMetaData, InstanceAdvisor.getMetaData
    * ClassAdvisor.getMethodMetaData (or field, or constructor)
    * ClassAdvisor.getDefaultMetaData
    */
   public Object getMetaData(Object group, Object attr)
   {
      Object val = super.getMetaData(group, attr);
      if (val != null) return val;

      if (getAdvisor() != null)
      {
         val = getAdvisor().getMethodMetaData().resolve(this, group, attr);
         if (val != null) return val;
      }

      if (getAdvisor() != null)
      {
         val = getAdvisor().getDefaultMetaData().resolve(this, group, attr);
         if (val != null) return val;
      }

      return null;
   }

   /**
    * Get a wrapper invocation object that can insert a new chain of interceptors
    * at runtime to the invocation flow.  CFlow makes use of this.
    * When the wrapper object finishes its invocation chain it delegates back to
    * the wrapped invocation.
    * @param newchain
    * @return
    */
   public Invocation getWrapper(Interceptor[] newchain)
   {
      MethodInvocationWrapper wrapper = new MethodInvocationWrapper(this, newchain);
      return wrapper;
   }

   /**
    * Copies complete state of Invocation object.
    * @return
    */
   public Invocation copy()
   {
      MethodInvocation wrapper = new MethodInvocation(info, interceptors);
      wrapper.metadata = this.metadata;
      wrapper.currentInterceptor = this.currentInterceptor;
      wrapper.instanceResolver = this.instanceResolver;
      wrapper.setTargetObject(this.getTargetObject());
      wrapper.setArguments(this.getArguments());
      return wrapper;
   }

   public Object[] getArguments()
   {
      return arguments;
   }

   public void setArguments(Object[] arguments)
   {
      this.arguments = arguments;
   }

   public Method getMethod()
   {
      return info.advisedMethod;
   }

   public Method getActualMethod()
   {
      return info.unadvisedMethod;
   }

   public long getMethodHash()
   {
      return methodHash;
   }

   public Advisor getAdvisor()
   {
      return info.advisor;
   }

}
