/***************************************
 *                                     *
 *  JBoss: The OpenSource J2EE WebOS   *
 *                                     *
 *  Distributable under LGPL license.  *
 *  See terms of license at gnu.org.   *
 *                                     *
 ***************************************/

package org.jboss.aop.metadata;
import org.jboss.aop.joinpoint.Invocation;
import org.jboss.aop.util.MarshalledValue;
import org.jboss.aop.util.PayloadKey;

import java.io.IOException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;

/**
 *
 * @author <a href="mailto:bill@jboss.org">Bill Burke</a>
 * @version $Revision: 1.6 $
 *
 */
public class SimpleMetaData implements MetaDataResolver, java.io.Externalizable
{
   protected HashMap metaData = new HashMap();

   public class MetaDataValue implements java.io.Serializable
   {
      
      public final PayloadKey type;
      public final Object value;

      public MetaDataValue(PayloadKey type, Object value)
      {
         this.type = type;
         this.value = value;
      }

      public Object get()
         throws java.io.IOException, ClassNotFoundException
      {
         if (value instanceof MarshalledValue)
         {
            return ((MarshalledValue)value).get();
         }
         return value;
      }

   }

   public synchronized HashSet tags()
   {
      return new HashSet(metaData.keySet());
   }

   public synchronized HashMap tag(String name)
   {
      HashMap map = (HashMap)metaData.get(name);
      if (map == null) return null;
      return (HashMap)map.clone();
   }

   public synchronized boolean hasTag(String name)
   {
      return metaData.get(name) != null;
   }

   /**
    * Tag metadata to structure.  Use for tags with no attributes
    * i.e. @Singleton, etc...
    * @param group
    */
   public void tag(Object tag)
   {
      addMetaData(tag, EMPTY_TAG, new Object(), PayloadKey.TRANSIENT);
   }

   public void addMetaData(Object tag, Object attr, Object value)
   {
      addMetaData(tag, attr, value, PayloadKey.MARSHALLED);
   }

   public synchronized void addMetaData(Object tag, Object attr, Object value, PayloadKey type)
   {
      HashMap groupData = (HashMap)metaData.get(tag);
      if (groupData == null)
      {
         groupData = new HashMap();
         metaData.put(tag, groupData);
      }
      MetaDataValue val = new MetaDataValue(type, value);
      groupData.put(attr, val);
   }

   public synchronized Object getMetaData(Object tag, Object attr)
   {
      try
      {
         HashMap groupData = (HashMap)metaData.get(tag);
         if (groupData == null) return null;
         MetaDataValue val = (MetaDataValue)groupData.get(attr);
         if (val == null) return null;
         return val.get();
      }
      catch (IOException ioex)
      {
         throw new RuntimeException("failed on MarshalledValue", ioex);
      }
      catch (ClassNotFoundException ex)
      {
         throw new RuntimeException("failed on MarshalledValue", ex);
      }
   }

   public synchronized void removeMetaData(Object tag, Object attr)
   {
      HashMap groupData = (HashMap)metaData.get(tag);
      if (groupData != null)
      {
         groupData.remove(attr);
      }
   }

   public synchronized void removeGroupData(Object group)
   {
      metaData.remove(group);
   }

   public synchronized void clear()
   {
      metaData.clear();
   }

   /**
    * merges incoming data.  Incoming data overrides existing data
    */
   public synchronized void mergeIn(SimpleMetaData data)
   {
      Iterator it = data.metaData.keySet().iterator();
      while (it.hasNext())
      {
         Object tag = it.next();
         HashMap attrs = (HashMap)data.metaData.get(tag);
         HashMap map = (HashMap)metaData.get(tag);
         if (map == null)
         {
            map = new HashMap();
            this.metaData.put(tag, map);
         }
         map.putAll(attrs);
      }
   }

   public synchronized Object resolve(Invocation invocation, Object tag, Object attr)
   {
      return getMetaData(tag, attr);
   }

   public SimpleMetaData getAllMetaData(Invocation invocation)
   {
      return this;
   }

   public void writeExternal(java.io.ObjectOutput out)
      throws IOException
   {
      //System.out.println("******** marshalling metadata");
      Iterator it = metaData.keySet().iterator();
      while (it.hasNext())
      {
         Object group = it.next();
         HashMap map = (HashMap)metaData.get(group);
         //System.out.println("******** marshalling group " + group + " size = " + map.size());
         if (map != null && map.size() > 0)
         {
            boolean groupWritten = false;
            Iterator attrs = map.keySet().iterator();
            while (attrs.hasNext())
            {
               Object attr = attrs.next();
               //System.out.println("******** marshalling attr: " + group + "." + attr);
               MetaDataValue value = (MetaDataValue)map.get(attr);
               if (value.type == PayloadKey.TRANSIENT) continue;
               if (!groupWritten)
               {
                  groupWritten = true;
                  out.writeObject(group);
               }
               out.writeObject(attr);
               if (value.type == PayloadKey.AS_IS)
               {
                  out.writeObject(value.value);
               }
               else
               {
                  out.writeObject(new MarshalledValue(value.value));
               }
            }
            if (groupWritten) out.writeObject(null); // placeholder for end of attributes
         }
      }
      out.writeObject(null); // place holder for end of marshall
   }

   public void readExternal(java.io.ObjectInput in)
      throws IOException, ClassNotFoundException
   {
      //System.out.println("******** unmarshalling metadata");
      metaData = new HashMap();
      Object group;
      while ((group = in.readObject()) != null)
      {
         //System.out.println("******** unmarshalling group: " + group);
         HashMap map = new HashMap();
         metaData.put(group, map);
         Object attr;
         while ((attr = in.readObject()) != null)
         {
            //System.out.println("******** unmarshalling attr: " + group + "." + attr);
            Object obj = in.readObject();
            if (obj instanceof MarshalledValue)
            {
               map.put(attr, new MetaDataValue(PayloadKey.MARSHALLED, obj));
            }
            else
            {
               map.put(attr, new MetaDataValue(PayloadKey.AS_IS, obj));
            }
         }
      }
   }


   

}
