/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.aop.pointcut;

import javassist.CtClass;
import javassist.CtConstructor;
import javassist.CtField;
import javassist.CtMethod;
import javassist.NotFoundException;
import org.jboss.aop.ClassAdvisor;
import org.jboss.aop.pointcut.ast.ASTAllParameter;
import org.jboss.aop.pointcut.ast.ASTAnd;
import org.jboss.aop.pointcut.ast.ASTAttribute;
import org.jboss.aop.pointcut.ast.ASTBoolean;
import org.jboss.aop.pointcut.ast.ASTClass;
import org.jboss.aop.pointcut.ast.ASTComposite;
import org.jboss.aop.pointcut.ast.ASTConstructor;
import org.jboss.aop.pointcut.ast.ASTField;
import org.jboss.aop.pointcut.ast.ASTHas;
import org.jboss.aop.pointcut.ast.ASTHasField;
import org.jboss.aop.pointcut.ast.ASTMethod;
import org.jboss.aop.pointcut.ast.ASTNot;
import org.jboss.aop.pointcut.ast.ASTOr;
import org.jboss.aop.pointcut.ast.ASTParameter;
import org.jboss.aop.pointcut.ast.ASTStart;
import org.jboss.aop.pointcut.ast.ASTSub;
import org.jboss.aop.pointcut.ast.Node;
import org.jboss.aop.pointcut.ast.SimpleNode;
import org.jboss.aop.pointcut.ast.TypeExpressionParserVisitor;

/**
 * Takes a Javassist object and sees if the expression
 * pertains to that Javassist object (CtMethod, CtField, CtClass, CtConstructor)
 *
 * @author <a href="mailto:bill@jboss.org">Bill Burke</a>
 * @version $Revision: 1.2 $
 *
 **/
public class AnnotationMatcher implements TypeExpressionParserVisitor
{
   private ClassAdvisor advisor;
   private Object javassist;

   public AnnotationMatcher(ClassAdvisor advisor, Object javassist)
   {
      this.advisor = advisor;
      this.javassist = javassist;
   }

   public Object visit(ASTStart node, Object data)
   {
      return node.jjtGetChild(0).jjtAccept(this, data);
   }

   public Object visit(ASTBoolean node, Object data)
   {
      return node.jjtGetChild(0).jjtAccept(this, data);
   }

   public Object visit(ASTComposite node, Object data)
   {
      return node.jjtGetChild(0).jjtAccept(this, data);
   }

   public Object visit(ASTNot node, Object data)
   {

      Boolean bool = (Boolean) node.jjtGetChild(0).jjtAccept(this, data);
      boolean val = bool.booleanValue();
      return val ? Boolean.FALSE : Boolean.TRUE;
   }

   public Object visit(ASTSub node, Object data)
   {
      for (int i = 0; i < node.jjtGetNumChildren(); i++)
      {
         data = node.jjtGetChild(i).jjtAccept(this, data);
      }
      return data;
   }

   public Object visit(ASTAnd node, Object left)
   {
      Node andChild = node.jjtGetChild(0); // should only have one child
      boolean val = ((Boolean) left).booleanValue();
      return new Boolean(val && ((Boolean) andChild.jjtAccept(this, Boolean.FALSE)).booleanValue());
   }

   public Object visit(ASTOr node, Object left)
   {
      Node orChild = node.jjtGetChild(0); // should only have one child
      boolean val = ((Boolean) left).booleanValue();
      return new Boolean(val || ((Boolean) orChild.jjtAccept(this, Boolean.FALSE)).booleanValue());
   }


   public Object visit(SimpleNode node, Object data)
   {
      return null;
   }

   public CtClass getDeclaringClass()
   {
      if (javassist instanceof CtClass) return (CtClass) javassist;
      if (javassist instanceof CtMethod) return ((CtMethod) javassist).getDeclaringClass();
      if (javassist instanceof CtField) return ((CtField) javassist).getDeclaringClass();
      if (javassist instanceof CtConstructor) return ((CtConstructor) javassist).getDeclaringClass();
      throw new RuntimeException("Unknown!");
   }

   public Object visit(ASTHas node, Object data)
   {
      Node n = node.jjtGetChild(0);
      CtClass clazz = getDeclaringClass();
      if (n instanceof ASTMethod)
      {
         return new Boolean(Util.has(clazz, (ASTMethod) n, advisor));
      }
      else
      {
         return new Boolean(Util.has(clazz, (ASTConstructor) n, advisor));
      }
   }

   public Object visit(ASTHasField node, Object data)
   {
      CtClass clazz = getDeclaringClass();
      ASTField f = (ASTField) node.jjtGetChild(0);
      return new Boolean(Util.has(clazz, f, advisor));
   }

   public Object visit(ASTClass node, Object data)
   {
      if (!(javassist instanceof CtClass)) return Boolean.FALSE;
      CtClass clazz = (CtClass)javassist;
      return new Boolean(Util.matchesClassExpr(node.getClazz(), clazz, advisor));
   }

   public Object visit(ASTMethod node, Object data)
   {
      if (!(javassist instanceof CtMethod)) return Boolean.FALSE;
      MethodMatcher methodMatcher = null;
      try
      {
         methodMatcher = new MethodMatcher(advisor, (CtMethod)javassist, null);
      }
      catch (NotFoundException e)
      {
         throw new RuntimeException(e);  //To change body of catch statement use Options | File Templates.
      }
      return methodMatcher.matches(node);
   }

   public Object visit(ASTConstructor node, Object data)
   {
      if (!(javassist instanceof CtConstructor)) return Boolean.FALSE;
      ConstructorMatcher conMatcher = null;
      try
      {
         conMatcher = new ConstructorMatcher(advisor, (CtConstructor)javassist, null);
      }
      catch (NotFoundException e)
      {
         throw new RuntimeException(e);  //To change body of catch statement use Options | File Templates.
      }
      return conMatcher.matches(node);
   }

   public Object visit(ASTField node, Object data)
   {
      if (!(javassist instanceof CtField)) return Boolean.FALSE;
      FieldMatcher fieldMatcher = null;
      try
      {
         fieldMatcher = new FieldMatcher(advisor, (CtField)javassist, null);
      }
      catch (NotFoundException e)
      {
         throw new RuntimeException(e);  //To change body of catch statement use Options | File Templates.
      }
      return node.jjtAccept(fieldMatcher, null);
   }

   public Object visit(ASTAttribute node, Object data)
   {
      return Boolean.FALSE;
   }

   public Object visit(ASTParameter node, Object data)
   {
      return Boolean.FALSE;
   }

   public Object visit(ASTAllParameter node, Object data)
   {
      return Boolean.FALSE;
   }

}
