/***************************************
 *                                     *
 *  JBoss: The OpenSource J2EE WebOS   *
 *                                     *
 *  Distributable under LGPL license.  *
 *  See terms of license at gnu.org.   *
 *                                     *
 ***************************************/

package org.jboss.aop.proxy;
import javassist.ClassPool;
import javassist.CtClass;
import javassist.CtField;
import javassist.CtMethod;
import javassist.CtNewMethod;
import javassist.Modifier;
import javassist.SerialVersionUID;
import org.jboss.aop.AspectManager;
import org.jboss.aop.ClassAdvisor;
import org.jboss.util.collection.WeakValueHashMap;

import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.WeakHashMap;


/**
 *
 * @author <a href="mailto:bill@jboss.org">Bill Burke</a>
 * @version $Revision: 1.18 $
 */
public class ClassProxyFactory
{
   private static Object maplock = new Object();
   private static WeakValueHashMap classnameMap = new WeakValueHashMap();
   private static WeakHashMap proxyCache = new WeakHashMap();
   private static WeakHashMap methodMapCache = new WeakHashMap();

   public static ClassProxy newInstance(Class clazz) throws Exception
   {
      // Don't make a proxy of a proxy !
      if (ClassProxy.class.isAssignableFrom(clazz)) clazz = clazz.getSuperclass();

      Class proxyClass = null;
      synchronized (maplock)
      {
         proxyClass = (Class)proxyCache.get(clazz);
         if (proxyClass == null) 
         {
            proxyClass = generateProxy(clazz);
            classnameMap.put(clazz.getName(), proxyClass);
            proxyCache.put(clazz, proxyClass);
            HashMap map = methodMap(clazz);
            methodMapCache.put(proxyClass, map);
         }
      }
      ClassProxy proxy = (ClassProxy)proxyClass.newInstance();
      return proxy;
      
   }

   public static HashMap getMethodMap(String classname)
   {
      synchronized (maplock)
      {
         Class clazz = (Class)classnameMap.get(classname);
         if (clazz == null) return null;
         return (HashMap)methodMapCache.get(clazz);
      }
   }

   public static HashMap getMethodMap(Class clazz)
   {
      HashMap map = getMethodMap(clazz.getName());
      if (map != null) return map;
      try
      {
         return methodMap(clazz);
      }
      catch (Exception e)
      {
         throw new RuntimeException(e);  //To change body of catch statement use Options | File Templates.
      }
   }

   private static Class generateProxy(Class clazz) throws Exception
   {
      ClassPool pool = AspectManager.instance().findClassPool(clazz.getClassLoader());
      if (pool == null) throw new NullPointerException("Could not find ClassPool");
      
      CtClass superclass = pool.get(clazz.getName());
      CtClass template = pool.get("org.jboss.aop.proxy.ClassProxyTemplate");
      CtMethod writeEx = template.getDeclaredMethod("writeExternal");
      CtMethod readEx = template.getDeclaredMethod("readExternal");
      CtMethod body = template.getDeclaredMethod("_proxy$invoke");
      CtMethod getInstanceAdvisor = template.getDeclaredMethod("_getInstanceAdvisor");
      CtMethod setInstanceAdvisor = template.getDeclaredMethod("_setInstanceAdvisor");
      CtMethod dynamicInvoke = template.getDeclaredMethod("_dynamicInvoke");

      CtClass proxy = pool.makeClass("org.jboss.aop.proxy$" + clazz.getName() , superclass);
      proxy.addInterface(pool.get("org.jboss.aop.proxy.ClassProxy"));
      proxy.addInterface(pool.get("java.io.Externalizable"));
      proxy.addInterface(pool.get("org.jboss.aop.instrument.Untransformable"));

      CtClass type = pool.get("org.jboss.aop.ClassInstanceAdvisor");
      CtField field = new CtField(type, "_proxy$instance$advisor", proxy);
      field.setModifiers(
         Modifier.PRIVATE);
      proxy.addField(field, 
                     CtField.Initializer.byExpr("new org.jboss.aop.ClassInstanceAdvisor(this)"));
      proxy.addMethod(CtNewMethod.copy(writeEx, proxy, null));
      proxy.addMethod(CtNewMethod.copy(readEx, proxy, null));
      proxy.addMethod(CtNewMethod.copy(getInstanceAdvisor, proxy, null));
      proxy.addMethod(CtNewMethod.copy(setInstanceAdvisor, proxy, null));
      proxy.addMethod(CtNewMethod.copy(dynamicInvoke, proxy, null));
      
      

      HashMap allMethods = org.jboss.aop.util.JavassistMethodHashing.getMethodMap(superclass);
      Iterator it = allMethods.entrySet().iterator();
      while (it.hasNext())
      {
         Map.Entry entry = (Map.Entry)it.next();
         CtMethod m = (CtMethod)entry.getValue();
         Long key = (Long)entry.getKey();
         if (Modifier.isPublic(m.getModifiers())
             && !Modifier.isStatic(m.getModifiers()))
         {
            CtMethod newMethod = CtNewMethod.wrapped(m.getReturnType(), m.getName(), m.getParameterTypes(), 
                                                     m.getExceptionTypes(), body, CtMethod.ConstParameter.integer(key.longValue()), proxy);
            newMethod.setModifiers(m.getModifiers());
            proxy.addMethod(newMethod);
         }
      }
      SerialVersionUID.setSerialVersionUID(proxy);
      return proxy.toClass();
   }

   private static void populateMethodTables(HashMap advised, Class superclass)
      throws Exception
   {
      if (superclass == null) return;
      if (superclass.getName().equals("java.lang.Object")) return;

      populateMethodTables(advised, superclass.getSuperclass());
                        
      Method[] declaredMethods = superclass.getDeclaredMethods();
      for (int i = 0; i < declaredMethods.length; i++)
      {
         if (ClassAdvisor.isAdvisable(declaredMethods[i]))
         {
            long hash = org.jboss.aop.util.MethodHashing.methodHash(declaredMethods[i]);
            advised.put(new Long(hash), declaredMethods[i]);
         }
      }
   }

   public static HashMap methodMap(Class clazz)
      throws Exception
   {
      HashMap methods = new HashMap();
      populateMethodTables(methods, clazz);
      return methods;
   }

}
