/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.aop.standalone;

import javassist.bytecode.ClassFile;
import org.jboss.aop.AspectManager;
import org.jboss.aop.Deployment;
import org.jboss.aop.instrument.TransformationException;

import java.io.BufferedReader;
import java.io.DataInputStream;
import java.io.File;
import java.io.FileFilter;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.lang.reflect.Field;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.StringTokenizer;

/**
 * takes jar or class files and adds needed jboss bytecode
 *
 * @author <a href="mailto:bill@jboss.org">Bill Burke</a>
 * @version $Revision: 1.7.2.1 $
 */
public class Compiler
{
   private static class ClassFileFilter implements FileFilter
   {
      public boolean accept(File pathname)
      {
         return pathname.getName().endsWith(".class");
      }
   }

   private static class DirectoryFilter implements FileFilter
   {
      public boolean accept(File pathname)
      {
         return pathname.isDirectory();
      }
   }

   public boolean verbose = false;

   public boolean isJarFile(File src)
   {
      return (src.isFile()
              && (src.getName().toLowerCase().endsWith(".jar")
              || src.getName().toLowerCase().endsWith(".zip"))
              );
   }

   public static void main(String[] args) throws Exception
   {
      Compiler c = new Compiler();
      try
      {
         c.compile(args);
      }
      catch (Exception e)
      {
         if (c.verbose) throw e;

         if (e instanceof TransformationException)
         {
            System.exit(1);
         }
         throw e;
      }
   }

   public void usage()
   {
      System.err.println("Usage: aopc [-cp <classpath>] [-classpath <classpath>] [-report] [-verbose] [-aoppath <xml files>] <dir>+");
   }

   // Make public and static so that transformers can locate it to do work
   // transformers may generate class files and they need to determine
   // file locations and such.  This will also be used as a flag to tell
   // transformers whether they are in compile or load-time mode.
   public static URLClassLoader loader;

   public void compile(String[] args) throws Exception
   {
      if (args.length == 0)
      {
         usage();
         System.exit(1);
         return;
      }
      ArrayList paths = new ArrayList();
      ArrayList files = new ArrayList();
      boolean report = false;
      for (int i = 0; i < args.length; i++)
      {
         if (args[i].equals("-verbose"))
         {
            verbose = true;
            continue;
         }
         else if (args[i].equals("-report"))
         {
            report = true;
            continue;
         }
         else if (args[i].equals("-cp") || args[i].equals("-classpath"))
         {
            if (i + 1 > args.length - 1)
            {
               usage();
               System.exit(1);
               return;
            }
            i++;
            StringTokenizer tokenizer = new StringTokenizer(args[i], File.pathSeparator);
            while (tokenizer.hasMoreTokens())
            {
               String cpath = tokenizer.nextToken();
               File f = new File(cpath);
               paths.add(f.toURL());
            }
            continue;
         }
         else if (args[i].equals("-aoppath"))
         {
            System.setProperty("jboss.aop.path", args[++i]);
            continue;
         }
         else if (args[i].equals("-aopclasspath"))
         {
            System.setProperty("jboss.aop.class.path", args[++i]);
            continue;
         }
         else if (args[i].equals("--SOURCEPATH"))
         {
            addFilesFromSourcePathFile(files, args[++i]);
            continue;
         }
         File f = new File(args[i]).getCanonicalFile();
         files.add(f);
      }


      URL[] urls = (URL[]) paths.toArray(new URL[paths.size()]);
      loader = new URLClassLoader(urls, Thread.currentThread().getContextClassLoader());

      Thread.currentThread().setContextClassLoader(loader);

      Deployment.searchClasspath = true; // turn on dynamic finding of DDs
      AspectManager.verbose = verbose;
      AspectManager.instance();

      if (report)
      {
         for (int i = 0; i < files.size(); i++)
         {
            File f = (File) files.get(i);
            loadFile(f);
         }
         FileOutputStream reportFile = new FileOutputStream("aop-report.xml");
         reportFile.write(XmlReport.toXml().getBytes());
         reportFile.close();
      }
      else
      {
         for (int i = 0; i < files.size(); i++)
         {
            File f = (File) files.get(i);
            if (f.isDirectory())
            {
               compileDirectory(f);
            }
            else
            {
               compileFile(f);
            }
         }
      }
   }

   private void addFilesFromSourcePathFile(ArrayList files, String sourcePathFile)
   {
      BufferedReader reader = null;

      try
      {
         reader = new BufferedReader(new FileReader(new File(sourcePathFile).getCanonicalFile()));

         String fileName = reader.readLine();
         while (fileName != null)
         {
            files.add(new File(fileName).getCanonicalFile());
            fileName = reader.readLine();
         }
      }
      catch (Exception e)
      {
         try
         {
            reader.close();
         }
         catch (IOException e1)
         {
         }
         throw new RuntimeException(e);
      }
   }

   public void loadFile(File file) throws Exception
   {
      DataInputStream is = new DataInputStream(new FileInputStream(file));
      ClassFile cf = new ClassFile(is);
      is.close();
      Class clazz = loader.loadClass(cf.getName());
      if (org.jboss.aop.Advised.class.isAssignableFrom(clazz))
      {
         Field f = clazz.getDeclaredField("aop$classAdvisor$aop");
         f.setAccessible(true);
         Object o = f.get(null);

      }
   }

   public void compileFile(File file) throws Exception
   {
      DataInputStream is = new DataInputStream(new FileInputStream(file));
      ClassFile cf = new ClassFile(is);
      is.close();
      String className = cf.getName();
      //System.err.println("[classname] " + className);
      URL classUrl = loader.getResource(className.replace('.', '/') + ".class");
      if (classUrl == null)
      {
         System.out.println("[warning] Unable to find " + file + " within classpath.  Make sure all transforming classes are within classpath.");
         return;
      }
      if (!classUrl.equals(file.toURL()))
      {
         System.out.println("[warning] Trying to compile " + file + " and found it also within " + classUrl + " will not proceed. ");
         return;
      }
      byte[] bytes = AspectManager.instance().transform(loader, className, null, null, null);
      if (bytes == null)
      {
         if (verbose) System.out.println("[no comp needed] " + file);
         return;
      }
      FileOutputStream os = new FileOutputStream(file);
      os.write(bytes);
      os.close();
      if (verbose) System.out.println("[compiled] " + file);
   }

   public void compileDirectory(File base) throws Exception
   {
      File[] directories = base.listFiles(new DirectoryFilter());
      File[] classFiles = base.listFiles(new ClassFileFilter());
      for (int i = 0; i < classFiles.length; i++)
      {
         compileFile(classFiles[i]);
      }
      for (int i = 0; i < directories.length; i++)
      {
         compileDirectory(directories[i]);
      }
   }

}
