/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */

package org.jboss.aspects.remoting;

import org.jboss.aop.util.PayloadKey;
import org.jboss.ha.framework.interfaces.GenericClusteringException;
import org.jboss.ha.framework.interfaces.LoadBalancePolicy;
import org.jboss.remoting.ConnectionFailedException;
import org.jboss.remoting.InvokerLocator;

import java.util.ArrayList;
/**
 * Pick an invocation target 
 *
 * @author <a href="mailto:bill@jboss.org">Bill Burke</a>
 * @version $Revision: 1.2 $
 */
public class ClusterChooserInterceptor implements org.jboss.aop.advice.Interceptor, ClusterConstants, java.io.Serializable
{
   public static final ClusterChooserInterceptor singleton = new ClusterChooserInterceptor();

   public String getName() { return "ClusterChooserInterceptor"; }

   public Object invoke(org.jboss.aop.joinpoint.Invocation invocation)
      throws Throwable
   {
      LoadBalancePolicy lb = (LoadBalancePolicy)invocation.getMetaData(CLUSTERED_REMOTING, LOADBALANCE_POLICY);
      FamilyWrapper family = (FamilyWrapper)invocation.getMetaData(CLUSTERED_REMOTING, CLUSTER_FAMILY_WRAPPER);

      // we give the opportunity, to any server interceptor, to know if this a
      // first invocation to a node or if it is a failovered call
      //
      int failoverCounter = 0;
      String familyName = family.get().getFamilyName();
      invocation.getMetaData().addMetaData(CLUSTERED_REMOTING, CLUSTER_FAMILY, familyName, PayloadKey.AS_IS);
      InvokerLocator target = (InvokerLocator)lb.chooseTarget(family.get());
      Throwable lastException = null;
      while (target != null)
      {
         invocation.getMetaData().addMetaData(CLUSTERED_REMOTING, FAILOVER_COUNTER, new Integer(failoverCounter), PayloadKey.AS_IS);
         invocation.getMetaData().addMetaData(InvokeRemoteInterceptor.REMOTING, InvokeRemoteInterceptor.INVOKER_LOCATOR, target, PayloadKey.TRANSIENT);
         invocation.getMetaData().addMetaData(CLUSTERED_REMOTING, CLUSTER_VIEW_ID, new Long(family.get().getCurrentViewId()), PayloadKey.AS_IS);

         boolean definitivlyRemoveNodeOnFailure = true;
         lastException = null;
         try
         {
            Object rsp = invocation.invokeNext();
            ArrayList newReplicants = (ArrayList)invocation.getResponseAttachment("replicants");
            if (newReplicants != null)
            {
               long newViewId = ((Long)invocation.getResponseAttachment("viewId")).longValue();
               family.get().updateClusterInfo(newReplicants, newViewId);
            }
            return rsp;
         }
         catch (ConnectionFailedException ex)
         {
            lastException = ex;
         }
         catch (GenericClusteringException gce)
         {
            lastException = gce;
            // this is a generic clustering exception that contain the
            // completion status: usefull to determine if we are authorized
            // to re-issue a query to another node
            //               
            if (gce.getCompletionStatus () == GenericClusteringException.COMPLETED_NO)
            {
               // we don't want to remove the node from the list of failed 
               // node UNLESS there is a risk to indefinitively loop
               //
               if (family.get().getTargets().size() >= failoverCounter)
               {
                  if (!gce.isDefinitive())
                     definitivlyRemoveNodeOnFailure = false;                     
               }
            }
            else
            {
               throw new RuntimeException("Clustering exception thrown", gce);
            }
         }
         // If we reach here, this means that we must fail-over
         family.get().removeDeadTarget(target);
         if (!definitivlyRemoveNodeOnFailure)            
         {            
            family.get().resetView();
         }
         
         target = (InvokerLocator)lb.chooseTarget(family.get());
         if (target == null) 
         {
            if (lastException != null)
            {
               throw new RuntimeException("cluster invocation failed, last exception was: ", lastException);
            }
            else
            {
               throw new RuntimeException("cluster invocation failed");
            }
         }
         failoverCounter++;
      }
      // if we get here this means list was exhausted
      throw new RuntimeException("Unreachable?: Service unavailable.");
   }
}
