/***************************************
 *                                     *
 *  JBoss: The OpenSource J2EE WebOS   *
 *                                     *
 *  Distributable under LGPL license.  *
 *  See terms of license at gnu.org.   *
 *                                     *
 ***************************************/

package org.jboss.aspects.remoting;

import org.jboss.aop.Advised;
import org.jboss.aop.Dispatcher;
import org.jboss.aop.InstanceAdvised;
import org.jboss.aop.InstanceAdvisor;
import org.jboss.aop.proxy.ClassProxy;
import org.jboss.aop.proxy.ClassProxyFactory;
import org.jboss.aop.util.PayloadKey;
import org.jboss.ha.framework.interfaces.ClusteringTargetsRepository;
import org.jboss.ha.framework.interfaces.FamilyClusterInfo;
import org.jboss.ha.framework.interfaces.HAPartition;
import org.jboss.ha.framework.interfaces.LoadBalancePolicy;
import org.jboss.ha.framework.server.HATarget;
import org.jboss.logging.Logger;
import org.jboss.remoting.InvokerLocator;

import javax.naming.InitialContext;

/**
 *
 * @author <a href="mailto:bill@jboss.org">Bill Burke</a>
 * @version $Revision: 1.2 $
 */

public class ClusteredRemoting implements ClusterConstants
{
   private static final Logger log = Logger.getLogger(ClusteredRemoting.class);

   public static ClassProxy clusterRegisteredObject(Object objectId, String partitionName, LoadBalancePolicy lb, String uri)
      throws Exception
   {
      return clusterRegisteredObject(objectId, partitionName, lb, new InvokerLocator(uri));
   }
   public static ClassProxy clusterRegisteredObject(Object objectId, String partitionName, LoadBalancePolicy lb, InvokerLocator locator)
      throws Exception
   {
      Object obj = Dispatcher.singleton.getRegistered(objectId);
      if (obj == null) throw new NotRegisteredException(objectId.toString() + " is not registered");

      String proxyFamilyName = objectId.toString() + locator.getProtocol() + partitionName;
      HAPartition partition = (HAPartition)new InitialContext().lookup("/HAPartition/" + partitionName);
      HATarget target = new HATarget(partition, proxyFamilyName, locator, HATarget.ENABLE_INVOCATIONS);
      FamilyClusterInfo familyClusterInfo = ClusteringTargetsRepository.initTarget (proxyFamilyName, target.getReplicants());
      

      InstanceAdvisor advisor = null;
      Class clazz;
      if (obj instanceof Advised)
      {
         advisor = ((Advised)obj)._getInstanceAdvisor();
         clazz = obj.getClass();
      }
      else if (obj instanceof ClassProxy)
      {
         advisor = ((ClassProxy)obj)._getInstanceAdvisor();
         clazz = obj.getClass().getSuperclass();
      }
      else
      {
         clazz = obj.getClass();
         ClassProxy proxy = ClassProxyFactory.newInstance(obj.getClass());
         advisor = proxy._getInstanceAdvisor();
         advisor.insertInterceptor(new ReplicantsManagerInterceptor(proxyFamilyName, target));
         advisor.insertInterceptor(new ForwardingInterceptor(obj));
         Dispatcher.singleton.registerTarget(objectId, proxy);
      }

      ClassProxy proxy = ClassProxyFactory.newInstance(clazz);
      InstanceAdvisor proxyAdvisor = proxy._getInstanceAdvisor();
      proxyAdvisor.insertInterceptor(IsLocalInterceptor.singleton);
      proxyAdvisor.insertInterceptor(MergeMetaDataInterceptor.singleton);
      proxyAdvisor.insertInterceptor(ClusterChooserInterceptor.singleton);
      proxyAdvisor.insertInterceptor(InvokeRemoteInterceptor.singleton);

      proxyAdvisor.getMetaData().addMetaData(CLUSTERED_REMOTING, 
                                        CLUSTER_FAMILY_WRAPPER,
                                        new FamilyWrapper(proxyFamilyName,target.getReplicants()),
                                        PayloadKey.AS_IS);

      // LB policy should be AS_IS so that remoting framework downloads the classes
      proxyAdvisor.getMetaData().addMetaData(CLUSTERED_REMOTING, 
                                        LOADBALANCE_POLICY,
                                        lb,
                                        PayloadKey.AS_IS);

      proxyAdvisor.getMetaData().addMetaData(InvokeRemoteInterceptor.REMOTING, 
                                        InvokeRemoteInterceptor.SUBSYSTEM,
                                        "AOP",
                                        PayloadKey.AS_IS);
      
      proxyAdvisor.getMetaData().addMetaData(Dispatcher.DISPATCHER,
                                        Dispatcher.OID,
                                        objectId,
                                        PayloadKey.AS_IS);

      return proxy;
   }

   public static ClassProxy registerClusteredObject(Object objectId, Object obj, String partitionName, LoadBalancePolicy lb, String uri)
      throws Exception
   {
      return registerClusteredObject(objectId, obj, partitionName, lb, new InvokerLocator(uri));
   }
   public static ClassProxy registerClusteredObject(Object objectId, Object obj, String partitionName, LoadBalancePolicy lb, InvokerLocator locator)
      throws Exception
   {
      if (Dispatcher.singleton.isRegistered(objectId)) throw new AlreadyRegisteredException(objectId.toString() + " is already registered");


      String proxyFamilyName = objectId.toString() + locator.getProtocol() + partitionName;
      HAPartition partition = (HAPartition)new InitialContext().lookup("/HAPartition/" + partitionName);
      HATarget target = new HATarget(partition, proxyFamilyName, locator, HATarget.ENABLE_INVOCATIONS);
      FamilyClusterInfo familyClusterInfo = ClusteringTargetsRepository.initTarget (proxyFamilyName, target.getReplicants());
      

      Object registeredObject = null;
      InstanceAdvisor advisor = null;
      if (obj instanceof Advised)
      {
         registeredObject = obj;
         advisor = ((Advised)obj)._getInstanceAdvisor();
      }
      else
      {
         ClassProxy proxy = ClassProxyFactory.newInstance(obj.getClass());
         advisor = proxy._getInstanceAdvisor();
         advisor.insertInterceptor(new ReplicantsManagerInterceptor(proxyFamilyName, target));
         advisor.insertInterceptor(new ForwardingInterceptor(obj));
         registeredObject = proxy;
      }
      advisor.getMetaData().addMetaData(CLUSTERED_REMOTING,
                                   HA_TARGET,
                                   target);
      Dispatcher.singleton.registerTarget(objectId, registeredObject);

      ClassProxy proxy = ClassProxyFactory.newInstance(obj.getClass());
      InstanceAdvisor proxyAdvisor = proxy._getInstanceAdvisor();
      proxyAdvisor.insertInterceptor(IsLocalInterceptor.singleton);
      proxyAdvisor.insertInterceptor(MergeMetaDataInterceptor.singleton);
      proxyAdvisor.insertInterceptor(ClusterChooserInterceptor.singleton);
      proxyAdvisor.insertInterceptor(InvokeRemoteInterceptor.singleton);

      proxyAdvisor.getMetaData().addMetaData(CLUSTERED_REMOTING, 
                                        CLUSTER_FAMILY_WRAPPER,
                                        new FamilyWrapper(proxyFamilyName,target.getReplicants()),
                                        PayloadKey.AS_IS);

      // LB policy should be AS_IS so that remoting framework downloads the classes
      proxyAdvisor.getMetaData().addMetaData(CLUSTERED_REMOTING, 
                                        LOADBALANCE_POLICY,
                                        lb,
                                        PayloadKey.AS_IS);

      proxyAdvisor.getMetaData().addMetaData(InvokeRemoteInterceptor.REMOTING, 
                                        InvokeRemoteInterceptor.SUBSYSTEM,
                                        "AOP",
                                        PayloadKey.AS_IS);
      
      proxyAdvisor.getMetaData().addMetaData(Dispatcher.DISPATCHER,
                                        Dispatcher.OID,
                                        objectId,
                                        PayloadKey.AS_IS);

      return proxy;
   }

   public static void unregisterClusteredObject(Object object)
   {
      try
      {
         ClassProxy proxy = (ClassProxy) object;
         InstanceAdvisor advisor = proxy._getInstanceAdvisor();

         String oid = (String) advisor.getMetaData().getMetaData(Dispatcher.DISPATCHER, Dispatcher.OID);
         InstanceAdvised registeredObject = (InstanceAdvised) Dispatcher.singleton.getRegistered(oid);
         if (registeredObject == null) throw new NotRegisteredException(oid.toString() + " is not registered");
         Dispatcher.singleton.unregisterTarget(oid);

         advisor = registeredObject._getInstanceAdvisor();
         HATarget target = (HATarget) advisor.getMetaData().getMetaData(CLUSTERED_REMOTING, HA_TARGET);
         target.destroy();
      }
      catch (Exception ignored)
      {
         log.trace("Ignored exception unregistering a clustered object", ignored);
      }
   }
}
