/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.aspects.tx;

import org.jboss.aop.advice.Interceptor;
import org.jboss.aop.joinpoint.Invocation;
import org.jboss.logging.Logger;
import org.jboss.util.deadlock.ApplicationDeadlockException;

import javax.transaction.Transaction;
import javax.transaction.TransactionManager;
import java.util.Random;

/**
 *  This interceptor handles transactions for AOP
 *
 *  @author <a href="mailto:d_jencks@users.sourceforge.net">David Jencks</a>
 *  @author <a href="mailto:bill@jboss.org">Bill Burke</a>
 *  @version $Revision: 1.4 $
 */
public class TxInterceptor implements Interceptor
{
   public static int MAX_RETRIES = 5;
   public static Random random = new Random();

   /**
    * Logging instance
    */
   protected Logger log = Logger.getLogger(this.getClass());

   private TransactionManager tm;

   public TxInterceptor(TransactionManager tm)
   {
      this.tm = tm;
   }

   public String getName()
   {
      return "TxInterceptor";
   }


   /**
    *  This method does calls internalInvoke up to the specified count of retries.
    *
    * @todo put the retry logic in a separate interceptor
    */
   public Object invoke(Invocation invocation) throws Throwable
   {
      Transaction oldTransaction = tm.getTransaction();
      for (int i = 0; i < MAX_RETRIES; i++)
      {
         try
         {
            return internalInvoke(invocation);
         }
         catch (Exception ex)
         {
            ApplicationDeadlockException deadlock = ApplicationDeadlockException.isADE(ex);
            if (deadlock != null)
            {
               if (!deadlock.retryable() ||
                       oldTransaction != null ||
                       i + 1 >= MAX_RETRIES)
               {
                  throw deadlock;
               }
               log.warn(deadlock.getMessage() + " retrying " + (i + 1));

               Thread.sleep(random.nextInt(1 + i), random.nextInt(1000));
            }
            else
            {
               throw ex;
            }
         }
      }
      throw new RuntimeException("Unreachable");
   }

   /**
    * The <code>internalInvoke</code> method chooses the TxSupport
    * subclass depending on the methods transaction support, then
    * delegates the server side tx handing to it.
    *
    * @param invocation an <code>Invocation</code> value
    * @return an <code>Object</code> value
    * @exception Exception if an error occurs
    */
   protected Object internalInvoke(Invocation invocation) throws Throwable
   {
      String txtype = resolveTxType(invocation);
      TxSupport txSupport = null;
      if (txtype == null)
      {
         txSupport = TxSupport.DEFAULT;
      }
      else
      {
         txSupport = TxSupport.byName(txtype);
      }
      if (txSupport == null)
      {
         txSupport = TxSupport.DEFAULT;
      } // end of if ()

      return txSupport.serverInvoke(invocation, tm);
   }

   protected String resolveTxType(Invocation invocation)
   {
      String txtype = (String) invocation.getMetaData("transaction", "trans-attribute");
      if (txtype == null)
      {
         Tx tx = (Tx)invocation.resolveAnnotation(Tx.class);
         if (tx == null)
         {
            tx = (Tx)invocation.resolveClassAnnotation(Tx.class);
         }
         if (tx != null)
         {
            txtype = tx.value().name();
         }
      }
      return txtype;
   }

}
