/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 *
 */

package org.jboss.aspects.tx;

import org.jboss.aop.joinpoint.Invocation;
import org.jboss.logging.Logger;
import org.jboss.util.UnreachableStatementException;

import javax.transaction.HeuristicMixedException;
import javax.transaction.HeuristicRollbackException;
import javax.transaction.RollbackException;
import javax.transaction.Status;
import javax.transaction.SystemException;
import javax.transaction.Transaction;
import javax.transaction.TransactionManager;
import javax.transaction.TransactionRolledbackException;
import java.io.ObjectStreamException;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.StringTokenizer;


/**
 * TxSupport.java encapsulates the transaction handling possibilities
 * from the ejb spec.  The Tx interceptors call the clientInvoke and
 * serverInvoke methods on the subclass determined by the method's
 * transaction support.
 *
 *
 * Created: Sun Feb  2 23:25:09 2003
 *
 * @author <a href="mailto:d_jencks@users.sourceforge.net">David Jencks</a>
 * @author <a href="mailto:bill@jboss.org">Bill Burke</a>
 * @version
 */

public abstract class TxSupport
        implements Serializable
{

   /**
    * The variable <code>values</code> has the acceptable TxSupport
    * instances to support the readResolve method.
    *
    */
   private static final List values = new ArrayList();

   protected final static Logger log = Logger.getLogger(TxSupport.class);

   public final static TxSupport NEVER = new Never("Never");
   public final static TxSupport NOT_SUPPORTED = new NotSupported("NotSupported");
   public final static TxSupport SUPPORTS = new Supports("Supports");
   public final static TxSupport REQUIRED = new Required("Required");
   public final static TxSupport REQUIRES_NEW = new RequiresNew("RequiresNew");
   public final static TxSupport MANDATORY = new Mandatory("Mandatory");

   public final static TxSupport DEFAULT = REQUIRED;//??

   private static int nextOrdinal = 0;
   private final int ordinal = nextOrdinal++;
   private final transient String name;

   private static final HashMap nameMap = new HashMap();

   static
   {
      nameMap.put("NEVER", NEVER);
      nameMap.put("NOTSUPPORTED", NOT_SUPPORTED);
      nameMap.put("SUPPORTS", SUPPORTS);
      nameMap.put("REQUIRED", REQUIRED);
      nameMap.put("REQUIRESNEW", REQUIRES_NEW);
      nameMap.put("MANDATORY", MANDATORY);
   }

   private TxSupport(String name)
   {
      this.name = name;
      values.add(this);
   }

   public static TxSupport byName(String name)
   {
      TxSupport val = (TxSupport) nameMap.get(name.trim().toUpperCase());
      if (val == null) throw new IllegalArgumentException("Unknown TxType: " + name);
      return val;
   }

   public String toString()
   {
      return name;
   }

   protected Object internalReadResolve() throws ObjectStreamException
   {
      return values.get(ordinal);
   }

   public abstract Object serverInvoke(org.jboss.aop.joinpoint.Invocation invocation, TransactionManager tm) throws Throwable;


   /**
    * The <code>invokeInNoTx</code> method implements the behavior in
    * the ejb 2.1 spec section 18.3 table 15 lines 5 and 6, in the
    * case where there is no transaction.
    *
    * @param invocation an <code>Invocation</code> value
    * @return an <code>Object</code> value
    * @exception Exception if an error occurs
    */
   protected Object invokeInNoTx(org.jboss.aop.joinpoint.Invocation invocation) throws Throwable
   {
      return invocation.invokeNext();
   }

   /**
    * The <code>invokeInOurTx</code> method implements the behavior
    * described in the ejb 2.1 spec, section 18.3, table 15, lines 3
    * and 4.
    *
    * @param invocation an <code>Invocation</code> value
    * @param tm a <code>TransactionManager</code> value
    * @return an <code>Object</code> value
    * @exception Exception if an error occurs
    */
   protected Object invokeInOurTx(org.jboss.aop.joinpoint.Invocation invocation, TransactionManager tm) throws Throwable
   {
      tm.begin();
      Transaction tx = tm.getTransaction();
      try
      {
         try
         {
            return invocation.invokeNext();
         }
         catch (Throwable t)
         {
            // if this is an ApplicationException, just rethrow it
            rethrowApplicationException(invocation, t);
            setRollbackOnly(tx);
            throw t;
         } // end of try-catch
      }
      finally
      {
         endTransaction(tm, tx);
      } // end of try-finally
   }

   /**
    * The <code>invokeInCallerTx</code> method implements the behavior
    * described in the ejb 2.1 spec, section 18.3, table 15, lines 1
    * and 2.
    *
    * @param invocation an <code>Invocation</code> value
    * @param tx a <code>Transaction</code> value
    * @return an <code>Object</code> value
    * @exception Exception if an error occurs
    */
   protected Object invokeInCallerTx(org.jboss.aop.joinpoint.Invocation invocation, Transaction tx) throws Throwable
   {
      try
      {
         return invocation.invokeNext();
      }
      catch (Throwable t)
      {
         // if this is an ApplicationException, just rethrow it
         rethrowApplicationException(invocation, t);
         setRollbackOnly(tx);
         //otherwise reformat if necessary and rethrow.
         rethrowAsTxRolledbackException(t);
         throw new UnreachableStatementException();
      } // end of try-catch
   }

   /**
    * The <code>endTransaction</code> method ends a transaction and
    * translates any exceptions into
    * TransactionRolledBack[Local]Exception or SystemException.
    *
    * @param invocation an <code>Invocation</code> value
    * @param tm a <code>TransactionManager</code> value
    * @param tx a <code>Transaction</code> value
    * @exception TransactionRolledbackException if an error occurs
    * @exception SystemException if an error occurs
    */
   protected void endTransaction(TransactionManager tm, Transaction tx)
           throws TransactionRolledbackException, SystemException
   {
      if (tx != tm.getTransaction())
      {
         throw new IllegalStateException("Wrong tx on thread: expected " + tx + ", actual " + tm.getTransaction());
      } // end of if ()

      try
      {
         if (tx.getStatus() == Status.STATUS_MARKED_ROLLBACK)
         {
            tx.rollback();
         }
         else
         {
            // Commit tx
            // This will happen if
            // a) everything goes well
            // b) app. exception was thrown
            tx.commit();
         }
      }
      catch (RollbackException e)
      {
         throw new TransactionRolledbackException(e.getMessage());
      }
      catch (HeuristicMixedException e)
      {
         throw new TransactionRolledbackException(e.getMessage());
      }
      catch (HeuristicRollbackException e)
      {
         throw new TransactionRolledbackException(e.getMessage());
      }
      catch (SystemException e)
      {
         throw new TransactionRolledbackException(e.getMessage());
      } // end of try-catch
   }

   /**
    * The <code>rethrowAsTxRolledbackException</code> method
    * translates the supplied exception into an appropriate
    * TransactionRolledback[Local]Exception, unwrapping as
    * appropriate.
    *
    * @param type an <code>InvocationType</code> value
    * @param e a <code>Throwable</code> value
    * @exception TransactionRolledbackException if an error occurs
    */
   protected void rethrowAsTxRolledbackException(Throwable cause)
           throws TransactionRolledbackException
   {
      if (cause instanceof TransactionRolledbackException)
      {
         throw (TransactionRolledbackException) cause;
      }
      else
      {
         TransactionRolledbackException ex =
                 new TransactionRolledbackException(cause.getMessage());
         ex.detail = cause;
         throw ex;
      }
   }

   /**
    * The <code>setRollbackOnly</code> method calls setRollbackOnly()
    * on the invocation's transaction and logs any exceptions than may
    * occur.
    *
    * @param invocation an <code>Invocation</code> value
    */
   protected void setRollbackOnly(Transaction tx)
   {
      try
      {
         tx.setRollbackOnly();
      }
      catch (SystemException ex)
      {
         log.error("SystemException while setting transaction " +
                 "for rollback only", ex);
      }
      catch (IllegalStateException ex)
      {
         log.error("IllegalStateException while setting transaction " +
                 "for rollback only", ex);
      }
   }

   /**
    * The <code>rethrowApplicationException</code> method determines
    * if the supplied Throwable is an application exception and
    * rethrows it if it is.
    *
    * @param e a <code>Throwable</code> value
    * @exception Exception if an error occurs
    */
   protected void rethrowApplicationException(Invocation inv, Throwable e) throws Throwable
   {
      Object applicationExceptions = inv.getMetaData("transaction", "application-exceptions");
      if (applicationExceptions == null) return;
      Class[] applicationExceptionsList;
      if (applicationExceptions instanceof String)
      {
         ArrayList tmpList = new ArrayList();
         String aes = (String)applicationExceptions;
         aes = aes.trim();
         StringTokenizer tokenizer = new StringTokenizer(aes, ",");
         while (tokenizer.hasMoreTokens())
         {
            String token = tokenizer.nextToken().trim();
            Class excClass = Thread.currentThread().getContextClassLoader().loadClass(token);
            tmpList.add(excClass);
         }
         applicationExceptionsList = (Class[])tmpList.toArray(new Class[tmpList.size()]);
      }
      else
      {
         applicationExceptionsList = (Class[])applicationExceptions;
      }
      for (int i = 0; i < applicationExceptionsList.length; i++)
      {
         if (applicationExceptionsList[i].isInstance(e)) throw e;
      }

   }

   public final static class Never extends TxSupport
   {
      private Never(String name)
      {
         super(name);
      }

      private Object readResolve() throws ObjectStreamException
      {
         return internalReadResolve();
      }

      public Object serverInvoke(org.jboss.aop.joinpoint.Invocation invocation,
                                 TransactionManager tm)
              throws Throwable
      {
         if (tm.getTransaction() != null)
         {
            throw new IllegalStateException("Transaction present on server in Never call");
         } // end of if ()
         return invokeInNoTx(invocation);
      }


   }

   public final static class NotSupported extends TxSupport
   {
      private NotSupported(String name)
      {
         super(name);
      }


      private Object readResolve() throws ObjectStreamException
      {
         return internalReadResolve();
      }

      public Object serverInvoke(org.jboss.aop.joinpoint.Invocation invocation,
                                 TransactionManager tm)
              throws Throwable
      {
         Transaction tx = tm.getTransaction();
         if (tx != null)
         {
            tm.suspend();
            try
            {
               return invokeInNoTx(invocation);
            }
            finally
            {
               tm.resume(tx);
            } // end of try-finally

         } // end of if ()
         else
         {
            return invokeInNoTx(invocation);
         } // end of else
      }

   }

   public final static class Supports extends TxSupport
   {
      private Supports(String name)
      {
         super(name);
      }


      private Object readResolve() throws ObjectStreamException
      {
         return internalReadResolve();
      }

      public Object serverInvoke(Invocation invocation,
                                 TransactionManager tm)
              throws Throwable
      {
         if (tm.getTransaction() == null)
         {
            return invokeInNoTx(invocation);
         } // end of if ()
         else
         {
            return invokeInCallerTx(invocation, tm.getTransaction());
         } // end of else
      }

   }

   public final static class Required extends TxSupport
   {
      private Required(String name)
      {
         super(name);
      }

      private Object readResolve() throws ObjectStreamException
      {
         return internalReadResolve();
      }

      public Object serverInvoke(Invocation invocation,
                                 TransactionManager tm)
              throws Throwable
      {
         Transaction tx = tm.getTransaction();
         if (tx == null)
         {
            return invokeInOurTx(invocation, tm);
         } // end of if ()

         else
         {
            return invokeInCallerTx(invocation, tx);
         } // end of else
      }

   }

   public final static class RequiresNew extends TxSupport
   {
      private RequiresNew(String name)
      {
         super(name);
      }

      private Object readResolve() throws ObjectStreamException
      {
         return internalReadResolve();
      }

      public Object serverInvoke(Invocation invocation,
                                 TransactionManager tm)
              throws Throwable
      {
         Transaction tx = tm.getTransaction();
         if (tx != null)
         {
            tm.suspend();
            try
            {
               return invokeInOurTx(invocation, tm);
            }
            finally
            {
               tm.resume(tx);
            } // end of try-finally
         } // end of if ()
         else
         {
            return invokeInOurTx(invocation, tm);
         } // end of else
      }

   }

   public final static class Mandatory extends TxSupport
   {
      private Mandatory(String name)
      {
         super(name);
      }

      private Object readResolve() throws ObjectStreamException
      {
         return internalReadResolve();
      }

      public Object serverInvoke(Invocation invocation,
                                 TransactionManager tm)
              throws Throwable
      {
         Transaction tx = tm.getTransaction();
         if (tx == null)
         {
            throw new RuntimeException("Transaction required");
         } // end of if ()
         return invokeInCallerTx(invocation, tx);
      }

   }
}// TxSupport
