package org.jboss.security.jacc;

import java.security.CodeSource;
import java.security.Permission;
import java.security.PermissionCollection;
import java.security.Policy;
import java.security.ProtectionDomain;
import java.util.HashMap;
import java.util.Iterator;
import javax.security.jacc.EJBMethodPermission;
import javax.security.jacc.EJBRoleRefPermission;
import javax.security.jacc.PolicyConfiguration;
import javax.security.jacc.PolicyContext;
import javax.security.jacc.PolicyContextException;
import javax.security.jacc.WebResourcePermission;
import javax.security.jacc.WebRoleRefPermission;
import javax.security.jacc.WebUserDataPermission;

import org.jboss.logging.Logger;

/** A JAAC Policy provider implementation that delegates any non-JACC permissions
 * to the java.security.Policy either passed in to the ctor, or the pre
 * existing Policy.getPolicy value. 
 * 
 * @author Scott.Stark@jboss.org
 * @version $Revision: 1.5.4.1 $
 */ 
public class DelegatingPolicy extends Policy
{
   private static Logger log = Logger.getLogger(DelegatingPolicy.class);
   private static DelegatingPolicy instance;

   /** The system Policy we delegate non-JACC checks to */
   private Policy delegate;
   /** HashMap<String, ContextPolicy> for the JACC context IDs */
   private HashMap contextPolicies = new HashMap();
   /** Flag indicating if our category is at trace level for logging */
   private boolean trace;
   /** The Policy proxy returned via the PolicyProxy attribute */
   private PolicyProxy policyProxy = new PolicyProxy(this);

   public synchronized static DelegatingPolicy getInstance()
   {
      if( instance == null )
      {
         instance = new DelegatingPolicy();
      }
      return instance;
   }

   public DelegatingPolicy()
   {
      this(null);
   }
   public DelegatingPolicy(Policy delegate)
   {
      if( delegate == null )
         delegate = Policy.getPolicy();
      this.delegate = delegate;
      this.trace = log.isTraceEnabled();
      if( instance == null )
         instance = this;
   }

   public PermissionCollection getPermissions(ProtectionDomain domain)
   {
      return super.getPermissions(domain);
   }

   public boolean implies(ProtectionDomain domain, Permission permission)
   {
      boolean isJaccPermission = permission instanceof EJBMethodPermission
         || permission instanceof EJBRoleRefPermission
         || permission instanceof WebResourcePermission
         || permission instanceof WebRoleRefPermission
         || permission instanceof WebUserDataPermission;
      if( trace )
      {
         log.trace("implies, domain="+domain+", permission="+permission
            +", isJaccPermission="+isJaccPermission);
      }
      boolean implied = false;
      if( isJaccPermission == false )
      {
         // Let the delegate policy handle the check
         implied = delegate.implies(domain, permission);
      }
      else
      {
         String contextID = PolicyContext.getContextID();
         ContextPolicy contextPolicy = (ContextPolicy) contextPolicies.get(contextID);
         if( contextPolicy != null )
            implied = contextPolicy.implies(domain, permission);
         else if( trace )
            log.trace("No PolicyContext found for contextID="+contextID);
      }
      if( trace )
      {
         log.trace("implied="+implied);
      }
      return implied;
   }

   public PermissionCollection getPermissions(CodeSource cs)
   {
      PermissionCollection pc = null;
      String contextID = PolicyContext.getContextID();
      if( contextID == null )
      {
         pc = delegate.getPermissions(cs);
      }
      else
      {
         ContextPolicy policy = (ContextPolicy) contextPolicies.get(contextID);
         if( policy != null )
         {
            pc = policy.getPermissions();
         }
         else
         {
            pc = delegate.getPermissions(cs);
         }
      }
      return pc;
   }

   /**
    * 
    */ 
   public void refresh()
   {
   }

   /**
    * 
    * @return A proxy for our Policy interface
    */ 
   public Policy getPolicyProxy()
   {
      return policyProxy;
   }

   // Policy configuration methods used by the PolicyConfiguration impl

   /**
    * Access the current ContextPolicy instances
    * @return Map<String, ContextPolicy> of the contextID to policy mappings
    */ 
   public String listContextPolicies()
   {
      HashMap copy = new HashMap(contextPolicies);
      StringBuffer tmp = new StringBuffer("<ContextPolicies>");
      Iterator iter = copy.keySet().iterator();
      while( iter.hasNext() )
      {
         String contextID = (String) iter.next();
         ContextPolicy cp = (ContextPolicy) contextPolicies.get(contextID);
         tmp.append(cp);
         tmp.append('\n');
      }
      tmp.append("</ContextPolicies>");
      return tmp.toString();
   }

   synchronized ContextPolicy getContextPolicy(String contextID)
      throws PolicyContextException
   {
      ContextPolicy policy = (ContextPolicy) contextPolicies.get(contextID);
      if( policy == null )
         throw new PolicyContextException("No ContextPolicy exists for contextID="+contextID);
      return policy;
   }
   /** Create or update a ContextPolicy for contextID.
    * @param contextID
    * @param remove
    * @throws PolicyContextException
    */ 
   synchronized void initPolicyConfiguration(String contextID, boolean remove)
      throws PolicyContextException
   {
      ContextPolicy policy = (ContextPolicy) contextPolicies.get(contextID);
      if( policy == null )
      {
         policy = new ContextPolicy(contextID);
         contextPolicies.put(contextID, policy);
      }
      if( remove == true )
         policy.clear();
   }

   void addToExcludedPolicy(String contextID, Permission permission)
      throws PolicyContextException
   {
      ContextPolicy policy = getContextPolicy(contextID);
      policy.addToExcludedPolicy(permission);
   }

   void addToExcludedPolicy(String contextID, PermissionCollection permissions)
      throws PolicyContextException
   {
      ContextPolicy policy = getContextPolicy(contextID);
      policy.addToExcludedPolicy(permissions);      
   }

   void addToRole(String contextID, String roleName, Permission permission)
      throws PolicyContextException
   {
      ContextPolicy policy = getContextPolicy(contextID);
      policy.addToRole(roleName, permission);
   }

   void addToRole(String contextID, String roleName, PermissionCollection permissions)
      throws PolicyContextException
   {
      ContextPolicy policy = getContextPolicy(contextID);
      policy.addToRole(roleName, permissions);
   }

   void addToUncheckedPolicy(String contextID, Permission permission)
      throws PolicyContextException
   {
      ContextPolicy policy = getContextPolicy(contextID);
      policy.addToUncheckedPolicy(permission);
   }

   void addToUncheckedPolicy(String contextID, PermissionCollection permissions)
      throws PolicyContextException
   {
      ContextPolicy policy = getContextPolicy(contextID);
      policy.addToUncheckedPolicy(permissions);
   }

   void linkConfiguration(String contextID, PolicyConfiguration link)
      throws PolicyContextException
   {
      ContextPolicy policy = getContextPolicy(contextID);
      ContextPolicy linkPolicy = getContextPolicy(link.getContextID());
      policy.linkConfiguration(linkPolicy);
   }

   public void commit(String contextID)
      throws PolicyContextException
   {
      ContextPolicy policy = getContextPolicy(contextID);
      policy.commit();
   }

   public void delete(String contextID)
      throws PolicyContextException
   {
      ContextPolicy policy = getContextPolicy(contextID);
      policy.delete();
      synchronized( contextPolicies )
      {
         contextPolicies.remove(contextID);
      }
   }

   void removeExcludedPolicy(String contextID)
      throws PolicyContextException
   {
      ContextPolicy policy = getContextPolicy(contextID);
      policy.removeExcludedPolicy();
   }

   void removeRole(String contextID, String roleName)
      throws PolicyContextException
   {
      ContextPolicy policy = getContextPolicy(contextID);
      policy.removeRole(roleName);
   }

   void removeUncheckedPolicy(String contextID)
      throws PolicyContextException
   {
      ContextPolicy policy = getContextPolicy(contextID);
      policy.removeUncheckedPolicy();
   }

   private static class PolicyProxy extends Policy
   {
      private Policy delegate;
      PolicyProxy(Policy delegate)
      {
         this.delegate = delegate;
      }

      public void refresh()
      {
         delegate.refresh();
      }

      public PermissionCollection getPermissions(CodeSource codesource)
      {
         return delegate.getPermissions(codesource);
      }

      public boolean implies(ProtectionDomain domain, Permission permission)
      {
         return delegate.implies(domain, permission);
      }

      public PermissionCollection getPermissions(ProtectionDomain domain)
      {
         return delegate.getPermissions(domain);
      }
   }
}
