/*
 * JBoss, the OpenSource J2EE webOS
 * 
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.test.messagedriven.support;

import java.util.ArrayList;

import javax.management.ObjectName;

import org.jboss.naming.Util;
import org.jboss.test.JBossTestCase;
import org.jboss.test.classloader.circularity.support.Base;
import org.jboss.test.messagedriven.beans.TestMessageDrivenManagementMBean;

/**
 * Basic tests of message driven beans 
 *
 * @author <a href="mailto:adrian@jboss.com>Adrian Brock</a>
 * @version <tt>$Revision: 1.4</tt>
 */
public abstract class BasicMessageDrivenUnitTest extends JBossTestCase
{
   protected static final long WAIT_TIME = 5000L;
   protected static final long REPEATED_WAIT = 4;

   protected BasicTest test;
   
   protected Thread thread;
   protected boolean running = false;
   
   protected ArrayList messages = new ArrayList(); 
   
   public BasicMessageDrivenUnitTest(String name)
   {
      super(name);
      test = setupTest();
   }
   
   public abstract BasicTest setupTest();

   public abstract Operation[] getOperations() throws Exception;
   
   public void testRun() throws Exception
   {
      test.run(this);
   }

   protected void setUp() throws Exception
   {
      if ("testServerFound".equals(getName()))
         return;
      getLog().info("Setting up: " + test);
      clearMessages(test.getJMXDestination());
      clearMessages(test.getDLQJMXDestination());
      deploy(test.getDeployment());
      startReceiverThread();
      getLog().info("=========== Setup: " + test);
   }

   protected void tearDown()
   {
      if ("testServerFound".equals(getName()))
         return;
      getLog().info("=========== Tearing down: " + test);
      stopReceiverThread();
      try
      {
         undeploy(test.getDeployment());
      }
      catch (Throwable t)
      {
         getLog().error("Error undeploying: " + test.getDeployment(), t);
      }
      try
      {
         clearMessages(test.getJMXDestination());
      }
      catch (Throwable t)
      {
         getLog().error("Error clearing messages: " + test.getJMXDestination(), t);
      }
      try
      {
         clearMessages(test.getDLQJMXDestination());
      }
      catch (Throwable t)
      {
         getLog().error("Error clearing messages: " + test.getDLQJMXDestination(), t);
      }
      getLog().info("Tore down: " + test);
   }
   
   protected void clearMessages(ObjectName name) throws Exception
   {
      if (name != null)
      {
         getLog().info("Clearing messages " + name);
         getServer().invoke(name, "removeAllMessages", new Object[0], new String[0]);
      }
   }
   
   protected void waitMessages(int expected) throws Exception
   {
      synchronized (this)
      {
         for (int i = 0; i < REPEATED_WAIT && messages.size() < expected; ++i)
         {
            wait(WAIT_TIME);
         }
      }
   }
   
   protected ArrayList getMessages() throws Exception
   {
      synchronized (this)
      {
         return new ArrayList(messages);
      }
   }
   
   protected void startReceiverThread()
   {
      synchronized (this)
      {
         thread = new Thread(new ReceiverRunnable(), test.toString());
         thread.start();
         running = true;
      }
   }
   
   protected void stopReceiverThread()
   {
      synchronized (this)
      {
         running = false;
         while (thread != null)
         {
            try
            {
               this.wait();
            }
            catch (Throwable t)
            {
               getLog().error("Error waiting for receiver thread to stop " + thread, t);
            }
         }
      }
   }

   protected Object getAttribute(ObjectName name, String attribute) throws Exception
   {
      return getServer().getAttribute(name, attribute);
   }
   
   protected Object lookup(String jndiName, Class clazz) throws Exception
   {
      return Util.lookup(getInitialContext(), jndiName, clazz);
   }
   
   public class ReceiverRunnable implements Runnable
   {
      public void run()
      {
         try
         {
            while (true)
            {
               ArrayList result = (ArrayList) getAttribute(TestMessageDrivenManagementMBean.OBJECT_NAME, "Messages");
               synchronized (BasicMessageDrivenUnitTest.this)
               {
                  if (running == false)
                     break;
                  if (result.size() > 0)
                  {
                     messages.addAll(result);
                     BasicMessageDrivenUnitTest.this.notifyAll();
                  }
                  BasicMessageDrivenUnitTest.this.wait(WAIT_TIME);
               }
            }
         }
         catch (Throwable t)
         {
            getLog().error("Error in receiver thread " + thread, t);
         }
         
         synchronized (BasicMessageDrivenUnitTest.this)
         {
            thread = null;
            BasicMessageDrivenUnitTest.this.notifyAll();
         }
      }
   }
}
